/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 This firmware is a mashup between the original Vellemann software, MARLIN and my own extensions

 It uses the hardware capabilities of the printer boards as much as possible.
 It is rewritten to easily implement service routines for adjustment and omtimisation of the pronter parameters.
 The SD library was exchanged to be capable for long filenames.
 The internationalisation is modified.
 The temperature control is implemented as PID control only.
 */

#include "MarlinHOH.h"

#ifdef ENABLE_AUTO_BED_LEVELING
#include "vector_3.h"
  #ifdef AUTO_BED_LEVELING_GRID
    #include "qr_solve.h"
  #endif
#endif // ENABLE_AUTO_BED_LEVELING

#include "stepper.h"
#include "temperature.h"
#include "motion_control.h"
#include "cardreader.h"
#include "watchdog.h"
#include "language.h"
#include "pins_arduino.h"
#include "math.h"
#include "parameter.h"
#include "printer.h"
#include "PositionControl.h"
#include "SdFat.h"
#include "Service.h"
#include "ultralcd.h"
#include "planner.h"

#ifdef BLINKM
#include "BlinkM.h"
#include "Wire.h"
#endif

#if NUM_SERVOS > 0
#include "Servo.h"
#endif

#if defined(DIGIPOTSS_PIN) && DIGIPOTSS_PIN > -1
#include <SPI.h>
#endif

#define VERSION_STRING  "Version 1.0.0\n"

// look here for descriptions of G-codes: http://linuxcnc.org/handbook/gcode/g-code.html
// http://objects.reprap.org/wiki/Mendel_User_Manual:_RepRapGCodes

//Implemented Codes
//-------------------
// G0  -> G1
// G1  - Coordinated Movement X Y Z E
// G2  - CW ARC
// G3  - CCW ARC
// G4  - Dwell S<seconds> or P<milliseconds>
// G10 - retract filament according to settings of M207
// G11 - retract recover filament according to settings of M208
// G28 - Home all Axis
// G29 - Detailed Z-Probe, probes the bed at 3 or more points.  Will fail if you haven't homed yet.
// G30 - Single Z Probe, probes bed at current XY location.
// G90 - Use Absolute Coordinates
// G91 - Use Relative Coordinates
// G92 - Set current position to coordinates given

// M Codes
// M0   - Unconditional stop - Wait for user to press a button on the LCD (Only if ULTRA_LCD is enabled)
// M1   - Same as M0
// M17  - Enable/Power all stepper motors
// M18  - Disable all stepper motors; same as M84
// M20  - List SD card
// M21  - Init SD card
// M22  - Release SD card
// M23  - Select SD file (M23 filename.g)
// M24  - Start/resume SD print
// M25  - Pause SD print
// M26  - Set SD position in bytes (M26 S12345)
// M27  - Report SD print status
// M28  - Start SD write (M28 filename.g)
// M29  - Stop SD write
// M30  - Delete file from SD (M30 filename.g)
// M31  - Output time since last M109 or SD card start to serial
// M32  - Select file and start SD print (Can be used _while_ printing from SD card files):
//        syntax "M32 /path/filename#", or "M32 S<startpos bytes> !filename#"
//        Call gcode file : "M32 P !filename#" and return to caller file after finishing (similar to #include).
//        The '#' is necessary when calling from within sd files, as it stops buffer prereading
// M42  - Change pin status via gcode Use M42 Px Sy to set pin x to value y, when omitting Px the onboard led will be used.
// M80  - Turn on Power Supply
// M81  - Turn off Power Supply
// M82  - Set E codes absolute (default)
// M83  - Set E codes relative while in Absolute Coordinates (G90) mode
// M84  - Disable steppers until next move,
//        or use S<seconds> to specify an inactivity timeout, after which the steppers will be disabled.  S0 to disable the timeout.
// M85  - Set inactivity shutdown timer with parameter S<seconds>. To disable set zero (default)
// M92  - Set axis_steps_per_unit - same syntax as G92
// M104 - Set extruder target temp
// M105 - Read current temp
// M106 - Fan on
// M107 - Fan off
// M109 - Sxxx Wait for extruder current temp to reach target temp. Waits only when heating
//        Rxxx Wait for extruder current temp to reach target temp. Waits when heating and cooling
//        IF AUTOTEMP is enabled, S<mintemp> B<maxtemp> F<factor>. Exit autotemp by any M109 without F
// M114 - Output current position to serial port
// M115 - Capabilities string
// M117 - display message
// M119 - Output Endstop status to serial port
// M126 - Solenoid Air Valve Open (BariCUDA support by jmil)
// M127 - Solenoid Air Valve Closed (BariCUDA vent to atmospheric pressure by jmil)
// M128 - EtoP Open (BariCUDA EtoP = electricity to air pressure transducer by jmil)
// M129 - EtoP Closed (BariCUDA EtoP = electricity to air pressure transducer by jmil)
// M140 - Set bed target temp
// M150 - Set BlinkM Color Output R: Red<0-255> U(!): Green<0-255> B: Blue<0-255> over i2c, G for green does not work.
// M190 - Sxxx Wait for bed current temp to reach target temp. Waits only when heating
//        Rxxx Wait for bed current temp to reach target temp. Waits when heating and cooling
// M200 D<millimeters>- set filament diameter and set E axis units to cubic millimeters (use S0 to set back to millimeters).
// M201 - Set max acceleration in units/s^2 for print moves (M201 X1000 Y1000)
// M202 - Set max acceleration in units/s^2 for travel moves (M202 X1000 Y1000) Unused in Marlin!!
// M203 - Set maximum feedrate that your machine can sustain (M203 X200 Y200 Z300 E10000) in mm/sec
// M204 - Set default acceleration: S normal moves T filament only moves (M204 S3000 T7000) in mm/sec^2  also sets minimum segment time in ms (B20000) to prevent buffer under-runs and M20 minimum feedrate
// M205 -  advanced settings:  minimum travel speed S=while printing T=travel only,  B=minimum segment time X= maximum xy jerk, Z=maximum Z jerk, E=maximum E jerk
// M206 - set additional homing offset
// M207 - set retract length S[positive mm] F[feedrate mm/min] Z[additional zlift/hop], stays in mm regardless of M200 setting
// M208 - set recover=unretract length S[positive mm surplus to the M207 S*] F[feedrate mm/sec]
// M209 - S<1=true/0=false> enable automatic retract detect if the slicer did not support G10/11: every normal extrude-only move will be classified as retract depending on the direction.
// M214 - set advanced parameters: T[tool <default: active extruder>] K[magic number]
// M218 - set hotend offset (in mm): T<extruder_number> X<offset_on_X> Y<offset_on_Y>
// M220 S<factor in percent>- set speed factor override percentage
// M221 S<factor in percent>- set extrude factor override percentage
// M226 P<pin number> S<pin state>- Wait until the specified pin reaches the state required
// M240 - Trigger a camera to take a photograph
// M250 - Set LCD contrast C<contrast value> (value 0..63)
// M280 - set servo position absolute. P: servo index, S: angle or microseconds
// M300 - Play beep sound S<frequency Hz> P<duration ms>
// M301 - Set PID parameters P I and D
// M302 - Allow cold extrudes, or set the minimum extrude S<temperature>.
// M303 - PID relay autotune S<temperature> sets the target temperature. (default target temperature = 150C)
// M304 - Set bed PID parameters P I and D
// M400 - Finish all moves
// M401 - Lower z-probe if present
// M402 - Raise z-probe if present
// M500 - stores parameters in EEPROM
// M501 - reads parameters from EEPROM (if you need reset them after you changed them temporarily).
// M502 - reverts to the default "factory settings".  You still need to store them in EEPROM afterwards if you want to.
// M503 - print the current settings (from memory not from EEPROM)
// M540 - Use S[0|1] to enable or disable the stop SD card print on endstop hit (requires ABORT_ON_ENDSTOP_HIT_FEATURE_ENABLED)
// M600 - Pause for filament change X[pos] Y[pos] Z[relative lift] E[initial retract] L[later retract distance for removal]
// M665 - set delta configurations
// M666 - set delta endstop adjustment
// M605 - Set dual x-carriage movement mode: S<mode> [ X<duplication x-offset> R<duplication temp offset> ]
// M907 - Set digital trimpot motor current using axis codes.
// M908 - Control digital trimpot directly.
// M350 - Set microstepping mode.
// M351 - Toggle MS1 MS2 pins directly.
// M928 - Start SD logging (M928 filename.g) - ended by M29
// M999 - Restart after being stopped by error

//Stepper Movement Variables

//===========================================================================
//=============================imported variables============================
//===========================================================================


//===========================================================================
//=============================public variables=============================
//===========================================================================


#ifdef SDSUPPORT
CardReader sdcard;
#endif

class MarlinSerial MSerial;
class parameter config;	// configuration settings stored in EEPROM
class Planner plan;
class stepper motor; // new stepper motor class
class printer my_printer; // Basic I/O interface
class Service service;
class temperatur_Control heater;
class Position_Control Position;
//class Ultra_LCD_HITACHI_HD44780 LCD;
float homing_feedrate[] = HOMING_FEEDRATE;
bool axis_relative_modes[] = AXIS_RELATIVE_MODES;
int feedmultiply=100; //100->1 200->2

int extrudemultiply=100; //100->1 200->2
int extruder_multiply[EXTRUDERS] = ARRAY_BY_EXTRUDERS1(100);
float volumetric_multiplier[EXTRUDERS] = ARRAY_BY_EXTRUDERS1(1.0) ;
float extruder_advance_k = 0;
//float current_position[NUM_AXIS] = { 0.0, 0.0, 0.0, 0.0 };
// float add_homeing[3]={0,0,0}; class parameter
#ifdef DELTA
float endstop_adj[3]={0,0,0};
#endif
float min_pos[3] = { X_MIN_POS, Y_MIN_POS, Z_MIN_POS };
float max_pos[3] = { X_MAX_POS, Y_MAX_POS, Z_MAX_POS };
//bool axis_known_position[3] = {false, false, false};
float zprobe_zoffset;

unsigned long codenum; //throw away variable

uint8_t active_extruder = 0;

#ifdef SERVO_ENDSTOPS
  int servo_endstops[] = SERVO_ENDSTOPS;
  int servo_endstop_angles[] = SERVO_ENDSTOP_ANGLES;
#endif
#ifdef BARICUDA
int ValvePressure=0;
int EtoPPressure=0;
#endif

#ifdef FWRETRACT
  bool autoretract_enabled=false;
  bool retracted=false;
  float retract_length = RETRACT_LENGTH;
  float retract_feedrate = RETRACT_FEEDRATE;
  float retract_zlift = RETRACT_ZLIFT;
  float retract_recover_length = RETRACT_RECOVER_LENGTH;
  float retract_recover_feedrate = RETRACT_RECOVER_FEEDRATE;
#endif

//#ifdef ULTIPANEL
//  #ifdef PS_DEFAULT_OFF
//    bool powersupply = false;
//  #else
//	  bool powersupply = true;
//  #endif
//#endif

#ifdef DELTA
  float delta[3] = {0.0, 0.0, 0.0};
  #define SIN_60 0.8660254037844386
  #define COS_60 0.5
  // these are the default values, can be overriden with M665
  float delta_radius= DELTA_RADIUS;
  float delta_tower1_x= -SIN_60*delta_radius; // front left tower
  float delta_tower1_y= -COS_60*delta_radius;	   
  float delta_tower2_x=  SIN_60*delta_radius; // front right tower
  float delta_tower2_y= -COS_60*delta_radius;	   
  float delta_tower3_x= 0.0;                  // back middle tower
  float delta_tower3_y= delta_radius;
  float delta_diagonal_rod= DELTA_DIAGONAL_ROD;
  float delta_diagonal_rod_2= sq(delta_diagonal_rod);
  float delta_segments_per_second= DELTA_SEGMENTS_PER_SECOND;
#endif					
//
// queue handling variables
//
  int cmd_queue_index_r = 0;
  int cmd_queue_index_w = 0;
  int commands_in_queue = 0;

//===========================================================================
//=============================Private Variables=============================
//===========================================================================
const char axis_codes[NUM_AXIS] = {'X', 'Y', 'Z', 'E'};
//static float destination[NUM_AXIS] = {  0.0, 0.0, 0.0, 0.0};
static float offset[3] = {0.0, 0.0, 0.0};
//static bool home_all_axis = true;
bool homex , homey, homez;

static long gcode_N, gcode_LastN, Stopped_gcode_LastN = 0;

static bool relative_mode = false;  //Determines Absolute or Relative Coordinates

static char cmdbuffer[BUFSIZE][MAX_CMD_SIZE];
static bool fromsd[BUFSIZE]; // internal comand flag: set when command is not given by serial input

//static int i = 0;
static char serial_char;
static int serial_count = 0;
static int service_count = 0;
static boolean comment_mode = false;
static char *strchr_pointer; // just a pointer to find chars in the command string like X, Y, Z, E, etc

const int sensitive_pins[] = SENSITIVE_PINS; // Sensitive pin list for M42

static uint8_t target_extruder;

bool CooldownNoWait = true;
bool target_direction;

#if defined PID_ADD_EXTRUSION_RATE
  int lpq_len = LPQ_LEN_DEFAULT;
#endif


//Insert variables if CHDK is defined
#ifdef CHDK
unsigned long chdkHigh = 0;
boolean chdkActive = false;
#endif

//===========================================================================
//=============================Routines======================================
//===========================================================================

void get_arc_coordinates();
bool setTargetedHotend(int code);

void serial_echopair_P(const char* s_P, int v)           { serialprintPGM(s_P); SERIAL_ECHO(v); }
void serial_echopair_P(const char* s_P, long v)          { serialprintPGM(s_P); SERIAL_ECHO(v); }
void serial_echopair_P(const char* s_P, float v)         { serialprintPGM(s_P); SERIAL_ECHO(v); }
void serial_echopair_P(const char* s_P, double v)        { serialprintPGM(s_P); SERIAL_ECHO(v); }
void serial_echopair_P(const char* s_P, unsigned long v) { serialprintPGM(s_P); SERIAL_ECHO(v); }

extern "C"{
  extern unsigned int __bss_end;
  extern unsigned int __heap_start;
  extern void *__brkval;

  int freeMemory() {
    int free_memory;

    if((int)__brkval == 0)
      free_memory = ((int)&free_memory) - ((int)&__bss_end);
    else
      free_memory = ((int)&free_memory) - ((int)__brkval);

    return free_memory;
  }
}

//adds an command to the main command buffer
//thats really done in a non-safe way.
//needs overworking someday
//void enquecommand(const char *cmd)
//{
//  if(commands_in_queue < BUFSIZE)
//  {
//    //this is dangerous if a mixing of serial and this happens
//    strcpy(&(cmdbuffer[cmd_queue_index_w][0]),cmd);
////    SERIAL_ECHO_START;
//    my_printer.echo("enqueing \"");
//    SERIAL_ECHO(cmdbuffer[cmd_queue_index_w]);
//    SERIAL_ECHOLNPGM("\"");
//    cmd_queue_index_w= (cmd_queue_index_w + 1)%BUFSIZE;
//    commands_in_queue += 1;
//  }
//}
//
//void enquecommand_P(const char *cmd)
//{
//  if(commands_in_queue < BUFSIZE)
//  {
//    //this is dangerous if a mixing of serial and this happens
//    strcpy_P(&(cmdbuffer[cmd_queue_index_w][0]),cmd);
////    SERIAL_ECHO_START;
//    my_printer.echo(PSTR("enqueing \\"));
// //   SERIAL_ECHO(cmdbuffer[cmd_queue_index_w]);
//    my_printer.message(cmdbuffer[cmd_queue_index_w]);
// //   SERIAL_ECHOLNPGM("\"");
//    my_printer.message(PSTR("\\\n"));
//    cmd_queue_index_w= (cmd_queue_index_w + 1)%BUFSIZE;
//    commands_in_queue += 1;
//  }
//}

void setup_killpin()
{
  #if defined(KILL_PIN) && KILL_PIN > -1
    pinMode(KILL_PIN,INPUT);
    WRITE(KILL_PIN,HIGH);
  #endif
}

void setup_photpin()
{
  #if defined(PHOTOGRAPH_PIN) && PHOTOGRAPH_PIN > -1
    SET_OUTPUT(PHOTOGRAPH_PIN);
    WRITE(PHOTOGRAPH_PIN, LOW);
  #endif
}

void setup_powerhold()
{
  #if defined(SUICIDE_PIN) && SUICIDE_PIN > -1
    SET_OUTPUT(SUICIDE_PIN);
    WRITE(SUICIDE_PIN, HIGH);
  #endif
  #if defined(PS_ON_PIN) && PS_ON_PIN > -1
    SET_OUTPUT(PS_ON_PIN);
	#if defined(PS_DEFAULT_OFF)
	  WRITE(PS_ON_PIN, PS_ON_ASLEEP);
    #else
	  WRITE(PS_ON_PIN, PS_ON_AWAKE);
	#endif
  #endif
}

void suicide()
{
  #if defined(SUICIDE_PIN) && SUICIDE_PIN > -1
    SET_OUTPUT(SUICIDE_PIN);
    WRITE(SUICIDE_PIN, LOW);
  #endif
}

void servo_init()
{
  #if (NUM_SERVOS >= 1) && defined(SERVO0_PIN) && (SERVO0_PIN > -1)
    servos[0].attach(SERVO0_PIN);
  #endif
  #if (NUM_SERVOS >= 2) && defined(SERVO1_PIN) && (SERVO1_PIN > -1)
    servos[1].attach(SERVO1_PIN);
  #endif
  #if (NUM_SERVOS >= 3) && defined(SERVO2_PIN) && (SERVO2_PIN > -1)
    servos[2].attach(SERVO2_PIN);
  #endif
  #if (NUM_SERVOS >= 4) && defined(SERVO3_PIN) && (SERVO3_PIN > -1)
    servos[3].attach(SERVO3_PIN);
  #endif
  #if (NUM_SERVOS >= 5)
    #error "TODO: enter initalisation code for more servos"
  #endif

  // Set position of Servo Endstops that are defined
  #ifdef SERVO_ENDSTOPS
  for(int8_t i = 0; i < 3; i++)
  {
    if(servo_endstops[i] > -1) {
      servos[servo_endstops[i]].write(servo_endstop_angles[i * 2 + 1]);
    }
  }
  #endif

  #if defined (ENABLE_AUTO_BED_LEVELING) && (PROBE_SERVO_DEACTIVATION_DELAY > 0)
  delay(PROBE_SERVO_DEACTIVATION_DELAY);
  servos[servo_endstops[Z_AXIS]].detach();
  #endif
}

void setup()
{
	my_printer.Init();
	my_printer.message(PSTR("start\n"));

	// Check startup - does nothing if bootloader sets MCUSR to 0
	//  byte mcu = MCUSR;
	//  if(mcu & 1) SERIAL_ECHOLNPGM(MSG_POWERUP);
	//  if(mcu & 2) SERIAL_ECHOLNPGM(MSG_EXTERNAL_RESET);
	//  if(mcu & 4) SERIAL_ECHOLNPGM(MSG_BROWNOUT_RESET);
	//  if(mcu & 8) SERIAL_ECHOLNPGM(MSG_WATCHDOG_RESET);
	//  if(mcu & 32) SERIAL_ECHOLNPGM(MSG_SOFTWARE_RESET);
	//  MCUSR=0;

	//  my_printer.message(PSTR(MSG_MARLIN));
	my_printer.message(T(MSG_Marlin));

	my_printer.message(PSTR(VERSION_STRING));

#ifdef STRING_VERSION_CONFIG_H
#ifdef STRING_CONFIG_H_AUTHOR
	//      SERIAL_ECHO_START;
	my_printer.echo(T(MSG_Configuration_Ver));
	my_printer.message(PSTR(STRING_VERSION_CONFIG_H));
	my_printer.message(T(MSG_Author));
	my_printer.message(PSTR(STRING_CONFIG_H_AUTHOR));
	my_printer.message(T(MSG_Compiled));
	my_printer.message(PSTR(__DATE__));
#endif
#endif
//  SERIAL_ECHO_START;
	my_printer.f_message(T(MSG_Free_Memory),freeMemory());
	//  SERIAL_ECHO(freeMemory());
	//  SERIAL_ECHOPGM(MSG_PLANNER_BUFFER_BYTES);
	my_printer.f_message(T(MSG_Planner_Buffer_Bytes),(int)sizeof(move_block)*BLOCK_BUFFER_SIZE);
	//  SERIAL_ECHOLN((int)sizeof(move_block)*BLOCK_BUFFER_SIZE);
	for(int8_t i = 0; i < BUFSIZE; i++)
	{
		fromsd[i] = false;  // clear internal command flag
	}

  // loads data from EEPROM if available else uses defaults (and resets step acceleration rate)
  config.Load_EEPROM();

  //tp_init();    // Initialize temperature loop
  heater.Init();
  plan.Init();  // Initialize planner;
  watchdog_init();
  motor.Initialize(); // Initialize stepper, this enables interrupts!
  setup_photpin();
  servo_init();

//  LCD.Init();
//  LCD.Firmware_screen();
  lcd_splashscreen();
  lcd_init();

  #if defined(CONTROLLERFAN_PIN) && CONTROLLERFAN_PIN > -1
    SET_OUTPUT(CONTROLLERFAN_PIN); //Set pin used for driver cooling fan
  #endif

  #ifdef DIGIPOT_I2C
    digipot_i2c_init();
  #endif

#ifdef PRESSURE_ADVANCE
    for(int8_t i = 0; i < EXTRUDERS; i++)
    {
    	config.advance_factor[i] = 0.0;
    	config.advance_delay[i]= 0;
    }

#endif

}


void loop() {
	if (commands_in_queue < (BUFSIZE - 1))
		get_command();
#ifdef SDSUPPORT
	sdcard.checkautostart(false);
#endif
	if (commands_in_queue) {
#ifdef SDSUPPORT
		if (sdcard.saving) {
			if (strstr_P(cmdbuffer[cmd_queue_index_r], PSTR("M29")) == NULL) {
				// M29 closes the file
				sdcard.closefile();
				my_printer.message(T(MSG_File_Saved));
			} else {
				sdcard.write_command(cmdbuffer[cmd_queue_index_r]);
				if (sdcard.logging) {
					process_commands();
				} else {
//					SERIAL_PROTOCOLLNPGM(MSG_OK);
					  my_printer.message(T(MSG_OK));
				}
			}
		} else {
			process_commands();
		}
#else
		process_commands();
#endif //SDSUPPORT
		commands_in_queue = (commands_in_queue - 1);
		cmd_queue_index_r = (cmd_queue_index_r + 1) % BUFSIZE;
	}
	//check heater every n milliseconds
	motor.CheckHitEndstops();
	my_printer.Go_Idle();
}

void get_command()
{
	while ( MSerial.available() > 0 && commands_in_queue < BUFSIZE)
	{
		serial_char = MSerial.read();
		if (serial_char == '\n' || serial_char == '\r'
				|| (serial_char == ':' && comment_mode == false)
				|| serial_count >= (MAX_CMD_SIZE - 1))
		{
			if (!serial_count)
			{ //if empty line
				comment_mode = false; //for new command
				return;
			}
			cmdbuffer[cmd_queue_index_w][serial_count] = 0; //terminate string

			if (!comment_mode)
			{
				comment_mode = false; //for new command
				fromsd[cmd_queue_index_w] = false;
				strchr_pointer = strchr(cmdbuffer[cmd_queue_index_w], 'N');
				if (strchr_pointer != NULL)
				{
					gcode_N = (strtol(
							&cmdbuffer[cmd_queue_index_w][strchr_pointer
														  - cmdbuffer[cmd_queue_index_w] + 1], NULL,
														  10));
					if (gcode_N != gcode_LastN + 1
							&& (strstr_P(cmdbuffer[cmd_queue_index_w],
									PSTR("M110")) == NULL))
					{
						my_printer.error();
						my_printer.f_message(T(MSG_Err_Line_No),(int)gcode_LastN);
//						SERIAL_ERRORLN(gcode_LastN);
//						MSerial.println(gcode_N);
						FlushSerialRequestResend();
						serial_count = 0;
						return;
					}
					strchr_pointer = strchr(cmdbuffer[cmd_queue_index_w],'*');
					if (strchr_pointer != NULL)
					{
						byte checksum = 0;
						byte count = 0;
						while (cmdbuffer[cmd_queue_index_w][count] != '*')
							checksum = checksum
							^ cmdbuffer[cmd_queue_index_w][count++];

						if ((int) (strtod(
								&cmdbuffer[cmd_queue_index_w][strchr_pointer
															  - cmdbuffer[cmd_queue_index_w] + 1],
															  NULL)) != checksum)
						{
							my_printer.error();
							my_printer.f_message(T(MSG_Err_Checksum_Mismatch),(int)gcode_LastN);
//							SERIAL_ERRORLN(gcode_LastN);
							FlushSerialRequestResend();
							serial_count = 0;
							return;
						}
						//if no errors, continue parsing
					}
					else
					{
						my_printer.error();
						my_printer.f_message(T(MSG_Err_No_Checksum),(int)gcode_LastN);
//						SERIAL_ERRORLN(gcode_LastN);
						FlushSerialRequestResend();
						serial_count = 0;
						return;
					}

					gcode_LastN = gcode_N;
					//if no errors, continue parsing
				}
				else  // if we don't receive 'N' but still see '*'
				{
					if ((strchr(cmdbuffer[cmd_queue_index_w], '*') != NULL))
					{
						my_printer.error();
						my_printer.f_message(T(MSG_Err_No_Linenumber_With_Checksum),(int)gcode_LastN);
//						SERIAL_ERRORLN(gcode_LastN);
						serial_count = 0;
						return;
					}
				}
				if ((strchr(cmdbuffer[cmd_queue_index_w], 'G') != NULL))
				{
					strchr_pointer = strchr(cmdbuffer[cmd_queue_index_w], 'G');
					switch ((int) ((strtod(
							&cmdbuffer[cmd_queue_index_w][strchr_pointer
														  - cmdbuffer[cmd_queue_index_w] + 1], NULL))))
					{
					case 0:
					case 1:
					case 2:
					case 3:
						if (my_printer.Stopped == false)
						{ // If printer is stopped by an error the G[0-3] codes are ignored.
#ifdef SDSUPPORT
							if (sdcard.saving)
								break;
#endif //SDSUPPORT
							my_printer.message(T(MSG_OK));
						}
						else
						{
							my_printer.error_ln(T(MSG_Err_Stopped));
							lcd_setstatuspgm(T(MSG_Stopped));
						}
						break;
					default:
						break;
					}

				}
				cmd_queue_index_w = (cmd_queue_index_w + 1) % BUFSIZE;
				commands_in_queue += 1;
			}
			serial_count = 0; //clear buffer
		}
		else if (serial_char == '\\')
		{  // Handle escapes
			if (MSerial.available() > 0 && commands_in_queue < BUFSIZE)
			{
				// if we have one more character, copy it over
				serial_char = MSerial.read();
				cmdbuffer[cmd_queue_index_w][serial_count++] = serial_char;
			}
			// otherwise do nothing
		}
		else
		{
			if (serial_char == ';')
				comment_mode = true;
			if (!comment_mode)
				cmdbuffer[cmd_queue_index_w][serial_count++] = serial_char;
		}
	}

	if (service.Printing() )
	{
		while (!service.Is_EOF() && commands_in_queue < BUFSIZE)
		{
			serial_char = service.Get();
			MSerial.print (serial_char,HEX);
			if (serial_char == '\n' || serial_char == '\r'
		//			|| (serial_char == ':' && comment_mode == false)
					|| service_count >= (MAX_CMD_SIZE - 1))
			{
				if (!service_count)
				{ //if empty line
					comment_mode = false; //for new command
					return;
				}
				cmdbuffer[cmd_queue_index_w][service_count] = 0; //terminate string
				fromsd[cmd_queue_index_w] = true;
				cmd_queue_index_w = (cmd_queue_index_w + 1) % BUFSIZE;
				commands_in_queue += 1;
				service_count = 0; //clear buffer
				comment_mode = false;
			}
			else
			{
				if (serial_char == ';')
					comment_mode = true;
				if (!comment_mode)
					cmdbuffer[cmd_queue_index_w][service_count++] = serial_char;
			}
		}
	}
#ifdef SDSUPPORT
	if ((sdcard.Status != Card_printing) || (serial_count != 0))
//		if ((sdcard.Status == Card_OK) )
	{
		return;
	}

	//'#' stops reading from SD to the buffer prematurely, so procedural macro calls are possible
	// if it occurs, stop_buffering is triggered and the buffer is ran dry.
	// this character _can_ occur in serial com, due to checksums. however, no checksums are used in SD printing

	static bool stop_buffering = false;
	if (commands_in_queue == 0)
		stop_buffering = false;

	while (!sdcard.eof() && commands_in_queue < BUFSIZE-1 && !stop_buffering)
	{
		int16_t n = sdcard.get();
		serial_char = (char) n;
//		MYSERIAL.print (serial_char);
		if (serial_char == '\n' || serial_char == '\r'
				|| (serial_char == '#' && comment_mode == false)
				|| (serial_char == ':' && comment_mode == false)
				|| serial_count >= (MAX_CMD_SIZE - 1) || n == -1)
		{
			if (sdcard.eof())
			{
				my_printer.message(T(MSG_File_Printed));
				my_printer.print_job_timer.stop();
				//				stoptime=millis();
				//				char time[30];
				//				unsigned long t=(stoptime-starttime)/1000;
				char time[30];
				uint16_t t = my_printer.print_job_timer.duration();
				int hours, minutes;
				minutes = (t / 60) % 60;
				hours = t / 60 / 60;
				sprintf_P(time, PSTR("%i hours %i minutes"), hours, minutes);
				my_printer.echo();
				SERIAL_ECHOLN(time);
	//			lcd_setstatus(time);


			}
//			if (serial_char == '#')
//				stop_buffering = true;

			comment_mode = false; //for new command
			if (!serial_count)
			{
				return; //if empty line
			}
			cmdbuffer[cmd_queue_index_w][serial_count] = 0; //terminate string
			fromsd[cmd_queue_index_w] = true;
			commands_in_queue += 1;
			cmd_queue_index_w = (cmd_queue_index_w + 1) % BUFSIZE;
			serial_count = 0; //clear buffer
		}
		else
		{
			if (serial_char == ';')
				comment_mode = true;
			if (!comment_mode)
				cmdbuffer[cmd_queue_index_w][serial_count++] = serial_char;
		}
	}
	if (sdcard.eof())
	{
		sdcard.printingHasFinished();
		sdcard.checkautostart(true);
	}

#endif //SDSUPPORT

}


float code_value()
{
	return (strtod(&cmdbuffer[cmd_queue_index_r][strchr_pointer - cmdbuffer[cmd_queue_index_r] + 1], NULL));
}

long code_value_long()
{
	return (strtol(&cmdbuffer[cmd_queue_index_r][strchr_pointer - cmdbuffer[cmd_queue_index_r] + 1], NULL, 10));
}

bool code_seen(char code)
{
	strchr_pointer = strchr(cmdbuffer[cmd_queue_index_r], code);
	return (strchr_pointer != NULL);  //Return True if a character was found
}









#ifdef DUAL_X_CARRIAGE
  #if EXTRUDERS == 1 || defined(COREXY) \
      || !defined(X2_ENABLE_PIN) || !defined(X2_STEP_PIN) || !defined(X2_DIR_PIN) \
      || !defined(X2_HOME_POS) || !defined(X2_MIN_POS) || !defined(X2_MAX_POS) \
      || !defined(X_MAX_PIN) || X_MAX_PIN < 0
    #error "Missing or invalid definitions for DUAL_X_CARRIAGE mode."
  #endif
  #if X_HOME_DIR != -1 || X2_HOME_DIR != 1
    #error "Please use canonical x-carriage assignment" // the x-carriages are defined by their homing directions
  #endif

#define DXC_FULL_CONTROL_MODE 0
#define DXC_AUTO_PARK_MODE    1
#define DXC_DUPLICATION_MODE  2
static int dual_x_carriage_mode = DEFAULT_DUAL_X_CARRIAGE_MODE;

static float x_home_pos(int extruder) {
  if (extruder == 0)
    return base_home_pos(X_AXIS) + add_homeing[X_AXIS];
  else
    // In dual carriage mode the extruder offset provides an override of the
    // second X-carriage offset when homed - otherwise X2_HOME_POS is used.
    // This allow soft recalibration of the second extruder offset position without firmware reflash
    // (through the M218 command).
    return (extruder_offset[X_AXIS][1] > 0) ? extruder_offset[X_AXIS][1] : X2_HOME_POS;
}

static int x_home_dir(int extruder) {
  return (extruder == 0) ? X_HOME_DIR : X2_HOME_DIR;
}

static float inactive_extruder_x_pos = X2_MAX_POS; // used in mode 0 & 1
static bool active_extruder_parked = false; // used in mode 1 & 2
static float raised_parked_position[NUM_AXIS]; // used in mode 1
static unsigned long delayed_move_time = 0; // used in mode 1
static float duplicate_extruder_x_offset = DEFAULT_DUPLICATION_X_OFFSET; // used in mode 2
static float duplicate_extruder_temp_offset = 0; // used in mode 2
bool extruder_duplication_enabled = false; // used in mode 2
#endif //DUAL_X_CARRIAGE



#ifdef ENABLE_AUTO_BED_LEVELING
#ifdef AUTO_BED_LEVELING_GRID
static void set_bed_level_equation_lsq(double *plane_equation_coefficients)
{
    vector_3 planeNormal = vector_3(-plane_equation_coefficients[0], -plane_equation_coefficients[1], 1);
    planeNormal.debug("planeNormal");
    plan_bed_level_matrix = matrix_3x3::create_look_at(planeNormal);
    //bedLevel.debug("bedLevel");

    //plan_bed_level_matrix.debug("bed level before");
    //vector_3 uncorrected_position = plan_get_position_mm();
    //uncorrected_position.debug("position before");

    vector_3 corrected_position = plan_get_position();
//    corrected_position.debug("position after");
    current_position[X_AXIS] = corrected_position.x;
    current_position[Y_AXIS] = corrected_position.y;
    current_position[Z_AXIS] = corrected_position.z;

    // but the bed at 0 so we don't go below it.
    current_position[Z_AXIS] = zprobe_zoffset; // in the lsq we reach here after raising the extruder due to the loop structure

    plan_set_position(current_position[X_AXIS], current_position[Y_AXIS], current_position[Z_AXIS], current_position[E_AXIS]);
}

#else // not AUTO_BED_LEVELING_GRID

static void set_bed_level_equation_3pts(float z_at_pt_1, float z_at_pt_2, float z_at_pt_3) {

    plan_bed_level_matrix.set_to_identity();

    vector_3 pt1 = vector_3(ABL_PROBE_PT_1_X, ABL_PROBE_PT_1_Y, z_at_pt_1);
    vector_3 pt2 = vector_3(ABL_PROBE_PT_2_X, ABL_PROBE_PT_2_Y, z_at_pt_2);
    vector_3 pt3 = vector_3(ABL_PROBE_PT_3_X, ABL_PROBE_PT_3_Y, z_at_pt_3);

    vector_3 from_2_to_1 = (pt1 - pt2).get_normal();
    vector_3 from_2_to_3 = (pt3 - pt2).get_normal();
    vector_3 planeNormal = vector_3::cross(from_2_to_1, from_2_to_3).get_normal();
    planeNormal = vector_3(planeNormal.x, planeNormal.y, abs(planeNormal.z));

    plan_bed_level_matrix = matrix_3x3::create_look_at(planeNormal);

    vector_3 corrected_position = plan_get_position();
    current_position[X_AXIS] = corrected_position.x;
    current_position[Y_AXIS] = corrected_position.y;
    current_position[Z_AXIS] = corrected_position.z;

    // put the bed at 0 so we don't go below it.
    current_position[Z_AXIS] = zprobe_zoffset;

    plan_set_position(current_position[X_AXIS], current_position[Y_AXIS], current_position[Z_AXIS], current_position[E_AXIS]);

}

#endif // AUTO_BED_LEVELING_GRID

static void run_z_probe() {
    plan_bed_level_matrix.set_to_identity();
    feedrate = homing_feedrate[Z_AXIS];

    // move down until you find the bed
    float zPosition = -10;
    plan_buffer_line(current_position[X_AXIS], current_position[Y_AXIS], zPosition, current_position[E_AXIS], feedrate/60, active_extruder);
    st_synchronize();

        // we have to let the planner know where we are right now as it is not where we said to go.
    zPosition = st_get_position_mm(Z_AXIS);
    plan_set_position(current_position[X_AXIS], current_position[Y_AXIS], zPosition, current_position[E_AXIS]);

    // move up the retract distance
    zPosition += home_retract_mm(Z_AXIS);
    plan_buffer_line(current_position[X_AXIS], current_position[Y_AXIS], zPosition, current_position[E_AXIS], feedrate/60, active_extruder);
    st_synchronize();

    // move back down slowly to find bed
    feedrate = homing_feedrate[Z_AXIS]/4;
    zPosition -= home_retract_mm(Z_AXIS) * 2;
    plan_buffer_line(current_position[X_AXIS], current_position[Y_AXIS], zPosition, current_position[E_AXIS], feedrate/60, active_extruder);
    st_synchronize();

    current_position[Z_AXIS] = st_get_position_mm(Z_AXIS);
    // make sure the planner knows where we are as it may be a bit different than we last said to move to
    plan_set_position(current_position[X_AXIS], current_position[Y_AXIS], current_position[Z_AXIS], current_position[E_AXIS]);
}

static void do_blocking_move_to(float x, float y, float z) {
    float oldFeedRate = feedrate;

    feedrate = XY_TRAVEL_SPEED;

    current_position[X_AXIS] = x;
    current_position[Y_AXIS] = y;
    current_position[Z_AXIS] = z;
    plan_buffer_line(current_position[X_AXIS], current_position[Y_AXIS], current_position[Z_AXIS], current_position[E_AXIS], feedrate/60, active_extruder);
    st_synchronize();

    feedrate = oldFeedRate;
}

static void do_blocking_move_relative(float offset_x, float offset_y, float offset_z) {
    do_blocking_move_to(current_position[X_AXIS] + offset_x, current_position[Y_AXIS] + offset_y, current_position[Z_AXIS] + offset_z);
}

static void setup_for_endstop_move() {
    saved_feedrate = feedrate;
    saved_feedmultiply = feedmultiply;
    feedmultiply = 100;
    previous_millis_cmd = millis();

    enable_endstops(true);
}

static void clean_up_after_endstop_move() {
#ifdef ENDSTOPS_ONLY_FOR_HOMING
    enable_endstops(false);
#endif

    feedrate = saved_feedrate;
    feedmultiply = saved_feedmultiply;
    previous_millis_cmd = millis();
}

static void engage_z_probe() {
    // Engage Z Servo endstop if enabled
    #ifdef SERVO_ENDSTOPS
    if (servo_endstops[Z_AXIS] > -1) {
#if defined (ENABLE_AUTO_BED_LEVELING) && (PROBE_SERVO_DEACTIVATION_DELAY > 0)
        servos[servo_endstops[Z_AXIS]].attach(0);
#endif
        servos[servo_endstops[Z_AXIS]].write(servo_endstop_angles[Z_AXIS * 2]);
#if defined (ENABLE_AUTO_BED_LEVELING) && (PROBE_SERVO_DEACTIVATION_DELAY > 0)
        delay(PROBE_SERVO_DEACTIVATION_DELAY);
        servos[servo_endstops[Z_AXIS]].detach();
#endif
    }
    #endif
}

static void retract_z_probe() {
    // Retract Z Servo endstop if enabled
    #ifdef SERVO_ENDSTOPS
    if (servo_endstops[Z_AXIS] > -1) {
#if defined (ENABLE_AUTO_BED_LEVELING) && (PROBE_SERVO_DEACTIVATION_DELAY > 0)
        servos[servo_endstops[Z_AXIS]].attach(0);
#endif
        servos[servo_endstops[Z_AXIS]].write(servo_endstop_angles[Z_AXIS * 2 + 1]);
#if defined (ENABLE_AUTO_BED_LEVELING) && (PROBE_SERVO_DEACTIVATION_DELAY > 0)
        delay(PROBE_SERVO_DEACTIVATION_DELAY);
        servos[servo_endstops[Z_AXIS]].detach();
#endif
    }
    #endif
}

/// Probe bed height at position (x,y), returns the measured z value
static float probe_pt(float x, float y, float z_before) {
  // move to right place
  do_blocking_move_to(current_position[X_AXIS], current_position[Y_AXIS], z_before);
  do_blocking_move_to(x - X_PROBE_OFFSET_FROM_EXTRUDER, y - Y_PROBE_OFFSET_FROM_EXTRUDER, current_position[Z_AXIS]);

  engage_z_probe();   // Engage Z Servo endstop if available
  run_z_probe();
  float measured_z = current_position[Z_AXIS];
  retract_z_probe();

  SERIAL_PROTOCOLPGM(MSG_BED);
  SERIAL_PROTOCOLPGM(" x: ");
  SERIAL_PROTOCOL(x);
  SERIAL_PROTOCOLPGM(" y: ");
  SERIAL_PROTOCOL(y);
  SERIAL_PROTOCOLPGM(" z: ");
  SERIAL_PROTOCOL(measured_z);
  SERIAL_PROTOCOLPGM("\n");
  return measured_z;
}

#endif // #ifdef ENABLE_AUTO_BED_LEVELING



void refresh_cmd_timeout(void)
{
	my_printer.previous_millis_cmd = millis();
}

#ifdef FWRETRACT
  void retract(bool retracting) {
    if(retracting && !retracted) {
      destination[X_AXIS]=current_position[X_AXIS];
      destination[Y_AXIS]=current_position[Y_AXIS];
      destination[Z_AXIS]=current_position[Z_AXIS];
      destination[E_AXIS]=current_position[E_AXIS];
      current_position[E_AXIS]+=retract_length/volumetric_multiplier[active_extruder];
      plan_set_e_position(current_position[E_AXIS]);
      float oldFeedrate = feedrate;
      feedrate=retract_feedrate*60;
      retracted=true;
      prepare_move();
      current_position[Z_AXIS]-=retract_zlift;
      plan_set_position(current_position[X_AXIS], current_position[Y_AXIS], current_position[Z_AXIS], current_position[E_AXIS]);
      prepare_move();
      feedrate = oldFeedrate;
    } else if(!retracting && retracted) {
      destination[X_AXIS]=current_position[X_AXIS];
      destination[Y_AXIS]=current_position[Y_AXIS];
      destination[Z_AXIS]=current_position[Z_AXIS];
      destination[E_AXIS]=current_position[E_AXIS];
      current_position[Z_AXIS]+=retract_zlift;
      plan_set_position(current_position[X_AXIS], current_position[Y_AXIS], current_position[Z_AXIS], current_position[E_AXIS]);
      //prepare_move();
      current_position[E_AXIS]-=(retract_length+retract_recover_length)/volumetric_multiplier[active_extruder]; 
      plan_set_e_position(current_position[E_AXIS]);
      float oldFeedrate = feedrate;
      feedrate=retract_recover_feedrate*60;
      retracted=false;
      prepare_move();
      feedrate = oldFeedrate;
    }
  } //retract
#endif //FWRETRACT

  /**
   * M24: Start SD Print
   */
  inline void gcode_M24() {
    sdcard.startFileprint();
    my_printer.print_job_timer.start();
  }

  /**
   * M31: Get the time since the start of SD Print (or last M109)
   */
  inline void gcode_M31() {
    millis_t t = my_printer.print_job_timer.duration();
    int min = t / 60, sec = t % 60;
    char time[30];
    sprintf_P(time, PSTR("%i min, %i sec"), min, sec);
    my_printer.echo();
  //  SERIAL_ECHOLN(time);
 //   lcd_setstatus(time);
    heater.autotempShutdown();
  }

  /**
   * M75: Start print timer
   */
  inline void gcode_M75() {
	  my_printer.print_job_timer.start();
  }

  /**
   * M76: Pause print timer
   */
  inline void gcode_M76() {
	  my_printer.print_job_timer.pause();
  }

  /**
   * M77: Stop print timer
   */
  inline void gcode_M77() {
	  my_printer.print_job_timer.stop();
  }

  /**
   * M104: Set hot end temperature
   */
  inline void gcode_M104()
  {
	  if (setTargetedHotend(104)) return;

	  if (code_seen('S')) {
		  float temp = code_value();
		  heater.setTargetHeater(temp, target_extruder);
	  }
  }

  /**
   * M105: Read hot end and bed temperature
   */
  inline void gcode_M105() {
	  if (setTargetedHotend(105)) return;

#if defined(TEMP_0_PIN) && TEMP_0_PIN > -1
//	  my_printer.message(PSTR(MSG_OK));
//	  SERIAL_PROTOCOLPGM(MSG_OK);
//	  print_heaterstates();
	  heater.Print_Heater_States();
#else // !HAS_TEMP_HOTEND && !HAS_TEMP_BED
	  SERIAL_ERROR_START;
	  SERIAL_ERRORLNPGM(MSG_ERR_NO_THERMISTORS);
#endif
//     SERIAL_PROTOCOLLN("");
//	  SERIAL_EOL;
	  fromsd[cmd_queue_index_r]= true;
  }

  // goes through command buffer and
  // check if next command is M105
  // all other commands are rejected
//  bool Check_M105()	{
//	  bool M105_found = false;
//
//	  get_command();
//	  while (commands_in_queue) {
//		  if(code_seen('M'))
//			  if ((int)code_value() == 105)
//				  M105_found = true;
//		  // every other command is discarded
//		  commands_in_queue = (commands_in_queue - 1);
//		  cmd_queue_index_r = (cmd_queue_index_r + 1) % BUFSIZE;
//		  if (M105_found)
//			  return true;
//	  }
//
//
//	  return false;
//  }
/**
* M109: Sxxx Wait for extruder(s) to reach temperature. Waits only when heating.
*       Rxxx Wait for extruder(s) to reach temperature. Waits when heating and cooling.
*       without explicit commands: R (Target-Temperature): waiting to reach the given target temperature
*/
  inline void gcode_M109()
  {
	  if(setTargetedHotend(109))
		  return;

	  //    if (DEBUGGING(DRYRUN)) return;

	  bool no_wait_for_cooling = code_seen('S');
	  if (no_wait_for_cooling || code_seen('R'))
	  {
		  float temp = code_value();
		  heater.setTargetHeater(temp, target_extruder);

		  /**
		   * We use half EXTRUDE_MINTEMP here to allow nozzles to be put into hot
		   * stand by mode, for instance in a dual extruder setup, without affecting
		   * the running print timer.
		   */
		  if (temp <= (EXTRUDE_MINTEMP)/2) {
			  my_printer.print_job_timer.stop();
			  lcd_setstatuspgm(T(MSG_Welcome));
		  }
		  /**
		   * We do not check if the timer is already running because this check will
		   * be done for us inside the Stopwatch::start() method thus a running timer
		   * will not restart.
		   */
		  else my_printer.print_job_timer.start();

		  if (temp > heater.current_temperature[target_extruder]) lcd_setstatuspgm(T(MSG_Heating));
	  }
	  else
	  {
		  my_printer.print_job_timer.start();

		  if (heater.target_temperature[target_extruder] > heater.current_temperature[target_extruder]) lcd_setstatuspgm(T(MSG_Heating));
	  }

	  //    #if ENABLED(AUTOTEMP)
	  //      autotemp_enabled = code_seen('F');
	  //      if (autotemp_enabled) autotemp_factor = code_value();
	  //      if (code_seen('S')) autotemp_min = code_value();
	  //      if (code_seen('B')) autotemp_max = code_value();
	  //    #endif

	  bool wants_to_cool = heater.isCoolingHotend(target_extruder);

	  // Exit if S<lower>, continue if S<higher>, R<lower>, or R<higher>
	  if (no_wait_for_cooling && wants_to_cool) return;
	  //
	  // Prevents a wait-forever situation if R is misused i.e. M109 R0
	  // Try to calculate a ballpark safe margin by halving EXTRUDE_MINTEMP
	  if (wants_to_cool && heater.degTargetHotend(target_extruder) < (EXTRUDE_MINTEMP)/2)
		  return;

	  heater.Wait_T_Stable(target_extruder, true);
	  lcd_setstatuspgm(T(MSG_Heating_complete));
  }

/**
 * M214: M214 advanced settings:
 *
 * T=tool number
 * K=magic number for selected extruder
 * D=delay between pressure initialise steps for selected extruder
 */
  inline void gcode_M214()
  {

	  int Tool_nb = active_extruder;
	  bool tool_selected = false;

	  if (code_seen('T'))
	  {
		  Tool_nb = code_value();
		  tool_selected = true;
	  }
	  if (Tool_nb >= EXTRUDERS) {
		  //		  my_printer.echo();
//		  my_printer.echo_ln("Illegal Tool Number");
		  my_printer.f_message(T(MSG_Invalid_Extruder_Num),Tool_nb);
		  return;
	  }
	  if (code_seen('K')) {
		  config.advance_factor[Tool_nb] = code_value();
		  tool_selected = true;
	  } else {
		  config.advance_factor[Tool_nb] = 0;
		  config.advance_delay[Tool_nb] = 0;
		  tool_selected = true;
	  }
	  if (code_seen('D')) {
		  config.advance_delay[Tool_nb] = code_value();
		  tool_selected = true;
	  }
	  for (int i = 0; i < EXTRUDERS; i++)
	  {
		  my_printer.echo();
		  if (!tool_selected || i == Tool_nb)
		  {
			  SERIAL_ECHO(" e:"); // specify extruder in serial output
			  SERIAL_ECHO(i);
			  SERIAL_ECHO(" K:");
			  SERIAL_ECHO(config.advance_factor[i]);
			  SERIAL_ECHO(" delay:");
			  SERIAL_ECHOLN(config.advance_delay[i]);
		  }
	  }
  }
  /**
   * M218: M218 set extruder offset:
   *
   * set hotend offset (in mm), T<extruder_number> X<offset_on_X> Y<offset_on_Y>
   */
    inline void gcode_M218() {

        if(setTargetedHotend(218)){
         return;
        }
#if EXTRUDERS > 1
        if(code_seen('X'))
        {
          config.extruder_offset[X_AXIS][target_extruder] = code_value();
        }
        if(code_seen('Y'))
        {
        	config.extruder_offset[Y_AXIS][target_extruder] = code_value();
        }
        my_printer.echo_ln(T(MSG_Hotend_Offset));
        for(target_extruder = 0; target_extruder < EXTRUDERS; target_extruder++)
        {
        	my_printer.f_message(PSTR(" T%i:  X %f  Y %f\n"),target_extruder,config.extruder_offset[X_AXIS][target_extruder],config.extruder_offset[Y_AXIS][target_extruder]);
        }
#endif
    }
#ifdef PIDTEMP

  /**
   * M301: Set PID parameters P I D (and optionally C, L)
   *
   *   P[float] Kp term
   *   I[float] Ki term (unscaled)
   *   D[float] Kd term (unscaled)
   *
   * With PID_ADD_EXTRUSION_RATE:
   *
   *   C[float] Kc term
   *   L[float] LPQ length
   *   F[float] Kf term
   */
  inline void gcode_M301() {

    // multi-extruder PID patch: M301 updates or prints a single extruder's PID values
    // default behaviour (omitting E parameter) is to update for extruder 0 only
	  // implementation of PID barameters for Bed: code "B"
    int e = code_seen('E') ? code_value() : 0; // extruder being updated
    if (code_seen('B')) e = HEAT_BED;

    if (e < EXTRUDERS) { // catch bad input value
      if (code_seen('P')) config.Set_Kp(code_value(),e);
      if (code_seen('I')) config.Set_Ki(code_value(),e);
      if (code_seen('D')) config.Set_Kd(code_value(),e);
      #ifdef PID_ADD_EXTRUSION_RATE
        if (code_seen('C')) config.Set_Kc(code_value(),e);
        if (code_seen('F')) config.Set_Kf(code_value(),e);
        if (code_seen('L')) lpq_len = code_value();
        NOMORE(lpq_len, LPQ_MAX_LEN);
      #endif

    heater.UpdatePID();
      my_printer.echo();
      #ifdef PID_PARAMS_PER_EXTRUDER
      my_printer.f_message(T(MSG_Extruder_Nb),e);
//        SERIAL_ECHO(" e:"); // specify extruder in serial output
//        SERIAL_ECHO(e);
      #endif // PID_PARAMS_PER_EXTRUDER
      my_printer.f_message(T(MSG_PID_Parameter),config.Get_Kp(e),config.Get_Ki(e),config.Get_Kd(e));
//
//      SERIAL_ECHO(" p:");
//      SERIAL_ECHO(config.Get_Kp(e));
//      SERIAL_ECHO(" i:");
//      SERIAL_ECHO(config.Get_Ki(e));
//      SERIAL_ECHO(" d:");
//      SERIAL_ECHO(config.Get_Kd( e));
      #ifdef PID_ADD_EXTRUSION_RATE
      my_printer.f_message(T(MSG_CF_Parameter),config.Get_Kc(e),config.Get_Kf(e));
//
//        SERIAL_ECHO(" c:");
//        //Kc does not have scaling applied above, or in resetting defaults
//        SERIAL_ECHO(config.Get_Kc(e));
//        SERIAL_ECHO(" F:");
//        SERIAL_ECHO(config.Get_Kf(e));
        SERIAL_ECHO(" l:");
        SERIAL_ECHO(lpq_len);
      #endif
//        SERIAL_PROTOCOLLN("");
//      SERIAL_EOL;
    }
    else {
    	my_printer.f_message  (T(MSG_Invalid_Extruder),301,e );
//      SERIAL_ERRORLN(MSG_INVALID_EXTRUDER);
    }
  }


  inline void gcode_M303()
  {
	  float temp = 150.0;
	  int e=0;
	  int Last_Extruder = 0;
	  int c=7;
	  bool update = false;
	  bool tune = false;

	  if (code_seen('E')) e=code_value();
	  if (e<0)
		  temp=70;
	  if (code_seen('S')) temp=code_value();
	  if (code_seen('C')) c=code_value();
	  if (code_seen('U')) update = true;
	  if (code_seen('T'))	tune = true;

	  heater.PID_autotune(temp, e, c, update);

	  if ((!Position.axis_known_position[X_AXIS] || !Position.axis_known_position[Y_AXIS] || !Position.axis_known_position[Z_AXIS]) && tune)
	  {
		  Position.home(true, true, true);
//		  my_printer.message(T(MSG_Temp_Tune_Failed));
	  }

	  if (tune)
	  {
		  my_printer.message(T(MSG_Temp_Tune_Started));
		  // set target temperature
		  Last_Extruder = active_extruder;
		  active_extruder = e;
		  heater.setTargetHeater(temp, e);
		  my_printer.message(T(MSG_Temp_Tune_Fan));
		  // go to center of the plate
		  // with short Z-distance
		  Position.destination[X_AXIS] = 100;
		  Position.destination[Y_AXIS] = 100;
		  Position.destination[Z_AXIS] = 1;
		  Position.destination[E_AXIS] = Position.current_position_mm[E_AXIS];
		  Position.Prepare_Move();
		  heater.Tune_Extruder_Fan(e, update);
		  // now get higher Z-Distand for extrusion
		  Position.destination[X_AXIS] = 100;
		  Position.destination[Y_AXIS] = 100;
		  Position.destination[Z_AXIS] = 100;
		  Position.destination[E_AXIS] = Position.current_position_mm[E_AXIS];
		  Position.Prepare_Move();
		  config.Set_Kc(0,e);
//		  PID_PARAM(Kc,e) = 0;
		  // initial filament feed to fill the nozzle
		  //
		  Position.destination[X_AXIS] = Position.current_position_mm[X_AXIS];
		  Position.destination[Y_AXIS] = Position.current_position_mm[Y_AXIS];
		  Position.destination[Z_AXIS] = Position.current_position_mm[Z_AXIS];
		  Position.destination[E_AXIS] = Position.current_position_mm[E_AXIS] + 20.0;
		  Position.Prepare_Move();
		  my_printer.message(T(MSG_Temp_Tune_Extrusion));
		  //start with reference heat
		  my_printer.echo();
		  temp = heater.Wait_T_Stable (e,true);
		  temp = heater.Wait_T_Stable (e,true);
		  my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp);
	   		unsigned long ref_heat = heater.Heat ( e, 25000);
			  temp = heater.Wait_T_Stable (e,true);
			  my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp);
		  // then 10mm extrusion
			  Position.destination[X_AXIS] = Position.current_position_mm[X_AXIS];
			  Position.destination[Y_AXIS] = Position.current_position_mm[Y_AXIS];
			  Position.destination[Z_AXIS] = Position.current_position_mm[Z_AXIS];
			  Position.destination[E_AXIS] = Position.current_position_mm[E_AXIS] + 10.0;
			  Position.Prepare_Move();
		   		unsigned long heat = heater.Heat ( e, 25000);
		  		unsigned long delta_10 = heat - ref_heat;
		  	unsigned int kc = delta_10/10;
		  	my_printer.f_message(T(MSG_Temp_Heat_Result),10,kc);
//		  	SERIAL_PROTOCOLPGM(" Kc 10mm "); SERIAL_PROTOCOLLN(kc);

		  // REPEAT with 5mm extrusion

			  temp = heater.Wait_T_Stable (e,true);
			  my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp);
		  //
			  Position.destination[X_AXIS] = Position.current_position_mm[X_AXIS];
			  Position.destination[Y_AXIS] = Position.current_position_mm[Y_AXIS];
			  Position.destination[Z_AXIS] = Position.current_position_mm[Z_AXIS];
			  Position.destination[E_AXIS] = Position.current_position_mm[E_AXIS] + 5.0;
			  Position.Prepare_Move();
		    		heat = heater.Heat ( e, 25000);
		    		unsigned long delta_5 = ((heat - ref_heat));
		    		kc = delta_5/5;
		  //  		SERIAL_PROTOCOLPGM(" Kc 5mm "); SERIAL_PROTOCOLLN(kc);
				  	my_printer.f_message(T(MSG_Temp_Heat_Result),5,kc);


		  //repeat with 15mm extrusion
					  temp = heater.Wait_T_Stable (e,true);
					  my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp);
					  //
					  Position.destination[X_AXIS] = Position.current_position_mm[X_AXIS];
					  Position.destination[Y_AXIS] = Position.current_position_mm[Y_AXIS];
					  Position.destination[Z_AXIS] = Position.current_position_mm[Z_AXIS];
					  Position.destination[E_AXIS] = Position.current_position_mm[E_AXIS] + 15.0;
					  Position.Prepare_Move();
		    		heat = heater.Heat ( e, 25000);
		    		unsigned long delta_15 = heat - ref_heat;
		    		kc = delta_15/15;
		  //  		SERIAL_PROTOCOLPGM(" Kc 15mm "); SERIAL_PROTOCOLLN(kc);
				  	my_printer.f_message(T(MSG_Temp_Heat_Result),15,kc);

		  // now take the mean value unknown

		    		kc += delta_10/10;
		    		kc += delta_5/5;
		    		kc /= 3;
		  //  		SERIAL_PROTOCOLPGM(" Kc = "); SERIAL_PROTOCOLLN(kc/128);
		    		config.Set_Kc(kc,e);
//		    		PID_PARAM(Kc,e) = kc*PID_dT;
my_printer.f_message(T(MSG_Temp_Heat_KC), (int)config.Get_Kc(e));

		  active_extruder = Last_Extruder;
	  }
  }

#endif // PIDTEMP





void process_commands()
{

  char *starpos = NULL;
#ifdef ENABLE_AUTO_BED_LEVELING
  float x_tmp, y_tmp, z_tmp, real_z;
#endif
  if(code_seen('G'))
  {
    switch((int)code_value())
    {
    case 0: // G0 -> G1
    case 1: // G1
      if(my_printer.Stopped == false) {
        get_coordinates(); // For X Y Z E F
          #ifdef FWRETRACT
            if(autoretract_enabled)
            if( !(code_seen('X') || code_seen('Y') || code_seen('Z')) && code_seen('E')) {
              float echange=destination[E_AXIS]-current_position[E_AXIS];
              if((echange<-MIN_RETRACT && !retracted) || (echange>MIN_RETRACT && retracted)) { //move appears to be an attempt to retract or recover
                  current_position[E_AXIS] = destination[E_AXIS]; //hide the slicer-generated retract/recover from calculations
                  plan_set_e_position(current_position[E_AXIS]); //AND from the planner
                  retract(!retracted);
                  return;
              }
            }
          #endif //FWRETRACT
            Position.Prepare_Move();
        ClearToSend();
        return;
      }
      break;
    case 2: // G2  - CW ARC
      if(my_printer.Stopped == false) {
        get_arc_coordinates();
        prepare_arc_move(true);
        return;
      }
      break;
    case 3: // G3  - CCW ARC
      if(my_printer.Stopped == false) {
        get_arc_coordinates();
        prepare_arc_move(false);
        return;
      }
      break;
    case 4: // G4 dwell
   //   lcd_setstatuspgm(MSG_DWELL);
      codenum = 0;
      if(code_seen('P')) codenum = code_value(); // milliseconds to wait
      if(code_seen('S')) codenum = code_value() * 1000; // seconds to wait

      motor.Synchronize();
      codenum += millis();  // keep track of when we started waiting
      my_printer.previous_millis_cmd = millis();
      while(millis()  < codenum )
      {
    	  my_printer.Go_Idle();
      }
      break;
      #ifdef FWRETRACT
      case 10: // G10 retract
        retract(true);
      break;
      case 11: // G11 retract_recover
        retract(false);
      break;
      #endif //FWRETRACT
    case 28: //G28 Home all Axis one at a time
#ifdef ENABLE_AUTO_BED_LEVELING
      plan_bed_level_matrix.set_to_identity();  //Reset the plane ("erase" all leveling data)
#endif //ENABLE_AUTO_BED_LEVELING

      homex = homey = homez = false;

      if (code_seen(axis_codes[X_AXIS]))
		{
		homex = true;
		};

	     if (code_seen(axis_codes[Y_AXIS]))
			{
			homey = true;
			}
		     if (code_seen(axis_codes[Z_AXIS]))
				{
				homez = true;
				}
	if (!homex && !homey && !homez)
	{
		// home all axes
		homex = homey = homez = true;
	}

    Position.home(homex, homey, homez);
      break;

#ifdef ENABLE_AUTO_BED_LEVELING
    case 29: // G29 Detailed Z-Probe, probes the bed at 3 or more points.
        {
            #if Z_MIN_PIN == -1
            #error "You must have a Z_MIN endstop in order to enable Auto Bed Leveling feature!!! Z_MIN_PIN must point to a valid hardware pin."
            #endif

            // Prevent user from running a G29 without first homing in X and Y
            if (! (axis_known_position[X_AXIS] && axis_known_position[Y_AXIS]) )
            {
                LCD_MESSAGEPGM(MSG_POSITION_UNKNOWN);
                SERIAL_ECHO_START;
                SERIAL_ECHOLNPGM(MSG_POSITION_UNKNOWN);
                break; // abort G29, since we don't know where we are
            }

            st_synchronize();
            // make sure the bed_level_rotation_matrix is identity or the planner will get it incorectly
            //vector_3 corrected_position = plan_get_position_mm();
            //corrected_position.debug("position before G29");
            plan_bed_level_matrix.set_to_identity();
            vector_3 uncorrected_position = plan_get_position();
            //uncorrected_position.debug("position durring G29");
            current_position[X_AXIS] = uncorrected_position.x;
            current_position[Y_AXIS] = uncorrected_position.y;
            current_position[Z_AXIS] = uncorrected_position.z;
            plan_set_position(current_position[X_AXIS], current_position[Y_AXIS], current_position[Z_AXIS], current_position[E_AXIS]);
            setup_for_endstop_move();

            feedrate = homing_feedrate[Z_AXIS];
#ifdef AUTO_BED_LEVELING_GRID
            // probe at the points of a lattice grid

            int xGridSpacing = (RIGHT_PROBE_BED_POSITION - LEFT_PROBE_BED_POSITION) / (AUTO_BED_LEVELING_GRID_POINTS-1);
            int yGridSpacing = (BACK_PROBE_BED_POSITION - FRONT_PROBE_BED_POSITION) / (AUTO_BED_LEVELING_GRID_POINTS-1);


            // solve the plane equation ax + by + d = z
            // A is the matrix with rows [x y 1] for all the probed points
            // B is the vector of the Z positions
            // the normal vector to the plane is formed by the coefficients of the plane equation in the standard form, which is Vx*x+Vy*y+Vz*z+d = 0
            // so Vx = -a Vy = -b Vz = 1 (we want the vector facing towards positive Z

            // "A" matrix of the linear system of equations
            double eqnAMatrix[AUTO_BED_LEVELING_GRID_POINTS*AUTO_BED_LEVELING_GRID_POINTS*3];
            // "B" vector of Z points
            double eqnBVector[AUTO_BED_LEVELING_GRID_POINTS*AUTO_BED_LEVELING_GRID_POINTS];


            int probePointCounter = 0;
            bool zig = true;

            for (int yProbe=FRONT_PROBE_BED_POSITION; yProbe <= BACK_PROBE_BED_POSITION; yProbe += yGridSpacing)
            {
              int xProbe, xInc;
              if (zig)
              {
                xProbe = LEFT_PROBE_BED_POSITION;
                //xEnd = RIGHT_PROBE_BED_POSITION;
                xInc = xGridSpacing;
                zig = false;
              } else // zag
              {
                xProbe = RIGHT_PROBE_BED_POSITION;
                //xEnd = LEFT_PROBE_BED_POSITION;
                xInc = -xGridSpacing;
                zig = true;
              }

              for (int xCount=0; xCount < AUTO_BED_LEVELING_GRID_POINTS; xCount++)
              {
                float z_before;
                if (probePointCounter == 0)
                {
                  // raise before probing
                  z_before = Z_RAISE_BEFORE_PROBING;
                } else
                {
                  // raise extruder
                  z_before = current_position[Z_AXIS] + Z_RAISE_BETWEEN_PROBINGS;
                }

                float measured_z = probe_pt(xProbe, yProbe, z_before);

                eqnBVector[probePointCounter] = measured_z;

                eqnAMatrix[probePointCounter + 0*AUTO_BED_LEVELING_GRID_POINTS*AUTO_BED_LEVELING_GRID_POINTS] = xProbe;
                eqnAMatrix[probePointCounter + 1*AUTO_BED_LEVELING_GRID_POINTS*AUTO_BED_LEVELING_GRID_POINTS] = yProbe;
                eqnAMatrix[probePointCounter + 2*AUTO_BED_LEVELING_GRID_POINTS*AUTO_BED_LEVELING_GRID_POINTS] = 1;
                probePointCounter++;
                xProbe += xInc;
              }
            }
            clean_up_after_endstop_move();

            // solve lsq problem
            double *plane_equation_coefficients = qr_solve(AUTO_BED_LEVELING_GRID_POINTS*AUTO_BED_LEVELING_GRID_POINTS, 3, eqnAMatrix, eqnBVector);

            SERIAL_PROTOCOLPGM("Eqn coefficients: a: ");
            SERIAL_PROTOCOL(plane_equation_coefficients[0]);
            SERIAL_PROTOCOLPGM(" b: ");
            SERIAL_PROTOCOL(plane_equation_coefficients[1]);
            SERIAL_PROTOCOLPGM(" d: ");
            SERIAL_PROTOCOLLN(plane_equation_coefficients[2]);


            set_bed_level_equation_lsq(plane_equation_coefficients);

            free(plane_equation_coefficients);

#else // AUTO_BED_LEVELING_GRID not defined

            // Probe at 3 arbitrary points
            // probe 1
            float z_at_pt_1 = probe_pt(ABL_PROBE_PT_1_X, ABL_PROBE_PT_1_Y, Z_RAISE_BEFORE_PROBING);

            // probe 2
            float z_at_pt_2 = probe_pt(ABL_PROBE_PT_2_X, ABL_PROBE_PT_2_Y, current_position[Z_AXIS] + Z_RAISE_BETWEEN_PROBINGS);

            // probe 3
            float z_at_pt_3 = probe_pt(ABL_PROBE_PT_3_X, ABL_PROBE_PT_3_Y, current_position[Z_AXIS] + Z_RAISE_BETWEEN_PROBINGS);

            clean_up_after_endstop_move();

            set_bed_level_equation_3pts(z_at_pt_1, z_at_pt_2, z_at_pt_3);


#endif // AUTO_BED_LEVELING_GRID
            st_synchronize();

            // The following code correct the Z height difference from z-probe position and hotend tip position.
            // The Z height on homing is measured by Z-Probe, but the probe is quite far from the hotend.
            // When the bed is uneven, this height must be corrected.
            real_z = float(st_get_position(Z_AXIS))/axis_steps_per_unit[Z_AXIS];  //get the real Z (since the auto bed leveling is already correcting the plane)
            x_tmp = current_position[X_AXIS] + X_PROBE_OFFSET_FROM_EXTRUDER;
            y_tmp = current_position[Y_AXIS] + Y_PROBE_OFFSET_FROM_EXTRUDER;
            z_tmp = current_position[Z_AXIS];

            apply_rotation_xyz(plan_bed_level_matrix, x_tmp, y_tmp, z_tmp);         //Apply the correction sending the probe offset
            current_position[Z_AXIS] = z_tmp - real_z + current_position[Z_AXIS];   //The difference is added to current position and sent to planner.
            plan_set_position(current_position[X_AXIS], current_position[Y_AXIS], current_position[Z_AXIS], current_position[E_AXIS]);
        }
        break;

    case 30: // G30 Single Z Probe
        {
            engage_z_probe(); // Engage Z Servo endstop if available

            st_synchronize();
            // TODO: make sure the bed_level_rotation_matrix is identity or the planner will get set incorectly
            setup_for_endstop_move();

            feedrate = homing_feedrate[Z_AXIS];

            run_z_probe();
            SERIAL_PROTOCOLPGM(MSG_BED);
            SERIAL_PROTOCOLPGM(" X: ");
            SERIAL_PROTOCOL(current_position[X_AXIS]);
            SERIAL_PROTOCOLPGM(" Y: ");
            SERIAL_PROTOCOL(current_position[Y_AXIS]);
            SERIAL_PROTOCOLPGM(" Z: ");
            SERIAL_PROTOCOL(current_position[Z_AXIS]);
            SERIAL_PROTOCOLPGM("\n");

            clean_up_after_endstop_move();

            retract_z_probe(); // Retract Z Servo endstop if available
        }
        break;
#endif // ENABLE_AUTO_BED_LEVELING
    case 90: // G90// G90 - Use Absolute Coordinates
      relative_mode = false;
      break;
    case 91: // G91// G91 - Use Relative Coordinates
      relative_mode = true;
      break;
    case 92: // G92// G92 - Set current position to coordinates given
//      if(!code_seen(axis_codes[E_AXIS]))
    	motor.Synchronize();
      for(AxisEnum a=X_AXIS; a <= E_AXIS; a = static_cast<AxisEnum>(a+1))
      {
        if(code_seen(axis_codes[a]))
        {
           if(a == E_AXIS) {
        	   Position.Set_Position_mm(E_AXIS, code_value());

           }
           else
           {
        	   Position.Set_Position_mm(a, code_value()+config.add_homeing[a]);
           }
        }
      }
      break;
    }
  }

  else if(code_seen('M'))
  {
    switch( (int)code_value() )
    {
#ifdef ULTIPANEL
    case 0: // M0 - Unconditional stop - Wait for user button press on LCD
    case 1: // M1 - Conditional stop - Wait for user button press on LCD
    {
    //  lcd_setstatuspgm(MSG_USERWAIT);
    my_printer.Status = Waiting;
      codenum = 0;
      if(code_seen('P')) codenum = code_value(); // milliseconds to wait
      if(code_seen('S')) codenum = code_value() * 1000; // seconds to wait

      motor.Synchronize();
      my_printer.previous_millis_cmd = millis();
      if (codenum > 0){
        codenum += millis();  // keep track of when we started waiting
        while(millis()  < codenum && !lcd_clicked())
        {
        	my_printer.Go_Idle();
        }
      }
      else
      {
        while(!lcd_clicked())
        {
        	my_printer.Go_Idle();
        }
      }
  //    lcd_setstatuspgm(MSG_RESUMING);
    }
    break;
#endif
    case 16:
    	service.Print(Service_Disable_Steppers);
    	break;
    case 17:
 //       lcd_setstatuspgm(MSG_NO_MOVE);
        enable_X();
        enable_Y();
        enable_Z();
        enable_E0();
        enable_E1();
        enable_E2();
      break;

#ifdef SDSUPPORT
    case 20: // M20 - list SD card
    	my_printer.message(T(MSG_Begin_File_List));
      sdcard.ls();
    	my_printer.message(T(MSG_End_File_List));
      break;
    case 21: // M21 - init SD card
      sdcard.initsd();
      break;
    case 22: //M22 - release SD card
      sdcard.release();
      break;
    case 23: //M23 - Select file
      starpos = (strchr(strchr_pointer + 4,'*'));
      if(starpos!=NULL)
        *(starpos)='\0';
      sdcard.openFile(strchr_pointer + 4,true);
      break;
    case 24: //M24 - Start SD print
      sdcard.startFileprint();
      my_printer.print_job_timer.reset();
      my_printer.print_job_timer.start();
      break;
    case 25: //M25 - Pause SD print
      sdcard.pauseSDPrint();
      break;
    case 26: //M26 - Set SD index
      if(sdcard.Status == Card_OK && code_seen('S')) {
        sdcard.setIndex(code_value_long());
      }
      break;
    case 27: //M27 - Get SD status
      sdcard.getStatus();
      break;
    case 28: //M28 - Start SD write
      starpos = (strchr(strchr_pointer + 4,'*'));
      if(starpos != NULL){
        char* npos = strchr(cmdbuffer[cmd_queue_index_r], 'N');
        strchr_pointer = strchr(npos,' ') + 1;
        *(starpos) = '\0';
      }
      sdcard.openFile(strchr_pointer+4,false);
      break;
    case 29: //M29 - Stop SD write
      //processed in write to file routine above
      sdcard.saving = false;
      break;
    case 30: //M30 <filename> Delete File
      if (sdcard.Status  == Card_OK){
        sdcard.closefile();
        starpos = (strchr(strchr_pointer + 4,'*'));
        if(starpos != NULL){
          char* npos = strchr(cmdbuffer[cmd_queue_index_r], 'N');
          strchr_pointer = strchr(npos,' ') + 1;
          *(starpos) = '\0';
        }
        sdcard.removeFile(strchr_pointer + 4);
      }
      break;
    case 32: //M32 - Select file and start SD print
    {
      if(sdcard.Status  == Card_OK) {
    	  motor.Synchronize();
      }
      my_printer.f_message(PSTR("M32 call: %s/n"),strchr_pointer);
      starpos = (strchr(strchr_pointer + 4,'*'));

      char* namestartpos = (strchr(strchr_pointer + 4,'!'));   //find ! to indicate filename string start.
      if(namestartpos==NULL)
      {
        namestartpos=strchr_pointer + 4; //default name position, 4 letters after the M
      }
      else
        namestartpos++; //to skip the '!'

      if(starpos!=NULL)
        *(starpos)='\0';

//      bool call_procedure=(code_seen('P'));
//
//      if(strchr_pointer>namestartpos)
//        call_procedure=false;  //false alert, 'P' found within filename

      if( sdcard.Status == Card_printing )
      {
    	  my_printer.f_message(PSTR("M32: %s %i"),namestartpos, 0);
        sdcard.openFile(namestartpos,true);
//        if(code_seen('S'))
//          if(strchr_pointer<namestartpos) //only if "S" is occuring _before_ the filename
//            sdcard.setIndex(code_value_long());
        sdcard.startFileprint();
        my_printer.print_job_timer.start();
//        if(!call_procedure) my_printer.print_job_timer.start();
//          starttime=millis(); //procedure calls count as normal print time.
      }
    } break;
    case 928: //M928 - Start SD write
      starpos = (strchr(strchr_pointer + 5,'*'));
      if(starpos != NULL){
        char* npos = strchr(cmdbuffer[cmd_queue_index_r], 'N');
        strchr_pointer = strchr(npos,' ') + 1;
        *(starpos-1) = '\0';
      }
      sdcard.openLogFile(strchr_pointer+5);
      break;

#endif //SDSUPPORT

    case 31: //M31 take time since the start of the SD print or an M109 command
    	gcode_M31();
//      {
//      stoptime=millis();
//      char time[30];
//      unsigned long t=(stoptime-starttime)/1000;
//      int sec,min;
//      min=t/60;
//      sec=t%60;
//      sprintf_P(time, PSTR("%i min, %i sec"), min, sec);
//      SERIAL_ECHO_START;
//      SERIAL_ECHOLN(time);
//      lcd_setstatus(time);
//      autotempShutdown();
//      }
      break;
    case 42: //M42 -Change pin status via gcode
      if (code_seen('S'))
      {
        int pin_status = code_value();
        int pin_number = LED_PIN;
        if (code_seen('P') && pin_status >= 0 && pin_status <= 255)
          pin_number = code_value();
        for(int8_t i = 0; i < (int8_t)sizeof(sensitive_pins); i++)
        {
          if (sensitive_pins[i] == pin_number)
          {
            pin_number = -1;
            break;
          }
        }
        if (pin_number > -1)
        {
          pinMode(pin_number, OUTPUT);
          digitalWrite(pin_number, pin_status);
          analogWrite(pin_number, pin_status);
        }
      }
     break;
    case 75: // Start print timer
      gcode_M75();
      break;

    case 76: // Pause print timer
      gcode_M76();
      break;

    case 77: // Stop print timer
      gcode_M77();
      break;
    case 104: // M104
    	gcode_M104();
      break;
    case 140: // M140 set bed temp
      if (code_seen('S')) heater.setTargetHeater(code_value(),HEAT_BED);
      break;
    case 105 : // M105
    	gcode_M105();
      break;
    case 109:
    {// M109 - Wait for extruder heater to reach target.
    	gcode_M109();
      }
      break;
    case 190: // M190 - Wait for bed heater to reach target.
    #if defined(TEMP_BED_PIN) && TEMP_BED_PIN > -1
 //       lcd_setstatuspgm(MSG_BED_HEATING);
        if (code_seen('S')) {
          heater.setTargetHeater(code_value(),HEAT_BED);
          CooldownNoWait = true;
        } else if (code_seen('R')) {
          heater.setTargetHeater(code_value(),HEAT_BED);
          CooldownNoWait = false;
        }
        codenum = millis();

        target_direction = heater.isHeatingBed(); // true if heating, false if cooling

        while ( target_direction ? (heater.isHeatingBed()) : (heater.isCoolingBed()&&(CooldownNoWait==false)) )
        {
          if(( millis() - codenum) > 1000 ) //Print Temp Reading every 1 second while heating up.
          {
        	  heater.Print_Heater_States();
            codenum = millis();
          }
          my_printer.Go_Idle();
        }
        lcd_setstatuspgm(T(MSG_Bed_Done));
        my_printer.previous_millis_cmd = millis();
    #endif
        break;

    #if defined(FAN_PIN) && FAN_PIN > -1
      case 106: //M106 Fan On
        if (code_seen('S'))
        {
        	heater.Set_FanSpeed(code_value());
        }
        else
        {
          heater.Set_FanSpeed(255);
        }
        break;
      case 107: //M107 Fan Off
        heater.Set_FanSpeed(0);
        break;
    #endif //FAN_PIN

      case 81: // M81 - Turn off Power Supply
        heater.Disable();
        motor.Synchronize();
        disable_E0();
        disable_E1();
        disable_E2();
        motor.FinishAndDisable();
        heater.Set_FanSpeed(0);
        delay(1000); // Wait a little before to switch off
      #ifdef ULTIPANEL
//        powersupply = false;
    //    LCD_MESSAGEPGM(MACHINE_NAME" "MSG_OFF".");
    lcd_update();
      #endif
	  break;

    case 82:
      axis_relative_modes[3] = false;
      break;
    case 83:
      axis_relative_modes[3] = true;
      break;
    case 18: //compatibility
    case 84: // M84
      if(code_seen('S')){
    	  my_printer.Reset_MotorTimeout(code_value() * 1000);
      }
      else
      {
        bool all_axis = !((code_seen(axis_codes[X_AXIS])) || (code_seen(axis_codes[Y_AXIS])) || (code_seen(axis_codes[Z_AXIS]))|| (code_seen(axis_codes[E_AXIS])));
        if(all_axis)
        {
          //finishAndDisableSteppers;
          motor.FinishAndDisable();
        }
        else
        {
        	motor.Synchronize();
          if(code_seen('X')) disable_X();
          if(code_seen('Y')) disable_Y();
          if(code_seen('Z')) disable_Z();
          #if ((E0_ENABLE_PIN != X_ENABLE_PIN) && (E1_ENABLE_PIN != Y_ENABLE_PIN)) // Only enable on boards that have seperate ENABLE_PINS
            if(code_seen('E')) {
              disable_E0();
              disable_E1();
              disable_E2();
            }
          #endif
        }
      }
      break;
    case 85: // M85
      code_seen('S');
      my_printer.Reset_Activity_Timeout(code_value() * 1000);
      break;
    case 92: // M92
      for(int8_t i=0; i < NUM_AXIS; i++)
      {
        if(code_seen(axis_codes[i]))
        {
          if(i == 3) { // E
            float value = code_value();
            if(value < 20.0) {
              float factor = config.axis_steps_per_unit[i] / value; // increase e constants if M92 E14 is given for netfab.
              config.max_jerk[E_AXIS] *= factor;
              config.max_feedrate_mm_s[i] *= factor;
              config.axis_steps_per_sqr_second[i] *= factor;
            }
            config.axis_steps_per_unit[i] = value;
          }
          else {
            config.axis_steps_per_unit[i] = code_value();
          }
        }
      }
      break;
    case 115: // M115
      my_printer.message(PSTR(VERSION_M115_REPORT));
      break;
    case 117: // M117 display message
      starpos = (strchr(strchr_pointer + 5,'*'));
      if(starpos!=NULL)
        *(starpos-1)='\0';
      lcd_setstatus(strchr_pointer + 5);
      break;
    case 114: // M114
      SERIAL_PROTOCOLPGM("X:");
      SERIAL_PROTOCOL(Position.current_position_mm[X_AXIS]);
      SERIAL_PROTOCOLPGM(" Y:");
      SERIAL_PROTOCOL(Position.current_position_mm[Y_AXIS]);
      SERIAL_PROTOCOLPGM(" Z:");
      SERIAL_PROTOCOL(Position.current_position_mm[Z_AXIS]);
      SERIAL_PROTOCOLPGM(" E:");
      SERIAL_PROTOCOL(Position.current_position_mm[E_AXIS]);

      SERIAL_PROTOCOLPGM("Count X:");
      SERIAL_PROTOCOL(float(motor.Get_Position(X_AXIS))/config.axis_steps_per_unit[X_AXIS]);
      SERIAL_PROTOCOLPGM(" Y:");
      SERIAL_PROTOCOL(float(motor.Get_Position(Y_AXIS))/config.axis_steps_per_unit[Y_AXIS]);
      SERIAL_PROTOCOLPGM(" Z:");
      SERIAL_PROTOCOL(float(motor.Get_Position(Z_AXIS))/config.axis_steps_per_unit[Z_AXIS]);

      SERIAL_PROTOCOLLN("");
      break;
    case 120: // M120
      motor.Enable_Endstops(false) ;
      break;
    case 121: // M121
    	motor.Enable_Endstops(true) ;
      break;
    case 119: // M119
        my_printer.message(T(MSG_M119_Report));
      #if defined(X_MIN_PIN) && X_MIN_PIN > -1
        SERIAL_PROTOCOLPGM(MSG_X_MIN);
        SERIAL_PROTOCOLLN(((READ(X_MIN_PIN)^X_MIN_ENDSTOP_INVERTING)?T(MSG_Endstop_Hit):T(MSG_Endstop_Open)));
      #endif
      #if defined(X_MAX_PIN) && X_MAX_PIN > -1
        my_printer.message(T(MSG_X_Max_Pin));
        my_printer.f_message(T((READ(X_MAX_PIN)^X_MAX_ENDSTOP_INVERTING)?MSG_Endstop_Hit:MSG_Endstop_Open));
//        SERIAL_PROTOCOLLN(((READ(X_MAX_PIN)^X_MAX_ENDSTOP_INVERTING)?T(MSG_Endstop_Hit):T(MSG_Endstop_Open));
      #endif
      #if defined(Y_MIN_PIN) && Y_MIN_PIN > -1
        SERIAL_PROTOCOLPGM(MSG_Y_MIN);
        SERIAL_PROTOCOLLN(((READ(Y_MIN_PIN)^Y_MIN_ENDSTOP_INVERTING)?MSG_ENDSTOP_HIT:MSG_ENDSTOP_OPEN));
      #endif
      #if defined(Y_MAX_PIN) && Y_MAX_PIN > -1
        my_printer.message(T(MSG_Y_Max_Pin));
        my_printer.f_message(T((READ(Y_MAX_PIN)^Y_MAX_ENDSTOP_INVERTING)?MSG_Endstop_Hit:MSG_Endstop_Open));
      #endif
      #if defined(Z_MIN_PIN) && Z_MIN_PIN > -1
        my_printer.message(T(MSG_Z_Min_Pin));
        my_printer.f_message(T((READ(Z_MIN_PIN)^Z_MIN_ENDSTOP_INVERTING)?MSG_Endstop_Hit:MSG_Endstop_Open));
      #endif
      #if defined(Z_MAX_PIN) && Z_MAX_PIN > -1
        SERIAL_PROTOCOLPGM(MSG_Z_MAX);
        SERIAL_PROTOCOLLN(((READ(Z_MAX_PIN)^Z_MAX_ENDSTOP_INVERTING)?MSG_ENDSTOP_HIT:MSG_ENDSTOP_OPEN));
      #endif
      break;
      //TODO: update for all axis, use for loop
    #ifdef BLINKM
    case 150: // M150
      {
        byte red;
        byte grn;
        byte blu;

        if(code_seen('R')) red = code_value();
        if(code_seen('U')) grn = code_value();
        if(code_seen('B')) blu = code_value();

        SendColors(red,grn,blu);
      }
      break;
    #endif //BLINKM
    case 200: // M200 D<millimeters> set filament diameter and set E axis units to cubic millimeters (use S0 to set back to millimeters).
      {
        float area = .0;
        float radius = .0;
        if(code_seen('D')) {
          radius = (float)code_value() * .5;
          if(radius == 0) {
            area = 1;
          } else {
            area = M_PI * pow(radius, 2);
          }
        } else {
          //reserved for setting filament diameter via UFID or filament measuring device
          break;
        }
        target_extruder = active_extruder;
        if(code_seen('T')) {
        	target_extruder = code_value();
          if(target_extruder >= EXTRUDERS) {
//            SERIAL_ECHO_START;
          	  my_printer.f_message(T(MSG_Invalid_Extruder),200,target_extruder);
//        	  my_printer.echo(MSG_M200_INVALID_EXTRUDER);
          }
        }
        volumetric_multiplier[target_extruder] = 1 / area;
      }
      break;
    case 201: // M201
      for(int8_t i=0; i < NUM_AXIS; i++)
      {
        if(code_seen(axis_codes[i]))
        {
          config.max_acceleration_units_per_sq_second[i] = code_value();
        }
      }
      // steps per sq second need to be updated to agree with the units per sq second (as they are what is used in the planner)
      //reset_acceleration_rates();
      config.Reset_AccelerationRates();
      break;
    #if 0 // Not used for Sprinter/grbl gen6
    case 202: // M202
      for(int8_t i=0; i < NUM_AXIS; i++) {
        if(code_seen(axis_codes[i])) axis_travel_steps_per_sqr_second[i] = code_value() * axis_steps_per_unit[i];
      }
      break;
    #endif
    case 203: // M203 max feedrate mm/sec
      for(int8_t i=0; i < NUM_AXIS; i++) {
        if(code_seen(axis_codes[i])) config.max_feedrate_mm_s[i] = code_value();
      }
      break;
    case 204: // M204 acclereration S normal moves T filmanent only moves
      {
        if(code_seen('S')) config.acceleration = code_value() ;
        if(code_seen('T')) config.retract_acceleration = code_value() ;
        if(code_seen('M')) config.travel_acceleration = code_value() ;
      }
      break;
    case 205: //M205 advanced settings:  minimum travel speed S=while printing T=travel only,  B=minimum segment time X= maximum xy jerk, Z=maximum Z jerk
    {
      if(code_seen('S')) config.min_feedrate_mm_s = code_value();
      if(code_seen('T')) config.min_travel_feedrate_mm_s = code_value();
      if(code_seen('B')) config.minsegmenttime = code_value() ;
      if(code_seen('X')) config.max_jerk[X_AXIS] = config.max_jerk[Y_AXIS] = code_value() ;
      if(code_seen('Z')) config.max_jerk[Z_AXIS] = code_value() ;
      if(code_seen('E')) config.max_jerk[E_AXIS] = code_value() ;
     }
    break;
    case 206: // M206 additional homeing offset
      for(int8_t i=0; i < 3; i++)
      {
        if(code_seen(axis_codes[i])) config.add_homeing[i] = code_value();
      }
      break;
    #ifdef DELTA
	case 665: // M665 set delta configurations L<diagonal_rod> R<delta_radius> S<segments_per_sec>
		if(code_seen('L')) {
			delta_diagonal_rod= code_value();
		}
		if(code_seen('R')) {
			delta_radius= code_value();
		}
		if(code_seen('S')) {
			delta_segments_per_second= code_value();
		}
		
		recalc_delta_settings(delta_radius, delta_diagonal_rod);
		break;
    case 666: // M666 set delta endstop adjustemnt
      for(int8_t i=0; i < 3; i++)
      {
        if(code_seen(axis_codes[i])) endstop_adj[i] = code_value();
      }
      break;
    #endif
    #ifdef FWRETRACT
    case 207: //M207 - set retract length S[positive mm] F[feedrate mm/min] Z[additional zlift/hop]
    {
      if(code_seen('S'))
      {
        retract_length = code_value() ;
      }
      if(code_seen('F'))
      {
        retract_feedrate = code_value()/60 ;
      }
      if(code_seen('Z'))
      {
        retract_zlift = code_value() ;
      }
    }break;
    case 208: // M208 - set retract recover length S[positive mm surplus to the M207 S*] F[feedrate mm/min]
    {
      if(code_seen('S'))
      {
        retract_recover_length = code_value() ;
      }
      if(code_seen('F'))
      {
        retract_recover_feedrate = code_value()/60 ;
      }
    }break;
    case 209: // M209 - S<1=true/0=false> enable automatic retract detect if the slicer did not support G10/11: every normal extrude-only move will be classified as retract depending on the direction.
    {
      if(code_seen('S'))
      {
        int t= code_value() ;
        switch(t)
        {
          case 0: autoretract_enabled=false;retracted=false;break;
          case 1: autoretract_enabled=true;retracted=false;break;
          default:
            SERIAL_ECHO_START;
            SERIAL_ECHOPGM(MSG_UNKNOWN_COMMAND);
            SERIAL_ECHO(cmdbuffer[cmd_queue_index_r]);
            SERIAL_ECHOLNPGM("\"");
        }
      }

    }break;
    #endif // FWRETRACT
#ifdef PRESSURE_ADVANCE
  // set advance factors for the extruders
    case 214:
     	gcode_M214();
     	break;
#endif

    #if EXTRUDERS > 1
    case 218: // M218 - set hotend offset (in mm), T<extruder_number> X<offset_on_X> Y<offset_on_Y>
    {
    	gcode_M218();
    }break;
    #endif
    case 220: // M220 S<factor in percent>- set speed factor override percentage
    {
      if(code_seen('S'))
      {
        feedmultiply = code_value() ;
      }
    }
    break;
    case 221: // M221 S<factor in percent>- set extrude factor override percentage
    {
      if(code_seen('S'))
      {
        int tmp_code = code_value();
        if (code_seen('T'))
        {
          if(setTargetedHotend(221)){
            break;
          }
          extruder_multiply[target_extruder] = tmp_code;
        }
        else
        {
          extrudemultiply = tmp_code ;
        }
      }
    }
    break;

	case 226: // M226 P<pin number> S<pin state>- Wait until the specified pin reaches the state required
	{
      if(code_seen('P')){
        int pin_number = code_value(); // pin number
        int pin_state = -1; // required pin state - default is inverted

        if(code_seen('S')) pin_state = code_value(); // required pin state

        if(pin_state >= -1 && pin_state <= 1){

          for(int8_t i = 0; i < (int8_t)sizeof(sensitive_pins); i++)
          {
            if (sensitive_pins[i] == pin_number)
            {
              pin_number = -1;
              break;
            }
          }

          if (pin_number > -1)
          {
        	  motor.Synchronize();

            pinMode(pin_number, INPUT);

            int target;
            switch(pin_state){
            case 1:
              target = HIGH;
              break;

            case 0:
              target = LOW;
              break;

            case -1:
              target = !digitalRead(pin_number);
              break;
            }

            while(digitalRead(pin_number) != target)
            {
            	my_printer.Go_Idle();
            }
          }
        }
      }
    }
    break;

    #if NUM_SERVOS > 0
    case 280: // M280 - set servo position absolute. P: servo index, S: angle or microseconds
      {
        int servo_index = -1;
        int servo_position = 0;
        if (code_seen('P'))
          servo_index = code_value();
        if (code_seen('S')) {
          servo_position = code_value();
          if ((servo_index >= 0) && (servo_index < NUM_SERVOS)) {
#if defined (ENABLE_AUTO_BED_LEVELING) && (PROBE_SERVO_DEACTIVATION_DELAY > 0)
		      servos[servo_index].attach(0);
#endif
            servos[servo_index].write(servo_position);
#if defined (ENABLE_AUTO_BED_LEVELING) && (PROBE_SERVO_DEACTIVATION_DELAY > 0)
              delay(PROBE_SERVO_DEACTIVATION_DELAY);
              servos[servo_index].detach();
#endif
          }
          else {
            SERIAL_ECHO_START;
            SERIAL_ECHO("Servo ");
            SERIAL_ECHO(servo_index);
            SERIAL_ECHOLN(" out of range");
          }
        }
        else if (servo_index >= 0) {
          SERIAL_PROTOCOL(MSG_OK);
          SERIAL_PROTOCOL(" Servo ");
          SERIAL_PROTOCOL(servo_index);
          SERIAL_PROTOCOL(": ");
          SERIAL_PROTOCOL(servos[servo_index].read());
          SERIAL_PROTOCOLLN("");
        }
      }
      break;
    #endif // NUM_SERVOS > 0

    #if (LARGE_FLASH == true && ( BEEPER > 0 || defined(ULTRALCD) || defined(LCD_USE_I2C_BUZZER)))
    case 300: // M300
    {
      int beepS = code_seen('S') ? code_value() : 110;
      int beepP = code_seen('P') ? code_value() : 1000;
      if (beepS > 0)
      {
        #if BEEPER > 0
          tone(BEEPER, beepS);
          delay(beepP);
          noTone(BEEPER);
        #elif defined(ULTRALCD)
		  lcd_buzz(beepS, beepP);
		#elif defined(LCD_USE_I2C_BUZZER)
		  lcd_buzz(beepP, beepS);
        #endif
      }
      else
      {
        delay(beepP);
      }
    }
    break;
    #endif // M300

    #ifdef PIDTEMP
    case 301: // M301
    	  gcode_M301();
      break;
    #endif //PIDTEMP
    #ifdef PIDTEMPBED
    case 304: // M304
      {
//        if(code_seen('P')) bedKp = code_value();
//        if(code_seen('I')) bedKi = scalePID_i(code_value());
//        if(code_seen('D')) bedKd = scalePID_d(code_value());
//
//        updatePID();
//        SERIAL_PROTOCOL(MSG_OK);
//        SERIAL_PROTOCOL(" p:");
//        SERIAL_PROTOCOL(bedKp);
//        SERIAL_PROTOCOL(" i:");
//        SERIAL_PROTOCOL(unscalePID_i(bedKi));
//        SERIAL_PROTOCOL(" d:");
//        SERIAL_PROTOCOL(unscalePID_d(bedKd));
//        SERIAL_PROTOCOLLN("");
      }
      break;
    #endif //PIDTEMP
    case 240: // M240  Triggers a camera by emulating a Canon RC-1 : http://www.doc-diy.net/photo/rc-1_hacked/
     {
     	#ifdef CHDK
       
         SET_OUTPUT(CHDK);
         WRITE(CHDK, HIGH);
         chdkHigh = millis();
         chdkActive = true;
       
       #else
     	
      	#if defined(PHOTOGRAPH_PIN) && PHOTOGRAPH_PIN > -1
	const uint8_t NUM_PULSES=16;
	const float PULSE_LENGTH=0.01524;
	for(int i=0; i < NUM_PULSES; i++) {
        WRITE(PHOTOGRAPH_PIN, HIGH);
        _delay_ms(PULSE_LENGTH);
        WRITE(PHOTOGRAPH_PIN, LOW);
        _delay_ms(PULSE_LENGTH);
        }
        delay(7.33);
        for(int i=0; i < NUM_PULSES; i++) {
        WRITE(PHOTOGRAPH_PIN, HIGH);
        _delay_ms(PULSE_LENGTH);
        WRITE(PHOTOGRAPH_PIN, LOW);
        _delay_ms(PULSE_LENGTH);
        }
      	#endif
      #endif //chdk end if
     }
    break;
#ifdef DOGLCD
    case 250: // M250  Set LCD contrast value: C<value> (value 0..63)
     {
	  if (code_seen('C')) {
	   lcd_setcontrast( ((int)code_value())&63 );
          }
          SERIAL_PROTOCOLPGM("lcd contrast value: ");
          SERIAL_PROTOCOL(lcd_contrast);
          SERIAL_PROTOCOLLN("");
     }
    break;
#endif
    #ifdef PREVENT_DANGEROUS_EXTRUDE
    case 302: // allow cold extrudes, or set the minimum extrude temperature
    {
	  float temp = .0;
	  if (code_seen('S')) temp=code_value();
      heater.set_extrude_min_temp(temp);
    }
    break;
	#endif
    case 303: // M303 PID autotune
    	gcode_M303();
    break;
    case 400: // M400 finish all moves
    {
    	motor.Synchronize();
    }
    break;
#if defined(ENABLE_AUTO_BED_LEVELING) && defined(SERVO_ENDSTOPS)
    case 401:
    {
        engage_z_probe();    // Engage Z Servo endstop if available
    }
    break;

    case 402:
    {
        retract_z_probe();    // Retract Z Servo endstop if enabled
    }
    break;
#endif
#ifdef LEDCONTROL
  case 420: //M420 - Rxxx Exxx Bxxx 		
  
    if (code_seen('R'))
      digitalWrite(REDPIN, code_value() != 0 ? 1 : 0);
      
    if (code_seen('E'))
      digitalWrite(GREENPIN, code_value() != 0 ? 1 : 0);
      
    if (code_seen('B'))
      digitalWrite(BLUEPIN, code_value() != 0 ? 1 : 0);

    break;
#endif
    case 500: // M500 Store settings in EEPROM
    {
    	config.Store_EEPROM();
    }
    break;
    case 501: // M501 Read settings from EEPROM
    {
    	config.Load_EEPROM();
    }
    break;
    case 502: // M502 Revert to default settings
    {
    	config.Load_Failsave();
    }
    break;
    case 503: // M503 print settings currently in memory
    {
      config.Print_Settings();
    }
    break;
    #ifdef ABORT_ON_ENDSTOP_HIT_FEATURE_ENABLED
    case 540:
    {
        if(code_seen('S')) abort_on_endstop_hit = code_value() > 0;
    }
    break;
    #endif

    #ifdef CUSTOM_M_CODE_SET_Z_PROBE_OFFSET
    case CUSTOM_M_CODE_SET_Z_PROBE_OFFSET:
    {
      float value;
      if (code_seen('Z'))
      {
        value = code_value();
        if ((Z_PROBE_OFFSET_RANGE_MIN <= value) && (value <= Z_PROBE_OFFSET_RANGE_MAX))
        {
          zprobe_zoffset = -value; // compare w/ line 278 of ConfigurationStore.cpp
          my_printer.echo();
          SERIAL_ECHOLNPGM(MSG_ZPROBE_ZOFFSET " " MSG_OK);
          SERIAL_PROTOCOLLN("");
        }
        else
        {
        	my_printer.echo();
          SERIAL_ECHOPGM(MSG_ZPROBE_ZOFFSET);
          SERIAL_ECHOPGM(MSG_Z_MIN);
          SERIAL_ECHO(Z_PROBE_OFFSET_RANGE_MIN);
          SERIAL_ECHOPGM(MSG_Z_MAX);
          SERIAL_ECHO(Z_PROBE_OFFSET_RANGE_MAX);
          SERIAL_PROTOCOLLN("");
        }
      }
      else
      {
    	  my_printer.echo();
    	  SERIAL_ECHOLNPGM(MSG_ZPROBE_ZOFFSET " : ");
          SERIAL_ECHO(-zprobe_zoffset);
          SERIAL_PROTOCOLLN("");
      }
      break;
    }
    #endif // CUSTOM_M_CODE_SET_Z_PROBE_OFFSET

    #ifdef FILAMENTCHANGEENABLE
    case 600: //Pause for filament change X[pos] Y[pos] Z[relative lift] E[initial retract] L[later retract distance for removal]
    {
        float target[4];
       float lastpos[4];
        target[X_AXIS]=Position.current_position_mm[X_AXIS];
        target[Y_AXIS]=Position.current_position_mm[Y_AXIS];
        target[Z_AXIS]=Position.current_position_mm[Z_AXIS];
        target[E_AXIS]=Position.current_position_mm[E_AXIS];
        lastpos[X_AXIS]=Position.current_position_mm[X_AXIS];
        lastpos[Y_AXIS]=Position.current_position_mm[Y_AXIS];
        lastpos[Z_AXIS]=Position.current_position_mm[Z_AXIS];
        lastpos[E_AXIS]=Position.current_position_mm[E_AXIS];
        //retract by E
        if(code_seen('E'))
        {
          target[E_AXIS]+= code_value();
        }
        else
        {
          #ifdef FILAMENTCHANGE_FIRSTRETRACT
            target[E_AXIS]+= FILAMENTCHANGE_FIRSTRETRACT ;
          #endif
        }
//        plan_buffer_line(target[X_AXIS], target[Y_AXIS], target[Z_AXIS], target[E_AXIS], feedrate/60, active_extruder);

        //lift Z
        if(code_seen('Z'))
        {
          target[Z_AXIS]+= code_value();
        }
        else
        {
          #ifdef FILAMENTCHANGE_ZADD
            target[Z_AXIS]+= FILAMENTCHANGE_ZADD ;
          #endif
        }
//        plan_buffer_line(target[X_AXIS], target[Y_AXIS], target[Z_AXIS], target[E_AXIS], feedrate/60, active_extruder);

        //move xy
        if(code_seen('X'))
        {
          target[X_AXIS]+= code_value();
        }
        else
        {
          #ifdef FILAMENTCHANGE_XPOS
            target[X_AXIS]= FILAMENTCHANGE_XPOS ;
          #endif
        }
        if(code_seen('Y'))
        {
          target[Y_AXIS]= code_value();
        }
        else
        {
          #ifdef FILAMENTCHANGE_YPOS
            target[Y_AXIS]= FILAMENTCHANGE_YPOS ;
          #endif
        }

//       plan_buffer_line(target[X_AXIS], target[Y_AXIS], target[Z_AXIS], target[E_AXIS], Position.Feedrate()/60, active_extruder);

        if(code_seen('L'))
        {
          target[E_AXIS]+= code_value();
        }
        else
        {
          #ifdef FILAMENTCHANGE_FINALRETRACT
            target[E_AXIS]+= FILAMENTCHANGE_FINALRETRACT ;
          #endif
        }

//        plan_buffer_line(target[X_AXIS], target[Y_AXIS], target[Z_AXIS], target[E_AXIS], feedrate/60, active_extruder);

        //finish moves
        motor.Synchronize();
        //disable extruder steppers so filament can be removed
        disable_E0();
        disable_E1();
        disable_E2();
        delay(100);
    //    LCD_ALERTMESSAGEPGM(MSG_FILAMENTCHANGE);
        uint8_t cnt=0;
        while(!lcd_clicked()){
          cnt++;
//          heater.Manage();
//          manage_inactivity();
//        lcd_update();
          my_printer.Go_Idle();
          if(cnt==0)
          {
          #if BEEPER > 0
            SET_OUTPUT(BEEPER);

            WRITE(BEEPER,HIGH);
            delay(3);
            WRITE(BEEPER,LOW);
            delay(3);
          #else
			#if !defined(LCD_FEEDBACK_FREQUENCY_HZ) || !defined(LCD_FEEDBACK_FREQUENCY_DURATION_MS)
   //           lcd_buzz(1000/6,100);
			#else
			  lcd_buzz(LCD_FEEDBACK_FREQUENCY_DURATION_MS,LCD_FEEDBACK_FREQUENCY_HZ);
			#endif
          #endif
          }
        }

        //return to normal
        if(code_seen('L'))
        {
          target[E_AXIS]+= -code_value();
        }
        else
        {
          #ifdef FILAMENTCHANGE_FINALRETRACT
            target[E_AXIS]+=(-1)*FILAMENTCHANGE_FINALRETRACT ;
          #endif
        }
//        current_position[E_AXIS]=target[E_AXIS]; //the long retract of L is compensated by manual filament feeding
//        plan_set_e_position(current_position[E_AXIS]);
//        plan_buffer_line(target[X_AXIS], target[Y_AXIS], target[Z_AXIS], target[E_AXIS], feedrate/60, active_extruder); //should do nothing
//        plan_buffer_line(lastpos[X_AXIS], lastpos[Y_AXIS], target[Z_AXIS], target[E_AXIS], feedrate/60, active_extruder); //move xy back
//        plan_buffer_line(lastpos[X_AXIS], lastpos[Y_AXIS], lastpos[Z_AXIS], target[E_AXIS], feedrate/60, active_extruder); //move z back
//        plan_buffer_line(lastpos[X_AXIS], lastpos[Y_AXIS], lastpos[Z_AXIS], lastpos[E_AXIS], feedrate/60, active_extruder); //final untretract
    }
    break;
    #endif //FILAMENTCHANGEENABLE
    #ifdef DUAL_X_CARRIAGE
    case 605: // Set dual x-carriage movement mode:
              //    M605 S0: Full control mode. The slicer has full control over x-carriage movement
              //    M605 S1: Auto-park mode. The inactive head will auto park/unpark without slicer involvement
              //    M605 S2 [Xnnn] [Rmmm]: Duplication mode. The second extruder will duplicate the first with nnn
              //                         millimeters x-offset and an optional differential hotend temperature of
              //                         mmm degrees. E.g., with "M605 S2 X100 R2" the second extruder will duplicate
              //                         the first with a spacing of 100mm in the x direction and 2 degrees hotter.
              //
              //    Note: the X axis should be homed after changing dual x-carriage mode.
    {
        st_synchronize();

        if (code_seen('S'))
          dual_x_carriage_mode = code_value();

        if (dual_x_carriage_mode == DXC_DUPLICATION_MODE)
        {
          if (code_seen('X'))
            duplicate_extruder_x_offset = max(code_value(),X2_MIN_POS - x_home_pos(0));

          if (code_seen('R'))
            duplicate_extruder_temp_offset = code_value();

          SERIAL_ECHO_START;
          SERIAL_ECHOPGM(MSG_HOTEND_OFFSET);
          SERIAL_ECHO(" ");
          SERIAL_ECHO(extruder_offset[X_AXIS][0]);
          SERIAL_ECHO(",");
          SERIAL_ECHO(extruder_offset[Y_AXIS][0]);
          SERIAL_ECHO(" ");
          SERIAL_ECHO(duplicate_extruder_x_offset);
          SERIAL_ECHO(",");
          SERIAL_ECHOLN(extruder_offset[Y_AXIS][1]);
        }
        else if (dual_x_carriage_mode != DXC_FULL_CONTROL_MODE && dual_x_carriage_mode != DXC_AUTO_PARK_MODE)
        {
          dual_x_carriage_mode = DEFAULT_DUAL_X_CARRIAGE_MODE;
        }

        active_extruder_parked = false;
        extruder_duplication_enabled = false;
        delayed_move_time = 0;
    }
    break;
    #endif //DUAL_X_CARRIAGE

    case 907: // M907 Set digital trimpot motor current using axis codes.
    {

    }
    break;
    case 908: // M908 Control digital trimpot directly.
    {

    }
    break;
    case 350: // M350 Set microstepping mode. Warning: Steps per unit remains unchanged. S code sets stepping mode for all drivers.
    {

    }
    break;
    case 351: // M351 Toggle MS1 MS2 pins directly, S# determines MS1 or MS2, X# sets the pin high/low.
    {

    }
    break;
    case 999: // M999: Restart after being stopped
    	my_printer.Stopped = false;
   //   lcd_reset_alert_level();
      gcode_LastN = Stopped_gcode_LastN;
      FlushSerialRequestResend();
    break;
    default:
 	  my_printer.f_message(T(MSG_Unknown_Command),cmdbuffer[cmd_queue_index_r]);
// 	  char * errchar = &cmdbuffer[cmd_queue_index_r][0];
// 	  while (*errchar)
// 		 MSerial.print (*(errchar++),HEX);

    }
  }

  else if(code_seen('T'))
  {
	  int tool = (int)code_value();
	  float feedrate = 0;

		if (code_seen('F'))
		{
		feedrate = code_value();
		}
	  Position.Set_Tool(tool,feedrate);
  }

  else
  {
    my_printer.f_message(T(MSG_Unknown_Command),cmdbuffer[cmd_queue_index_r]);
  }

  ClearToSend();
}

void FlushSerialRequestResend()
{
  //char cmdbuffer[cmd_queue_index_r][100]="Resend:";
	MSerial.flush();
//  SERIAL_PROTOCOLPGM(MSG_RESEND);
//  SERIAL_PROTOCOLLN(gcode_LastN + 1);
  ClearToSend();
}

void ClearToSend()
{
	my_printer.previous_millis_cmd = millis();
  #ifdef SDSUPPORT
  if(fromsd[cmd_queue_index_r])
    return;
  #endif //SDSUPPORT
  my_printer.message(T(MSG_OK));
  my_printer.message (PSTR("\n"));
}

void get_coordinates()
{
  // bool seen[4]={false,false,false,false};
  for(int8_t i=0; i < NUM_AXIS; i++) {
    if(code_seen(axis_codes[i]))
    {
    	Position.destination[i] = (float)code_value();
    	if (axis_relative_modes[i] || relative_mode)
    		Position.destination[i] += Position.current_position_mm[i];
     // seen[i]=true;
    }
    else
    	Position.destination[i] = Position.current_position_mm[i];
  }
  if(code_seen('F'))
  {
    float next_feedrate = code_value();
    if(next_feedrate > 0.0)
    	Position.Set_Feedrate( next_feedrate);
  }
}

void get_arc_coordinates()
{
#ifdef SF_ARC_FIX
   bool relative_mode_backup = relative_mode;
   relative_mode = true;
#endif
   get_coordinates();
#ifdef SF_ARC_FIX
   relative_mode=relative_mode_backup;
#endif

   if(code_seen('I')) {
     offset[0] = code_value();
   }
   else {
     offset[0] = 0.0;
   }
   if(code_seen('J')) {
     offset[1] = code_value();
   }
   else {
     offset[1] = 0.0;
   }
}

void clamp_to_software_endstops(float target[3])
{
  if (min_software_endstops) {
    if (target[X_AXIS] < min_pos[X_AXIS]) target[X_AXIS] = min_pos[X_AXIS];
    if (target[Y_AXIS] < min_pos[Y_AXIS]) target[Y_AXIS] = min_pos[Y_AXIS];
    if (target[Z_AXIS] < min_pos[Z_AXIS]) target[Z_AXIS] = min_pos[Z_AXIS];
  }

  if (max_software_endstops) {
    if (target[X_AXIS] > max_pos[X_AXIS]) target[X_AXIS] = max_pos[X_AXIS];
    if (target[Y_AXIS] > max_pos[Y_AXIS]) target[Y_AXIS] = max_pos[Y_AXIS];
    if (target[Z_AXIS] > max_pos[Z_AXIS]) target[Z_AXIS] = max_pos[Z_AXIS];
  }
}

#ifdef DELTA
void recalc_delta_settings(float radius, float diagonal_rod)
{
	 delta_tower1_x= -SIN_60*radius; // front left tower
	 delta_tower1_y= -COS_60*radius;	   
	 delta_tower2_x=  SIN_60*radius; // front right tower
	 delta_tower2_y= -COS_60*radius;	   
	 delta_tower3_x= 0.0;                  // back middle tower
	 delta_tower3_y= radius;
	 delta_diagonal_rod_2= sq(diagonal_rod);
}

void calculate_delta(float cartesian[3])
{
  delta[X_AXIS] = sqrt(delta_diagonal_rod_2
                       - sq(delta_tower1_x-cartesian[X_AXIS])
                       - sq(delta_tower1_y-cartesian[Y_AXIS])
                       ) + cartesian[Z_AXIS];
  delta[Y_AXIS] = sqrt(delta_diagonal_rod_2
                       - sq(delta_tower2_x-cartesian[X_AXIS])
                       - sq(delta_tower2_y-cartesian[Y_AXIS])
                       ) + cartesian[Z_AXIS];
  delta[Z_AXIS] = sqrt(delta_diagonal_rod_2
                       - sq(delta_tower3_x-cartesian[X_AXIS])
                       - sq(delta_tower3_y-cartesian[Y_AXIS])
                       ) + cartesian[Z_AXIS];
  /*
  SERIAL_ECHOPGM("cartesian x="); SERIAL_ECHO(cartesian[X_AXIS]);
  SERIAL_ECHOPGM(" y="); SERIAL_ECHO(cartesian[Y_AXIS]);
  SERIAL_ECHOPGM(" z="); SERIAL_ECHOLN(cartesian[Z_AXIS]);

  SERIAL_ECHOPGM("delta x="); SERIAL_ECHO(delta[X_AXIS]);
  SERIAL_ECHOPGM(" y="); SERIAL_ECHO(delta[Y_AXIS]);
  SERIAL_ECHOPGM(" z="); SERIAL_ECHOLN(delta[Z_AXIS]);
  */
}
#endif


void prepare_arc_move(char isclockwise) {
  float r = hypot(offset[X_AXIS], offset[Y_AXIS]); // Compute arc radius for mc_arc

  // Trace the arc
  mc_arc(Position.current_position_mm, Position.destination, offset, X_AXIS, Y_AXIS, Z_AXIS, Position.Feedrate()*feedmultiply/60/100.0, r, isclockwise, active_extruder);

  // As far as the parser is concerned, the position is now == target. In reality the
  // motion control system might still be processing the action and the real tool position
  // in any intermediate location.
  for(int8_t i=0; i < NUM_AXIS; i++) {
	  Position.current_position_mm[i] = Position.destination[i];
  }
  my_printer.previous_millis_cmd = millis();
}

//#if defined(CONTROLLERFAN_PIN) && CONTROLLERFAN_PIN > -1
//
//#if defined(FAN_PIN)
//  #if CONTROLLERFAN_PIN == FAN_PIN
//    #error "You cannot set CONTROLLERFAN_PIN equal to FAN_PIN"
//  #endif
//#endif
//
//
//
//void controllerFan()
//{
//  if ((millis() - my_printer.lastMotorCheck) >= 2500) //Not a time critical function, so we only check every 2500ms
//  {
//	  my_printer.lastMotorCheck = millis();
//
//   // if(!READ(X_ENABLE_PIN) || !READ(Y_ENABLE_PIN) || !READ(Z_ENABLE_PIN) || (soft_pwm_bed > 0)
//    	    if(!READ(X_ENABLE_PIN) || !READ(Y_ENABLE_PIN) || !READ(Z_ENABLE_PIN)
//    #if EXTRUDERS > 2
//       || !READ(E2_ENABLE_PIN)
//    #endif
//    #if EXTRUDER > 1
//      #if defined(X2_ENABLE_PIN) && X2_ENABLE_PIN > -1
//       || !READ(X2_ENABLE_PIN)
//      #endif
//       || !READ(E1_ENABLE_PIN)
//    #endif
//       || !READ(E0_ENABLE_PIN)) //If any of the drivers are enabled...
//    {
//    	    	my_printer.lastMotor = millis(); //... set time to NOW so the fan will turn on
//    }
//
//    if ((millis() - my_printer.lastMotor) >= (CONTROLLERFAN_SECS*1000UL) || my_printer.lastMotor == 0) //If the last time any driver was enabled, is longer since than CONTROLLERSEC...
//    {
//        digitalWrite(CONTROLLERFAN_PIN, 0);
//        analogWrite(CONTROLLERFAN_PIN, 0);
//    }
//    else
//    {
//        // allows digital or PWM fan output to be used (see M42 handling)
//        digitalWrite(CONTROLLERFAN_PIN, CONTROLLERFAN_SPEED);
//        analogWrite(CONTROLLERFAN_PIN, CONTROLLERFAN_SPEED);
//    }
//  }
//}
//#endif

#ifdef TEMP_STAT_LEDS
static bool blue_led = false;
static bool red_led = false;
static uint32_t stat_update = 0;

void handle_status_leds(void) {
  float max_temp = 0.0;
  if(millis() > stat_update) {
    stat_update += 500; // Update every 0.5s
    for (int8_t cur_extruder = 0; cur_extruder < EXTRUDERS; ++cur_extruder) {
       max_temp = max(max_temp, degHotend(cur_extruder));
       max_temp = max(max_temp, degTargetHotend(cur_extruder));
    }
    #if defined(TEMP_BED_PIN) && TEMP_BED_PIN > -1
      max_temp = max(max_temp, degTargetBed());
      max_temp = max(max_temp, degBed());
    #endif
    if((max_temp > 55.0) && (red_led == false)) {
      digitalWrite(STAT_LED_RED, 1);
      digitalWrite(STAT_LED_BLUE, 0);
      red_led = true;
      blue_led = false;
    }
    if((max_temp < 54.0) && (blue_led == false)) {
      digitalWrite(STAT_LED_RED, 0);
      digitalWrite(STAT_LED_BLUE, 1);
      red_led = false;
      blue_led = true;
    }
  }
}
#endif

//void manage_inactivity()
//{
//  if( (millis() - my_printer.previous_millis_cmd) >  my_printer.max_inactive_time )
//    if(my_printer.max_inactive_time)
//      kill();
//  if(my_printer.stepper_inactive_time)  {
//    if( (millis() - my_printer.previous_millis_cmd) >  my_printer.stepper_inactive_time )
//    {
//      if(plan.Blocks_Queued() == false)
//      {
//        disable_X();
//        disable_Y();
//       disable_Z();
//       disable_E0();
//        disable_E1();
//      }
//    }
//  }
//
//  #ifdef CHDK //Check if pin should be set to LOW after M240 set it to HIGH
//    if (chdkActive)
//    {
//      chdkActive = false;
//      if (millis()-chdkHigh < CHDK_DELAY) return;
//      WRITE(CHDK, LOW);
//    }
//  #endif
//
//  #if defined(KILL_PIN) && KILL_PIN > -1
//    if( 0 == READ(KILL_PIN) )
//      kill();
//  #endif
//  #if defined(CONTROLLERFAN_PIN) && CONTROLLERFAN_PIN > -1
//    controllerFan(); //Check if fan should be turned on to cool stepper drivers down
//  #endif
//  #ifdef EXTRUDER_RUNOUT_PREVENT
//    if( (millis() - previous_millis_cmd) >  EXTRUDER_RUNOUT_SECONDS*1000 )
//    if(degHotend(active_extruder)>EXTRUDER_RUNOUT_MINTEMP)
//    {
//     bool oldstatus=READ(E0_ENABLE_PIN);
//     enable_e0();
//     float oldepos=current_position[E_AXIS];
//     float oldedes=destination[E_AXIS];
//     plan_buffer_line(destination[X_AXIS], destination[Y_AXIS], destination[Z_AXIS],
//                      destination[E_AXIS]+EXTRUDER_RUNOUT_EXTRUDE*EXTRUDER_RUNOUT_ESTEPS/axis_steps_per_unit[E_AXIS],
//                      EXTRUDER_RUNOUT_SPEED/60.*EXTRUDER_RUNOUT_ESTEPS/axis_steps_per_unit[E_AXIS], active_extruder);
//     current_position[E_AXIS]=oldepos;
//     destination[E_AXIS]=oldedes;
//     plan_set_e_position(oldepos);
//     previous_millis_cmd=millis();
//     st_synchronize();
//     WRITE(E0_ENABLE_PIN,oldstatus);
//    }
//  #endif
//  #if defined(DUAL_X_CARRIAGE)
//    // handle delayed move timeout
//    if (delayed_move_time != 0 && (millis() - delayed_move_time) > 1000 && Stopped == false)
//    {
//      // travel moves have been received so enact them
//      delayed_move_time = 0xFFFFFFFFUL; // force moves to be done
//      memcpy(destination,current_position,sizeof(destination));
//      prepare_move();
//    }
//  #endif
//  #ifdef TEMP_STAT_LEDS
//      handle_status_leds();
//  #endif
//  plan.Check_Axes_Activity();
//}

void kill()
{
  //cli(); // Stop interrupts
 // disable_heater();
  heater.Disable();

  disable_X();
  disable_Y();
  disable_Z();
  disable_E0();
  disable_E1();
  disable_E2();

#if defined(PS_ON_PIN) && PS_ON_PIN > -1
  pinMode(PS_ON_PIN,INPUT);
#endif
  my_printer.error();
  my_printer.error_ln(T(MSG_Err_Killed));
//  LCD_ALERTMESSAGEPGM(MSG_KILLED);
  suicide();
  while(1) { /* Intentionally left empty */ } // Wait for reset
}

void Stop()
{
  heater.Disable();
  if(my_printer.Stopped == false) {
	  my_printer.Stopped = true;
    Stopped_gcode_LastN = gcode_LastN; // Save last g_code for restart
    my_printer.error();
    my_printer.error_ln(T(MSG_Err_Stopped));
    lcd_setstatuspgm(T(MSG_Stopped));
  }
}

bool IsStopped() { return my_printer.Stopped; }

#ifdef FAST_PWM_FAN
void setPwmFrequency(uint8_t pin, int val)
{
  val &= 0x07;
  switch(digitalPinToTimer(pin))
  {

    #if defined(TCCR0A)
    case TIMER0A:
    case TIMER0B:
//         TCCR0B &= ~(_BV(CS00) | _BV(CS01) | _BV(CS02));
//         TCCR0B |= val;
         break;
    #endif

    #if defined(TCCR1A)
    case TIMER1A:
    case TIMER1B:
//         TCCR1B &= ~(_BV(CS10) | _BV(CS11) | _BV(CS12));
//         TCCR1B |= val;
         break;
    #endif

    #if defined(TCCR2)
    case TIMER2:
    case TIMER2:
         TCCR2 &= ~(_BV(CS10) | _BV(CS11) | _BV(CS12));
         TCCR2 |= val;
         break;
    #endif

    #if defined(TCCR2A)
    case TIMER2A:
    case TIMER2B:
         TCCR2B &= ~(_BV(CS20) | _BV(CS21) | _BV(CS22));
         TCCR2B |= val;
         break;
    #endif

    #if defined(TCCR3A)
    case TIMER3A:
    case TIMER3B:
    case TIMER3C:
         TCCR3B &= ~(_BV(CS30) | _BV(CS31) | _BV(CS32));
         TCCR3B |= val;
         break;
    #endif

    #if defined(TCCR4A)
    case TIMER4A:
    case TIMER4B:
    case TIMER4C:
         TCCR4B &= ~(_BV(CS40) | _BV(CS41) | _BV(CS42));
         TCCR4B |= val;
         break;
   #endif

    #if defined(TCCR5A)
    case TIMER5A:
    case TIMER5B:
    case TIMER5C:
         TCCR5B &= ~(_BV(CS50) | _BV(CS51) | _BV(CS52));
         TCCR5B |= val;
         break;
   #endif

  }
}
#endif //FAST_PWM_FAN

bool setTargetedHotend(int code)
{
	target_extruder = active_extruder;
  if(code_seen('T')) {
	  target_extruder = code_value();
    if(target_extruder >= EXTRUDERS) {
  	  my_printer.f_message(T(MSG_Invalid_Extruder),code,target_extruder);
      return true;
    }
  }
  return false;
}

