/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include "ultralcd.h"
#ifdef ULTRA_LCD
#include "MarlinHOH.h"
#include "language.h"
#include "cardreader.h"
#include "temperature.h"
#include "stepper.h"
#include "parameter.h"
#include "Service.h"
#include "temperature.h"
#include "planner.h"

/**
 * REVERSE_MENU_DIRECTION
 *
 * To reverse the menu direction we need a general way to reverse
 * the direction of the encoder everywhere. So encoderDirection is
 * added to allow the encoder to go the other way.
 *
 * This behavior is limited to scrolling Menus and SD card listings,
 * and is disabled in other contexts.
 */
#ifdef REVERSE_MENU_DIRECTION
  int8_t encoderDirection = 1;
  #define ENCODER_DIRECTION_NORMAL() (encoderDirection = 1)
  #define ENCODER_DIRECTION_MENUS() (encoderDirection = -1)
#else
  #define ENCODER_DIRECTION_NORMAL() ;
  #define ENCODER_DIRECTION_MENUS() ;
#endif

int8_t encoderDiff; /* encoderDiff is updated from interrupt context and added to encoderPosition every LCD update */
bool encoderRateMultiplierEnabled;


#ifdef ULTIPANEL
const float manual_feedrate[] = MANUAL_FEEDRATE;
#endif // ULTIPANEL

/* !Configuration settings */

//Function pointer to menu functions.
typedef void (*menuFunc_t)();

uint8_t lcd_status_message_level;
char lcd_status_message[3*LCD_WIDTH+1] ;

//#ifdef DOGLCD
// #include "dogm_lcd_implementation.h"
//#endif

#ifndef DOGLCD
 #include "ultralcd_implementation_hitachi_HD44780.h"
#endif

/** forward declarations **/

//void copy_and_scalePID_i(int e);
//void copy_and_scalePID_d(int e);

/* Different menus */
static void lcd_status_screen();
#ifdef ULTIPANEL
//extern bool powersupply;
static void lcd_main_menu();
static void lcd_tune_menu();
static void lcd_prepare_menu();
static void lcd_move_menu();
static void lcd_load_menu();
static void lcd_unload_menu();
static void lcd_control_menu();
static void lcd_control_advanced_menu();
static void lcd_led_menu();
static void lcd_control_temperature_menu();
static void lcd_control_temperature_preheat_pla_settings_menu();
static void lcd_control_temperature_preheat_abs_settings_menu();
static void lcd_control_motion_menu();
#ifdef DOGLCD
 static void lcd_set_contrast();
#endif
static void lcd_control_retract_menu();
static void lcd_sdcard_menu();
static void lcd_firmware_menu();

static void lcd_flow_extrude_start();
static void lcd_flow_extrude_stop();
static void lcd_timer_menu();

static void lcd_quick_feedback();//Cause an LCD refresh, and give the user visual or audible feedback that something has happened

/* Different types of actions that can be used in menu items. */
static void menu_action_back(menuFunc_t data);
static void menu_action_submenu(menuFunc_t data);
static void menu_action_function(menuFunc_t data);
static void menu_action_service(enum Services servicenr);
static void menu_action_sdfile( char* filename);
static void menu_action_sddirectory(const char* filename);
static void menu_action_setting_edit_bool(const char* pstr, bool* ptr);

template <class T> static void menu_action_setting_edit (const char* pstr, T* ptr, float minValue, float maxValue , float scale);
template <class T> static void menu_action_setting_edit_callback(const char* pstr, T* ptr, float minValue, float maxValue,float nscale, menuFunc_t callback);


#define ENCODER_FEEDRATE_DEADZONE 2

#if !defined(LCD_I2C_VIKI)
  #ifndef ENCODER_STEPS_PER_MENU_ITEM
    #define ENCODER_STEPS_PER_MENU_ITEM 1
  #endif
  #ifndef ENCODER_PULSES_PER_STEP
    #define ENCODER_PULSES_PER_STEP 4
  #endif
#else
  #ifndef ENCODER_STEPS_PER_MENU_ITEM
    #define ENCODER_STEPS_PER_MENU_ITEM 2 // VIKI LCD rotary encoder uses a different number of steps per rotation
  #endif
  #ifndef ENCODER_PULSES_PER_STEP
    #define ENCODER_PULSES_PER_STEP 1
  #endif
#endif

/** Used variables to keep track of the menu */
#ifndef REPRAPWORLD_KEYPAD
volatile uint8_t buttons;//Contains the bits of the currently pressed buttons.
#else
volatile uint8_t buttons_reprapworld_keypad; // to store the reprapworld_keypad shift register values
#endif
#ifdef LCD_HAS_SLOW_BUTTONS
volatile uint8_t slow_buttons;//Contains the bits of the currently pressed buttons.
#endif
uint8_t currentMenuViewOffset;              /* scroll offset in the current menu */
uint32_t blocking_enc;
uint8_t lastEncoderBits;
uint32_t encoderPosition;
#if (SDCARDDETECT > 0)
bool lcd_oldcardstatus;
#endif


bool itemSelected;
  uint8_t encoderLine;
  uint8_t lcdDrawUpdate = 2;                  /* Set to none-zero when the LCD needs to draw, decreased after every draw. Set to 2 in LCD routines so the LCD gets at least 1 full redraw (first redraw is partial) */

///* Helper macros for menus */
/**
 * START_MENU generates the init code for a menu function
 */
#define START_MENU() do { \
  ENCODER_DIRECTION_MENUS(); \
  encoderRateMultiplierEnabled = false; \
  if (encoderPosition > 0x8000) encoderPosition = 0; \
  encoderLine = encoderPosition / ENCODER_STEPS_PER_MENU_ITEM; \
  NOMORE(currentMenuViewOffset, encoderLine); \
  uint8_t _lineNr = currentMenuViewOffset, _menuItemNr; \
  bool wasClicked = LCD_CLICKED; \
  for (uint8_t _drawLineNr = 0; _drawLineNr < LCD_HEIGHT; _drawLineNr++, _lineNr++) { \
    _menuItemNr = 0;

/**
 * MENU_ITEM generates draw & handler code for a menu item, potentially calling:
 *
 *   lcd_implementation_drawmenu_[type](sel, row, label, arg3...)
 *   menu_action_[type](arg3...)
 *
 * Examples:
 *   MENU_ITEM(back, MSG_WATCH)
 *     lcd_implementation_drawmenu_back(sel, row, PSTR(MSG_WATCH))
 *     menu_action_back()
 *
 *   MENU_ITEM(function, MSG_PAUSE_PRINT, lcd_sdcard_pause)
 *     lcd_implementation_drawmenu_function(sel, row, PSTR(MSG_PAUSE_PRINT), lcd_sdcard_pause)
 *     menu_action_function(lcd_sdcard_pause)
 *
 *   MENU_ITEM_EDIT(int3, MSG_SPEED, &feedrate_multiplier, 10, 999)
 *   MENU_ITEM(setting_edit_int3, MSG_SPEED, PSTR(MSG_SPEED), &feedrate_multiplier, 10, 999)
 *     lcd_implementation_drawmenu_setting_edit_int3(sel, row, PSTR(MSG_SPEED), PSTR(MSG_SPEED), &feedrate_multiplier, 10, 999)
 *     menu_action_setting_edit_int3(PSTR(MSG_SPEED), &feedrate_multiplier, 10, 999)
 *
 */
#define _MENU_ITEM_PART_1(type, label, args...) \
  if (_menuItemNr == _lineNr) { \
    itemSelected = encoderLine == _menuItemNr; \
    if (lcdDrawUpdate) \
       {lcd_implementation_drawmenu_ ## type(itemSelected, _drawLineNr, label, ## args);} \
     if (wasClicked && itemSelected) { \
      lcd_quick_feedback();

#define _MENU_ITEM_PART_2(type, args...) \
      menu_action_ ## type(args); \
      return; \
    } \
  } \
  _menuItemNr++

#define _MENU_ITEM_T_PART_1(type, label, args...) \
  if (_menuItemNr == _lineNr) { \
    itemSelected = encoderLine == _menuItemNr; \
    if (lcdDrawUpdate) \
       {lcd_implementation_drawmenu_ ## type(itemSelected, _drawLineNr, label, ## args);} \
     if (wasClicked && itemSelected) { \
      lcd_quick_feedback();

#define _MENU_ITEM_T_PART_2(type, args...) \
      menu_action_ ## type(args); \
      return; \
    } \
  } \
  _menuItemNr++

#define MENU_ITEM(type, label, args...) do { \
    _MENU_ITEM_PART_1(type, label, ## args); \
    _MENU_ITEM_PART_2(type, ## args); \
  } while(0)

#define MENU_ITEM_T(type, label, args...) do { \
    _MENU_ITEM_T_PART_1(type, label, ## args); \
    _MENU_ITEM_T_PART_2(type, ## args); \
  } while(0)



#ifdef ENCODER_RATE_MULTIPLIER

  //#define ENCODER_RATE_MULTIPLIER_DEBUG  // If defined, output the encoder steps per second value

  /**
   * MENU_MULTIPLIER_ITEM generates drawing and handling code for a multiplier menu item
   */
  #define MENU_MULTIPLIER_ITEM(type, label, args...) do { \
      _MENU_ITEM_PART_1(type, label, ## args); \
      encoderRateMultiplierEnabled = true; \
      lastEncoderMovementMillis = 0; \
      _MENU_ITEM_PART_2(type, ## args); \
    } while(0)

#endif //ENCODER_RATE_MULTIPLIER

#define MENU_ITEM_DUMMY() do { _menuItemNr++; } while(0)
//#define MENU_ITEM_EDIT(type, label, args...) MENU_ITEM(setting_edit_ ## type, label, PSTR(label), ## args)
#define MENU_ITEM_EDIT_T(type, label, args...) MENU_ITEM_T(setting_edit, label, label, ## args)
//#define MENU_ITEM_EDIT_CALLBACK(type, label, args...) MENU_ITEM(setting_edit_callback_ ## type, label, PSTR(label), ## args)
#define MENU_ITEM_EDIT_CALLBACK_T(type, label, args...) MENU_ITEM(setting_edit_callback, label, label, ## args)

#ifdef ENCODER_RATE_MULTIPLIER
  #define MENU_MULTIPLIER_ITEM_EDIT(type, label, args...) MENU_MULTIPLIER_ITEM(setting_edit_ ## type, label, PSTR(label), ## args)
  #define MENU_MULTIPLIER_ITEM_EDIT_CALLBACK(type, label, args...) MENU_MULTIPLIER_ITEM(setting_edit_callback_ ## type, label, PSTR(label), ## args)
#else //!ENCODER_RATE_MULTIPLIER
  #define MENU_MULTIPLIER_ITEM_EDIT(type, label, args...) MENU_ITEM(setting_edit_ ## type, label, PSTR(label), ## args)
  #define MENU_MULTIPLIER_ITEM_EDIT_CALLBACK(type, label, args...) MENU_ITEM(setting_edit_callback_ ## type, label, label, ## args)
#define MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(type, label, args...) MENU_ITEM_T(setting_edit_callback, label, label, ## args)
#endif //!ENCODER_RATE_MULTIPLIER
#define END_MENU() \
    if (encoderLine >= _menuItemNr) { encoderPosition = _menuItemNr * (ENCODER_STEPS_PER_MENU_ITEM) - 1; encoderLine = _menuItemNr - 1; }\
    if (encoderLine >= currentMenuViewOffset + LCD_HEIGHT) { currentMenuViewOffset = encoderLine - (LCD_HEIGHT) + 1; lcdDrawUpdate = LCDVIEW_CALL_REDRAW_NEXT; _lineNr = currentMenuViewOffset - 1; _drawLineNr = -1; } \
    } } while(0)


#endif//ULTIPANEL

menuFunc_t currentMenu = lcd_status_screen; /* function pointer to the currently active menu */
uint8_t lcd_status_update_delay;

//prevMenu and prevEncoderPosition are used to store the previous menu location when editing settings.
menuFunc_t prevMenu = NULL;
uint16_t prevEncoderPosition;
//Variables used when editing values.
const char* editLabel;
void* editValue;
float scale;
int32_t minEditValue, maxEditValue;
menuFunc_t callbackFunc;

enum LCDViewAction {
  LCDVIEW_NONE,
  LCDVIEW_REDRAW_NOW,
  LCDVIEW_CALL_REDRAW_NEXT,
  LCDVIEW_CLEAR_CALL_REDRAW,
  LCDVIEW_CALL_NO_REDRAW
};

// place-holders for Ki and Kd edits
// used for offset menu too
float raw_Ki, raw_Kd, raw_Kp;
int eselect;

/* Main status screen. It's up to the implementation specific part to show what is needed. As this is very display dependent */
static void lcd_status_screen()
{
    if (lcd_status_update_delay)
        lcd_status_update_delay--;
    else
        lcdDrawUpdate = 1;
    if (lcdDrawUpdate)
    {
        lcd_implementation_status_screen();
        lcd_status_update_delay = 10;   /* redraw the main screen every second. This is easier then trying keep track of all things that change on the screen */
    }
#ifdef ULTIPANEL
    if (LCD_CLICKED)
    {
        currentMenu = lcd_main_menu;
        encoderPosition = 0;
        lcd_quick_feedback();
        lcd_implementation_init(); // to maybe revive the LCD if static electricity killed it.
    }

#ifdef ULTIPANEL_FEEDMULTIPLY
    // Dead zone at 100% feedrate
    if ((feedmultiply < 100 && (feedmultiply + int(encoderPosition)) > 100) ||
            (feedmultiply > 100 && (feedmultiply + int(encoderPosition)) < 100))
    {
        encoderPosition = 0;
        feedmultiply = 100;
    }

    if (feedmultiply == 100 && int(encoderPosition) > ENCODER_FEEDRATE_DEADZONE)
    {
        feedmultiply += int(encoderPosition) - ENCODER_FEEDRATE_DEADZONE;
        encoderPosition = 0;
    }
    else if (feedmultiply == 100 && int(encoderPosition) < -ENCODER_FEEDRATE_DEADZONE)
    {
        feedmultiply += int(encoderPosition) + ENCODER_FEEDRATE_DEADZONE;
        encoderPosition = 0;
    }
    else if (feedmultiply != 100)
    {
        feedmultiply += int(encoderPosition);
        encoderPosition = 0;
    }
#endif//ULTIPANEL_FEEDMULTIPLY

    if (feedmultiply < 10)
        feedmultiply = 10;
    if (feedmultiply > 999)
        feedmultiply = 999;
#endif//ULTIPANEL
}


#ifdef ULTIPANEL
static void lcd_return_to_status()
{
    encoderPosition = 0;
    currentMenu = lcd_status_screen;
}

static void lcd_sdcard_pause()
{
    sdcard.pauseSDPrint();
}
static void lcd_sdcard_resume()
{
    sdcard.startFileprint();
}

static void lcd_sdcard_stop()
{
    sdcard.Status = Card_OK;
    sdcard.closefile();
    motor.QuickStop();
    if(SD_FINISHED_STEPPERRELEASE)
    {
    	service.Print(Service_SD_Finished_Release);
//        enquecommand_P(PSTR(SD_FINISHED_RELEASECOMMAND));
    }
    heater.autotempShutdown();
}

typedef int time ;

static time Extrude_time = 0;

//static void set_flow_time()
//{
//	print_job_timer.stop();
//	Extrude_time = print_job_timer.duration();
//}

static void lcd_flow_extrude_start()
{
		my_printer.print_job_timer.reset();
		my_printer.print_job_timer.start();
		lcdDrawUpdate = 2;
		encoderPosition = 0;
		service.Print(Service_Extrude_For_Timer);
		encoderPosition = 3;
		my_printer.Reset_Status_Timeout(60000);
}

static void lcd_flow_extrude_stop()
{
	my_printer.print_job_timer.stop();
		Extrude_time = my_printer.print_job_timer.duration();
		my_printer.print_job_timer.reset();
		lcd_quick_feedback();
		currentMenu = lcd_timer_menu;
		encoderPosition = 2;
		my_printer.Reset_Status_Timeout(60000);
}
/* Menu implementation */
//static void lcd_set_led_white()
//{
//   digitalWrite(REDPIN, 1);
//   digitalWrite(GREENPIN, 1);
//   digitalWrite(BLUEPIN, 1);
//}
//static void lcd_set_led_red()
//{
//   digitalWrite(REDPIN, 1);
//   digitalWrite(GREENPIN, 0);
//   digitalWrite(BLUEPIN, 0);
//}
//static void lcd_set_led_green()
//{
//   digitalWrite(REDPIN, 0);
//   digitalWrite(GREENPIN, 1);
//   digitalWrite(BLUEPIN, 0);
//}
//static void lcd_set_led_blue()
//{
//   digitalWrite(REDPIN, 0);
//   digitalWrite(GREENPIN, 0);
//   digitalWrite(BLUEPIN, 1);
//}
//static void lcd_set_led_lightblue()
//{
//   digitalWrite(REDPIN, 0);
//   digitalWrite(GREENPIN, 1);
//   digitalWrite(BLUEPIN, 1);
//}
//static void lcd_set_led_purple()
//{
//   digitalWrite(REDPIN, 1);
//   digitalWrite(GREENPIN, 0);
//   digitalWrite(BLUEPIN, 1);
//}
//static void lcd_set_led_yellow()
//{
//   digitalWrite(REDPIN, 1);
//   digitalWrite(GREENPIN, 1);
//   digitalWrite(BLUEPIN, 0);
//}
//static void lcd_set_led_off()
//{
//   digitalWrite(REDPIN, 0);
//   digitalWrite(GREENPIN, 0);
//   digitalWrite(BLUEPIN, 0);
//}
//
//static void lcd_led_menu()
//{
//  START_MENU();
//    MENU_ITEM(back, MSG_MAIN, lcd_main_menu);
//    MENU_ITEM(function, MSG_LED_WHITE, lcd_set_led_white);
//    MENU_ITEM(function, MSG_LED_RED, lcd_set_led_red);
//    MENU_ITEM(function, MSG_LED_GREEN, lcd_set_led_green);
//    MENU_ITEM(function, MSG_LED_BLUE, lcd_set_led_blue);
//    MENU_ITEM(function, MSG_LED_LIGHTBLUE, lcd_set_led_lightblue);
//    MENU_ITEM(function, MSG_LED_PURPLE, lcd_set_led_purple);
//    MENU_ITEM(function, MSG_LED_YELLOW, lcd_set_led_yellow);
//    MENU_ITEM(function, MSG_LED_OFF, lcd_set_led_off);
//  END_MENU();
//}

static void lcd_main_menu()
{
	START_MENU();
	MENU_ITEM(back, T(MSG_Info), lcd_status_screen);

	if (plan.Moves_Planned() || IS_SD_PRINTING)
	{
		MENU_ITEM(submenu, T(MSG_Tune), lcd_tune_menu);
	}else{
		MENU_ITEM(submenu, T(MSG_Prepare), lcd_prepare_menu);
	}
	MENU_ITEM(submenu, T(MSG_Control), lcd_control_menu);
#ifdef SDSUPPORT
if (sdcard.Status != No_Card)
{
	if (sdcard.isFileOpen())
	{
		if (sdcard.Status == Card_printing)
			MENU_ITEM(function, T(MSG_Pause_Print), lcd_sdcard_pause);
		else
			MENU_ITEM(function, T(MSG_Resume_Print), lcd_sdcard_resume);
		MENU_ITEM(function, T(MSG_Stop_Print), lcd_sdcard_stop);
	}else{
		MENU_ITEM(submenu, T(MSG_Card_Menu), lcd_sdcard_menu);
#if SDCARDDETECT < 1
MENU_ITEM(service, T(MSG_Change_SDcard), Service_Init_SDcard);  // SD-card changed by user
#endif
	}
}else{
	MENU_ITEM(submenu, T(MSG_No_Card), lcd_sdcard_menu);
#if SDCARDDETECT < 1
	MENU_ITEM(service, T(MSG_Init_SDcard), Service_Init_SDcard); // Manually initialize the SD-card via user interface
#endif
}
#endif

if (plan.Moves_Planned() || IS_SD_PRINTING)
{    }
else{
	MENU_ITEM(submenu, T(MSG_Firmware), lcd_firmware_menu);
}

END_MENU();
}

#ifdef SDSUPPORT
static void lcd_autostart_sd()
{
    sdcard.lastnr=0;
    sdcard.setroot();
    sdcard.checkautostart(true);
}
#endif

#ifdef BABYSTEPPING
static void lcd_babystep_x()
{
    if (encoderPosition != 0)
    {
        babystepsTodo[X_AXIS]+=(int)encoderPosition;
        encoderPosition=0;
        lcdDrawUpdate = 1;
    }
    if (lcdDrawUpdate)
    {
        lcd_implementation_drawedit(PSTR(MSG_BABYSTEPPING_X),"");
    }
    if (LCD_CLICKED)
    {
        lcd_quick_feedback();
        currentMenu = lcd_tune_menu;
        encoderPosition = 0;
    }
}

static void lcd_babystep_y()
{
    if (encoderPosition != 0)
    {
        babystepsTodo[Y_AXIS]+=(int)encoderPosition;
        encoderPosition=0;
        lcdDrawUpdate = 1;
    }
    if (lcdDrawUpdate)
    {
        lcd_implementation_drawedit(PSTR(MSG_BABYSTEPPING_Y),"");
    }
    if (LCD_CLICKED)
    {
        lcd_quick_feedback();
        currentMenu = lcd_tune_menu;
        encoderPosition = 0;
    }
}

static void lcd_babystep_z()
{
    if (encoderPosition != 0)
    {
        babystepsTodo[Z_AXIS]+=BABYSTEP_Z_MULTIPLICATOR*(int)encoderPosition;
        encoderPosition=0;
        lcdDrawUpdate = 1;
    }
    if (lcdDrawUpdate)
    {
        lcd_implementation_drawedit(PSTR(MSG_BABYSTEPPING_Z),"");
    }
    if (LCD_CLICKED)
    {
        lcd_quick_feedback();
        currentMenu = lcd_tune_menu;
        encoderPosition = 0;
    }
}
#endif //BABYSTEPPING

/**
 * Watch temperature callbacks
 */
#if defined(THERMAL_PROTECTION_HOTENDS)
  #if TEMP_SENSOR_0 != 0
    void watch_temp_callback_E0() { start_watching_heater(0); }
  #endif
  #if EXTRUDERS > 1 && TEMP_SENSOR_1 != 0
    void watch_temp_callback_E1() { start_watching_heater(1); }
  #endif // EXTRUDERS > 1
  #if EXTRUDERS > 2 && TEMP_SENSOR_2 != 0
    void watch_temp_callback_E2() { start_watching_heater(2); }
  #endif // EXTRUDERS > 2
  #if EXTRUDERS > 3 && TEMP_SENSOR_3 != 0
    void watch_temp_callback_E3() { start_watching_heater(3); }
  #endif // EXTRUDERS > 3
#else
  #if TEMP_SENSOR_0 != 0
    void watch_temp_callback_E0() {}
  #endif
  #if EXTRUDERS > 1 && TEMP_SENSOR_1 != 0
    void watch_temp_callback_E1() {}
  #endif // EXTRUDERS > 1
  #if EXTRUDERS > 2 && TEMP_SENSOR_2 != 0
    void watch_temp_callback_E2() {}
  #endif // EXTRUDERS > 2
  #if EXTRUDERS > 3 && TEMP_SENSOR_3 != 0
    void watch_temp_callback_E3() {}
  #endif // EXTRUDERS > 3
#endif


static void lcd_tune_menu()
{
    START_MENU();
    MENU_ITEM(back, T(MSG_Main_Menu), lcd_main_menu);
//#ifdef LEDCONTROL
//    MENU_ITEM(submenu, MSG_LED_MENU, lcd_led_menu);
//#endif
    MENU_ITEM_EDIT_T(int, T(MSG_Speed), &feedmultiply, 10, 999,1);
#if EXTRUDERS == 1
#if TEMP_SENSOR_0 != 0
        	MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(int3, T(MSG_Nozzle1),  &heater.target_temperature[0], 0, HEATER_0_MAXTEMP - 15, 1, watch_temp_callback_E0);
#endif
#else //EXTRUDERS > 1
#if TEMP_SENSOR_0 != 0
        	MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(int, T(MSG_Nozzle1), &heater.target_temperature[0], 0, HEATER_0_MAXTEMP - 15, 1 , watch_temp_callback_E0);
#endif
#if TEMP_SENSOR_1 != 0
        	MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(int, T(MSG_Nozzle2), &heater.target_temperature[1], 0, HEATER_1_MAXTEMP - 15, 1, watch_temp_callback_E1);
#endif
#endif // EXTRUDERS > 1
#if TEMP_SENSOR_BED != 0
    MENU_ITEM_EDIT_T(int, T(MSG_Bed), &heater.target_temperature_bed, 0, BED_MAXTEMP - 15,1);
#endif
    MENU_ITEM_EDIT_T(int, T(MSG_FanSpeed), &heater.fanSpeed, 0, 255,1.0);

    MENU_ITEM_EDIT_T(int, T(MSG_Flow), &extrudemultiply, 10, 999 ,1); // ToDo : reduzieren
    MENU_ITEM_EDIT_T(int, T(MSG_Flow1), &extruder_multiply[0], 10, 999 ,1);
#if TEMP_SENSOR_1 != 0
    MENU_ITEM_EDIT_T(int, T(MSG_Flow2), &extruder_multiply[1], 10, 999, 1);
#endif
//#if TEMP_SENSOR_2 != 0
//    MENU_ITEM_EDIT(int3, MSG_FLOW2, &extruder_multiply[2], 10, 999);
//#endif

#ifdef BABYSTEPPING
    #ifdef BABYSTEP_XY
      MENU_ITEM(submenu, MSG_BABYSTEP_X, lcd_babystep_x);
      MENU_ITEM(submenu, MSG_BABYSTEP_Y, lcd_babystep_y);
    #endif //BABYSTEP_XY
    MENU_ITEM(submenu, MSG_BABYSTEP_Z, lcd_babystep_z);
#endif
#ifdef FILAMENTCHANGEENABLE
    MENU_ITEM(service, T(MSG_Filamentchange), Service_Filamentchange);

//     MENU_ITEM(gcode, MSG_FILAMENTCHANGE, PSTR("M600"));
#endif
    END_MENU();
}

void lcd_preheat_pla0()
{
    heater.setTargetHeater(config.plaPreheatHotendTemp,0);
    heater.setTargetHeater(config.plaPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.plaPreheatFanSpeed;
    lcd_return_to_status();
//    setWatch(); // heater sanity check timer
}

void lcd_preheat_abs0()
{
    heater.setTargetHeater(config.absPreheatHotendTemp,0);
    heater.setTargetHeater(config.absPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.absPreheatFanSpeed;
    lcd_return_to_status();
 //   setWatch(); // heater sanity check timer
}

#if TEMP_SENSOR_1 >= 0 //2nd extruder preheat
void lcd_preheat_pla1()
{
    heater.setTargetHeater(config.plaPreheatHotendTemp,1);
    heater.setTargetHeater(config.plaPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.plaPreheatFanSpeed;
    lcd_return_to_status();
//    setWatch(); // heater sanity check timer
}

void lcd_preheat_abs1()
{
    heater.setTargetHeater(config.absPreheatHotendTemp,1);
    heater.setTargetHeater(config.absPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.absPreheatFanSpeed;
    lcd_return_to_status();
//    setWatch(); // heater sanity check timer
}
#endif //2nd extruder preheat

#if TEMP_SENSOR_2 >= 0 //3 extruder preheat
void lcd_preheat_pla2()
{
    heater.setTargetHeater(config.plaPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.plaPreheatFanSpeed;
    lcd_return_to_status();
//    setWatch(); // heater sanity check timer
}

void lcd_preheat_abs2()
{
    heater.setTargetHeater(config.absPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.absPreheatFanSpeed;
    lcd_return_to_status();
//    setWatch(); // heater sanity check timer
}
#endif //3 extruder preheat

#if TEMP_SENSOR_1 >= 0 || TEMP_SENSOR_2 != 0 //more than one extruder present
void lcd_preheat_pla012()
{
    heater.setTargetHeater(config.plaPreheatHotendTemp,0);
    heater.setTargetHeater(config.plaPreheatHotendTemp,1);
   // setTargetHeater2(config.plaPreheatHotendTemp);
    heater.setTargetHeater(config.plaPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.plaPreheatFanSpeed;
    lcd_return_to_status();
//    setWatch(); // heater sanity check timer
}

void lcd_preheat_abs012()
{
    heater.setTargetHeater(config.absPreheatHotendTemp,0);
    heater.setTargetHeater(config.absPreheatHotendTemp,1);
  //  setTargetHeater2(config.absPreheatHotendTemp);
    heater.setTargetHeater(config.absPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.absPreheatFanSpeed;
    lcd_return_to_status();
//    setWatch(); // heater sanity check timer
}
#endif //more than one extruder present

void lcd_preheat_pla_bedonly()
{
	heater.setTargetHeater(config.plaPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.plaPreheatFanSpeed;
    lcd_return_to_status();
//    setWatch(); // heater sanity check timer
}

void lcd_preheat_abs_bedonly()
{
	heater.setTargetHeater(config.absPreheatHPBTemp,HEAT_BED);
    heater.fanSpeed = config.absPreheatFanSpeed;
    lcd_return_to_status();
//    setWatch(); // heater sanity check timer
}

static void lcd_preheat_pla_menu()
{
    START_MENU();
    MENU_ITEM(back,  T(MSG_Prepare), lcd_prepare_menu);
    MENU_ITEM(function, T(MSG_Preheat_PLA_EX1), lcd_preheat_pla0);
#if TEMP_SENSOR_1 >= 0 //2 extruder preheat
    MENU_ITEM(function, T(MSG_Preheat_PLA_EX2), lcd_preheat_pla1);
#endif //2 extruder preheat
//#if TEMP_SENSOR_2 >= 0 //3 extruder preheat
//    MENU_ITEM(function, MSG_PREHEAT_PLA2, lcd_preheat_pla2);
//#endif //3 extruder preheat
#if TEMP_SENSOR_1 >= 0 //|| TEMP_SENSOR_2 >= 0 //all extruder preheat
    MENU_ITEM(function, T(MSG_Preheat_PLA_EXALL), lcd_preheat_pla012);
#endif //2 extruder preheat
#if TEMP_SENSOR_BED != 0
    MENU_ITEM(function, MSG_PREHEAT_PLA_BEDONLY, lcd_preheat_pla_bedonly);
#endif
    END_MENU();
}

static void lcd_preheat_abs_menu()
{
    START_MENU();
    MENU_ITEM(back,  T(MSG_Prepare), lcd_prepare_menu);
    MENU_ITEM(function, T(MSG_Preheat_ABS_EX1), lcd_preheat_abs0);
#if TEMP_SENSOR_1 >= 0 //2 extruder preheat
    MENU_ITEM(function, T(MSG_Preheat_ABS_EX2), lcd_preheat_abs1);
#endif //2 extruder preheat
//#if TEMP_SENSOR_2 >= 0 //3 extruder preheat
//    MENU_ITEM(function, MSG_PREHEAT_ABS2, lcd_preheat_abs2);
//#endif //3 extruder preheat
#if TEMP_SENSOR_1 >= 0 //|| TEMP_SENSOR_2 != 0 //all extruder preheat
    MENU_ITEM(function, T(MSG_Preheat_ABS_EXALL), lcd_preheat_abs012);
#endif //2 extruder preheat
#if TEMP_SENSOR_BED != 0
    MENU_ITEM(function, MSG_PREHEAT_ABS_BEDONLY, lcd_preheat_abs_bedonly);
#endif
    END_MENU();
}

void lcd_cooldown()
{
    heater.setTargetHeater(0,0);
    heater.setTargetHeater(0,1);
    heater.setTargetHeater(0,HEAT_BED);
    heater.fanSpeed = 0;
    lcd_return_to_status();
}

void Save_Extruder_Offsets()
{
#if EXTRUDERS > 1
	  // raw_ki = x-offset, raw_kd = y-offset
	  if (raw_Ki >= 0)
	  {
		  config.extruder_offset[X_AXIS][0] =raw_Ki;
		  config.extruder_offset[X_AXIS][1] =0;
	  }
	  else
	  {	// should never occur
		  config.extruder_offset[X_AXIS][0] =0;
		  config.extruder_offset[X_AXIS][1] =raw_Ki;
	  }
	  if (raw_Kd >= 0)
	  {
		  config.extruder_offset[Y_AXIS][0] =raw_Kd;
		  config.extruder_offset[Y_AXIS][1] =0;
	  }
	  else
	  {	// should never occur
		  config.extruder_offset[Y_AXIS][0] =0;
		  config.extruder_offset[Y_AXIS][1] =raw_Kd;
	  }
#endif
}

#if EXTRUDERS > 1
static void lcd_extruder_offset_menu()
{
	// raw_Ki and raw_Kd are re-used for offset menu
	raw_Ki = config.extruder_offset[X_AXIS][0] - config.extruder_offset[X_AXIS][1] ;
	raw_Kd = config.extruder_offset[Y_AXIS][0] - config.extruder_offset[Y_AXIS][1] ;
    START_MENU();
    MENU_ITEM(back,  T(MSG_Prepare), lcd_prepare_menu);
	MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(float, T(MSG_Extruder_offset_X), &raw_Ki, 20, 35, 20 , Save_Extruder_Offsets);
	MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(float, T(MSG_Extruder_offset_Y), &raw_Kd, -10, 10, 20 , Save_Extruder_Offsets);
    END_MENU();
}
#endif

static void lcd_prepare_menu()
{
    START_MENU();
    MENU_ITEM(back, T(MSG_Main_Menu), lcd_main_menu);
    //MENU_ITEM(gcode, T(MSG_Auto_Home), PSTR("G28"));
    MENU_ITEM(service, T(MSG_Auto_Home), Service_Home);
//#ifdef LEDCONTROL
//    MENU_ITEM(submenu, MSG_LED_MENU, lcd_led_menu);
//#endif
    MENU_ITEM(service, T(MSG_Disable_Steppers),Service_Disable_Steppers);
    MENU_ITEM(function, T(MSG_Cooldown), lcd_cooldown);
    MENU_ITEM(submenu, T(MSG_Load_Filament), lcd_load_menu);
    MENU_ITEM(submenu, T(MSG_Unload_Filament), lcd_unload_menu);
#if TEMP_SENSOR_0 != 0
  #if TEMP_SENSOR_1 != 0 || TEMP_SENSOR_2 != 0 || TEMP_SENSOR_BED != 0
    MENU_ITEM(submenu, T(MSG_Preheat_PLA), lcd_preheat_pla_menu);
    MENU_ITEM(submenu, T(MSG_Preheat_ABS), lcd_preheat_abs_menu);
  #else
    MENU_ITEM(function, MSG_PREHEAT_PLA, lcd_preheat_pla0);
    MENU_ITEM(function, MSG_PREHEAT_ABS, lcd_preheat_abs0);
  #endif
#endif
MENU_ITEM(submenu, T(MSG_Move_Axis), lcd_move_menu);
#ifdef SDSUPPORT
    #ifdef MENU_ADDAUTOSTART
      MENU_ITEM(function, MSG_AUTOSTART, lcd_autostart_sd);
    #endif
#endif
#if EXTRUDERS > 1
      MENU_ITEM(submenu, T(MSG_Extruder_offsets), lcd_extruder_offset_menu);
#endif
#if PS_ON_PIN > -1
    if (powersupply)
    {
        MENU_ITEM(gcode, MSG_SWITCH_PS_OFF, PSTR("M81"));
    }else{
        MENU_ITEM(gcode, MSG_SWITCH_PS_ON, PSTR("M80"));
    }
#endif
    END_MENU();
}

float move_menu_scale;
static void lcd_move_menu_axis();

static void lcd_move_x()
{
    if (encoderPosition != 0)
    {
        refresh_cmd_timeout();
        Position.current_position_mm[X_AXIS] += float((int)encoderPosition) * move_menu_scale;
        if (min_software_endstops && Position.current_position_mm[X_AXIS] < X_MIN_POS)
        	Position.current_position_mm[X_AXIS] = X_MIN_POS;
        if (max_software_endstops && Position.current_position_mm[X_AXIS] > X_MAX_POS)
        	Position.current_position_mm[X_AXIS] = X_MAX_POS;
        encoderPosition = 0;
        #ifdef DELTA
        calculate_delta(current_position_mm);
        plan_buffer_line(delta[X_AXIS], delta[Y_AXIS], delta[Z_AXIS], current_position_mm[E_AXIS], manual_feedrate[X_AXIS]/60, active_extruder);
        #else
        plan.Buffer_Line(Position.current_position_mm[X_AXIS], Position.current_position_mm[Y_AXIS], Position.current_position_mm[Z_AXIS], Position.current_position_mm[E_AXIS], manual_feedrate[X_AXIS]/60, active_extruder);
        #endif
        lcdDrawUpdate = 1;
    }
    if (lcdDrawUpdate)
    {
        lcd_implementation_drawedit(PSTR("X"), ftostr31(Position.current_position_mm[X_AXIS]));
    }
    if (LCD_CLICKED)
    {
        lcd_quick_feedback();
        currentMenu = lcd_move_menu_axis;
        encoderPosition = 0;
    }
}
static void lcd_move_y()
{
    if (encoderPosition != 0)
    {
        refresh_cmd_timeout();
        Position.current_position_mm[Y_AXIS] += float((int)encoderPosition) * move_menu_scale;
        if (min_software_endstops && Position.current_position_mm[Y_AXIS] < Y_MIN_POS)
        	Position.current_position_mm[Y_AXIS] = Y_MIN_POS;
        if (max_software_endstops && Position.current_position_mm[Y_AXIS] > Y_MAX_POS)
        	Position.current_position_mm[Y_AXIS] = Y_MAX_POS;
        encoderPosition = 0;
        #ifdef DELTA
        calculate_delta(current_position_mm);
        plan_buffer_line(delta[X_AXIS], delta[Y_AXIS], delta[Z_AXIS], current_position_mm[E_AXIS], manual_feedrate[Y_AXIS]/60, active_extruder);
        #else
        plan.Buffer_Line(Position.current_position_mm[X_AXIS], Position.current_position_mm[Y_AXIS], Position.current_position_mm[Z_AXIS], Position.current_position_mm[E_AXIS], manual_feedrate[Y_AXIS]/60, active_extruder);
        #endif
        lcdDrawUpdate = 1;
    }
    if (lcdDrawUpdate)
    {
        lcd_implementation_drawedit(PSTR("Y"), ftostr31(Position.current_position_mm[Y_AXIS]));
    }
    if (LCD_CLICKED)
    {
        lcd_quick_feedback();
        currentMenu = lcd_move_menu_axis;
        encoderPosition = 0;
    }
}
static void lcd_move_z()
{
    if (encoderPosition != 0)
    {
        refresh_cmd_timeout();
        Position.current_position_mm[Z_AXIS] += float((int)encoderPosition) * move_menu_scale;
        if (min_software_endstops && Position.current_position_mm[Z_AXIS] < Z_MIN_POS)
        	Position.current_position_mm[Z_AXIS] = Z_MIN_POS;
        if (max_software_endstops && Position.current_position_mm[Z_AXIS] > Z_MAX_POS)
        	Position.current_position_mm[Z_AXIS] = Z_MAX_POS;
        encoderPosition = 0;
        #ifdef DELTA
        calculate_delta(current_position_mm);
        plan_buffer_line(delta[X_AXIS], delta[Y_AXIS], delta[Z_AXIS], current_position_mm[E_AXIS], manual_feedrate[Z_AXIS]/60, active_extruder);
        #else
        plan.Buffer_Line(Position.current_position_mm[X_AXIS], Position.current_position_mm[Y_AXIS], Position.current_position_mm[Z_AXIS], Position.current_position_mm[E_AXIS], manual_feedrate[Z_AXIS]/60, active_extruder);
        #endif
        lcdDrawUpdate = 1;
    }
    if (lcdDrawUpdate)
    {
        lcd_implementation_drawedit(PSTR("Z"), ftostr31(Position.current_position_mm[Z_AXIS]));
    }
    if (LCD_CLICKED)
    {
        lcd_quick_feedback();
        currentMenu = lcd_move_menu_axis;
        encoderPosition = 0;
    }
}
static void lcd_move_e()
{
    if (encoderPosition != 0)
    {
    	Position.current_position_mm[E_AXIS] += float((int)encoderPosition) * move_menu_scale;
        encoderPosition = 0;
        #ifdef DELTA
        calculate_delta(current_position_mm);
        plan_buffer_line(delta[X_AXIS], delta[Y_AXIS], delta[Z_AXIS], current_position_mm[E_AXIS], manual_feedrate[E_AXIS]/60, active_extruder);
        #else
        plan.Buffer_Line(Position.current_position_mm[X_AXIS], Position.current_position_mm[Y_AXIS], Position.current_position_mm[Z_AXIS], Position.current_position_mm[E_AXIS], manual_feedrate[E_AXIS]/60, active_extruder);
        #endif
        lcdDrawUpdate = 1;
    }
    if (lcdDrawUpdate)
    {
        lcd_implementation_drawedit(PSTR("Extruder"), ftostr31(Position.current_position_mm[E_AXIS]));
    }
    if (LCD_CLICKED)
    {
        lcd_quick_feedback();
        currentMenu = lcd_move_menu_axis;
        encoderPosition = 0;
    }
}

static void lcd_move_menu_axis()
{
    START_MENU();
    MENU_ITEM(back, T(MSG_Move_Axis), lcd_move_menu);
    MENU_ITEM(submenu, T(MSG_Move_X), lcd_move_x);
    MENU_ITEM(submenu, T(MSG_Move_Y), lcd_move_y);
    if (move_menu_scale < 10.0)
    {
        MENU_ITEM(submenu, T(MSG_Move_Z), lcd_move_z);
        MENU_ITEM(submenu, T(MSG_Move_E), lcd_move_e);
        //MENU_ITEM(submenu, MSG_MOVE_E1, lcd_move_e1);
    }
    END_MENU();
}

static void lcd_move_menu_10mm()
{
    move_menu_scale = 10.0;
    lcd_move_menu_axis();
}
static void lcd_move_menu_1mm()
{
    move_menu_scale = 1.0;
    lcd_move_menu_axis();
}
static void lcd_move_menu_01mm()
{
    move_menu_scale = 0.1;
    lcd_move_menu_axis();
}

static void lcd_move_menu()
{
    START_MENU();  
    MENU_ITEM(back,  T(MSG_Prepare), lcd_prepare_menu);
    MENU_ITEM(submenu, T(MSG_Move_10mm), lcd_move_menu_10mm);
    MENU_ITEM(submenu, T(MSG_Move_1mm), lcd_move_menu_1mm);
    MENU_ITEM(submenu, T(MSG_Move_01mm), lcd_move_menu_01mm);
    //TODO:X,Y,Z,E
    END_MENU();
}
static void lcd_load_menu_EXT1_ABS_go()
{
	heater.setTargetHeater(config.absPreheatHotendTemp,0);
	heater.Set_FanSpeed(config.absPreheatFanSpeed);
	service.Print(Service_Load_ABS_EX1);
  lcd_return_to_status();
}

static void lcd_load_menu_EXT1_ABS()
{
  START_MENU();
  MENU_ITEM(function, T(MSG_Load_TEXT0), lcd_load_menu_EXT1_ABS_go);
  MENU_ITEM(function, T(MSG_Load_TEXT1), lcd_load_menu_EXT1_ABS_go);
  MENU_ITEM(function, T(MSG_Load_TEXT2), lcd_load_menu_EXT1_ABS_go);
  MENU_ITEM(function, T(MSG_Load_TEXT3), lcd_load_menu_EXT1_ABS_go);
  END_MENU();
}

static void lcd_load_menu_EXT1_PLA_go()
{
	heater.setTargetHeater(config.plaPreheatHotendTemp,0);
	heater.Set_FanSpeed(config.plaPreheatFanSpeed);
	service.Print(Service_Load_PLA_EX1);
  lcd_return_to_status();
}


static void lcd_load_menu_EXT1_PLA()
{
  START_MENU();
  MENU_ITEM(function, T(MSG_Load_TEXT0), lcd_load_menu_EXT1_PLA_go);
  MENU_ITEM(function, T(MSG_Load_TEXT1), lcd_load_menu_EXT1_PLA_go);
  MENU_ITEM(function, T(MSG_Load_TEXT2), lcd_load_menu_EXT1_PLA_go);
  MENU_ITEM(function, T(MSG_Load_TEXT3), lcd_load_menu_EXT1_PLA_go);
  END_MENU();
}
static void lcd_load_menu_EXT2_ABS_go()
{
	heater.setTargetHeater(config.absPreheatHotendTemp,1);
	heater.Set_FanSpeed(config.absPreheatFanSpeed);
	service.Print(Service_Load_ABS_EX2);
  lcd_return_to_status();
}


static void lcd_load_menu_EXT2_ABS()
{
  START_MENU();
  MENU_ITEM(function, T(MSG_Load_TEXT0), lcd_load_menu_EXT2_ABS_go);
  MENU_ITEM(function, T(MSG_Load_TEXT1), lcd_load_menu_EXT2_ABS_go);
  MENU_ITEM(function, T(MSG_Load_TEXT2), lcd_load_menu_EXT2_ABS_go);
  MENU_ITEM(function, T(MSG_Load_TEXT3), lcd_load_menu_EXT2_ABS_go);
  END_MENU();
}
static void lcd_load_menu_EXT2_PLA_go()
{
	heater.setTargetHeater(config.plaPreheatHotendTemp,1);
	heater.Set_FanSpeed(config.plaPreheatFanSpeed);
	service.Print(Service_Load_PLA_EX2);
  lcd_return_to_status();
}


static void lcd_load_menu_EXT2_PLA()
{
  START_MENU();
  MENU_ITEM(function, T(MSG_Load_TEXT0), lcd_load_menu_EXT2_PLA_go);
  MENU_ITEM(function, T(MSG_Load_TEXT1), lcd_load_menu_EXT2_PLA_go);
  MENU_ITEM(function, T(MSG_Load_TEXT2), lcd_load_menu_EXT2_PLA_go);
  MENU_ITEM(function, T(MSG_Load_TEXT3), lcd_load_menu_EXT2_PLA_go);
  END_MENU();
}

static void lcd_unload_menu_EXT1_ABS()
{
	heater.setTargetHeater(config.absPreheatHotendTemp,0);
	heater.Set_FanSpeed(config.absPreheatFanSpeed);
	service.Print(Service_Unload_ABS_EX1);
  lcd_return_to_status();
}


static void lcd_unload_menu_EXT1_PLA()
{
	heater.setTargetHeater(config.plaPreheatHotendTemp,0);
	heater.Set_FanSpeed(config.plaPreheatFanSpeed);
	service.Print(Service_Unload_PLA_EX1);
  lcd_return_to_status();
}


static void lcd_unload_menu_EXT2_ABS()
{
	heater.setTargetHeater(config.absPreheatHotendTemp,1);
	heater.Set_FanSpeed(config.absPreheatFanSpeed);
	service.Print(Service_Unload_ABS_EX2);
  lcd_return_to_status();
}


static void lcd_unload_menu_EXT2_PLA()
{
	heater.setTargetHeater(config.plaPreheatHotendTemp,1);
	heater.Set_FanSpeed(config.plaPreheatFanSpeed);
	service.Print(Service_Unload_PLA_EX2);
  lcd_return_to_status();

}


static void lcd_load_menu()
{
    START_MENU();
      MENU_ITEM(back,  T(MSG_Prepare), lcd_prepare_menu);
      MENU_ITEM(submenu, T(MSG_Load_ABS_EX1), lcd_load_menu_EXT1_ABS);
      MENU_ITEM(submenu, T(MSG_Load_PLA_EX1), lcd_load_menu_EXT1_PLA);
#if EXTRUDERS > 1
      MENU_ITEM(submenu, T(MSG_Load_ABS_EX2), lcd_load_menu_EXT2_ABS);
      MENU_ITEM(submenu, T(MSG_Load_PLA_EX2), lcd_load_menu_EXT2_PLA);
#endif
    END_MENU();
}

static void lcd_unload_menu()
{
    START_MENU(); 
      MENU_ITEM(back,  T(MSG_Prepare), lcd_prepare_menu);
      MENU_ITEM(function, T(MSG_Unload_ABS_EX1), lcd_unload_menu_EXT1_ABS);
      MENU_ITEM(function, T(MSG_Unload_PLA_EX1), lcd_unload_menu_EXT1_PLA);
  #if EXTRUDERS > 1
      MENU_ITEM(function, T(MSG_Unload_ABS_EX2), lcd_unload_menu_EXT2_ABS);
      MENU_ITEM(function, T(MSG_Unload_PLA_EX2), lcd_unload_menu_EXT2_PLA);
  #endif
    END_MENU();
}

static void load_EEPROM()
{
	config.Load_EEPROM();
}

static void store_EEPROM()
{
	config.Store_EEPROM();
}

static void load_failsave()
{
	config.Load_Failsave();
}

static void lcd_control_menu()
{
    START_MENU();
    MENU_ITEM(back, T(MSG_Main_Menu), lcd_main_menu);
    MENU_ITEM(submenu, T(MSG_Temperature), lcd_control_temperature_menu);
    MENU_ITEM(submenu, T(MSG_Motion), lcd_control_motion_menu);
#ifdef DOGLCD
//    MENU_ITEM_EDIT(int3, MSG_CONTRAST, &lcd_contrast, 0, 63);
    MENU_ITEM(submenu, MSG_CONTRAST, lcd_set_contrast);
#endif
#ifdef FWRETRACT
    MENU_ITEM(submenu, MSG_RETRACT, lcd_control_retract_menu);
#endif
    MENU_ITEM(submenu, T(MSG_Advanced), lcd_control_advanced_menu);
#ifdef EEPROM_SETTINGS
    MENU_ITEM(function, T(MSG_Store_Eprom), store_EEPROM);
    MENU_ITEM(function, T(MSG_Load_Eprom), load_EEPROM);
#endif
    MENU_ITEM(function, T(MSG_Restore_Failsave), load_failsave);
    END_MENU();
}



#ifdef PIDTEMP

  // Helpers for editing PID Ki & Kd values
  // grab the PID value out of the temp variable; scale it; then update the PID driver
//  void copy_and_scalePID_i(int e) {
//    #ifndef PID_PARAMS_PER_EXTRUDER
//      UNUSED(e);
//    #endif
//    config.Set_Ki(raw_Ki,e);
//    heater.UpdatePID();
//  }
//  void copy_and_scalePID_d(int e) {
//    #ifndef PID_PARAMS_PER_EXTRUDER
//      UNUSED(e);
//    #endif
//    config.Set_Kd(raw_Kd,e);
//    heater.UpdatePID();
//  }


	void copy_and_scalePID_p_E0() { config.Set_Kp(raw_Kp,0); }
  void copy_and_scalePID_i_E0() { config.Set_Ki(raw_Ki,0); }
  void copy_and_scalePID_d_E0() { config.Set_Kd(raw_Kd,0); }
	void copy_and_scalePID_p_E1() { config.Set_Kp(raw_Kp,1); }
  void copy_and_scalePID_i_E1() { config.Set_Ki(raw_Ki,1); }
  void copy_and_scalePID_d_E1() { config.Set_Kd(raw_Kd,1); }
	void copy_and_scalePID_p_E99() { config.Set_Kp(raw_Kp,-1); }
void copy_and_scalePID_i_E99() { config.Set_Ki(raw_Ki,-1); }
void copy_and_scalePID_d_E99() { config.Set_Kd(raw_Kd,-1); }
//#ifdef PID_AUTOTUNE_MENU
//  #define _PIDTEMP_FUNCTIONS(eindex) \
//    _PIDTEMP_BASE_FUNCTIONS(eindex); \
//    void lcd_autotune_callback_E ## eindex() { _lcd_autotune(eindex); }
//#else
//  #define _PIDTEMP_FUNCTIONS(eindex) _PIDTEMP_BASE_FUNCTIONS(eindex)
//#endif
//
//_PIDTEMP_FUNCTIONS(0);
//#ifdef PID_PARAMS_PER_EXTRUDER
//  #if EXTRUDERS > 1
//    _PIDTEMP_FUNCTIONS(1);
//    #if EXTRUDERS > 2
//      _PIDTEMP_FUNCTIONS(2);
//      #if EXTRUDERS > 3
//        _PIDTEMP_FUNCTIONS(3);
//      #endif //EXTRUDERS > 3
//    #endif //EXTRUDERS > 2
//  #endif //EXTRUDERS > 1
//#endif //PID_PARAMS_PER_EXTRUDER
#endif //PIDTEMP

        static void lcd_control_temperature_menu()
        {
        	START_MENU();
        	MENU_ITEM(back, T(MSG_Control), lcd_control_menu);

        	//
        	// Nozzle:
        	// Nozzle [1-4]:
        	//
#if EXTRUDERS == 1
#if TEMP_SENSOR_0 != 0
        	MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(int, T(MSG_Nozzle1), &heater.target_temperature[0], 0, HEATER_0_MAXTEMP - 15, 1, watch_temp_callback_E0);
#endif
#else //EXTRUDERS > 1
#if TEMP_SENSOR_0 != 0
        	MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(int, T(MSG_Nozzle1), &heater.target_temperature[0], 0, HEATER_0_MAXTEMP - 15, 1, watch_temp_callback_E0);
#endif
#if TEMP_SENSOR_1 != 0
        	MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(int, T(MSG_Nozzle2), &heater.target_temperature[1], 0, HEATER_1_MAXTEMP - 15, 1, watch_temp_callback_E1);
#endif
#endif // EXTRUDERS > 1

#if TEMP_SENSOR_BED != 0
        	MENU_ITEM_EDIT_T(int, T(MSG_Bed), &heater.target_temperature_bed, 0, BED_MAXTEMP - 15,1);
#endif
        	//
        	// Fan Speed:
        	//
        	MENU_ITEM_EDIT_T(int, T(MSG_FanSpeed), &heater.fanSpeed, 0, 255,1.0);

#ifdef AUTOTEMP
        	MENU_ITEM_EDIT(bool, MSG_AUTOTEMP, &autotemp_enabled);
        	MENU_ITEM_EDIT(float3, MSG_MIN, &autotemp_min, 0, HEATER_0_MAXTEMP - 15);
        	MENU_ITEM_EDIT(float3, MSG_MAX, &autotemp_max, 0, HEATER_0_MAXTEMP - 15);
        	MENU_ITEM_EDIT(float32, MSG_FACTOR, &autotemp_factor, 0.0, 1.0);
#endif

        	//
        	// PID-P, PID-I, PID-D, PID-C, PID Autotune
        	// PID-P E1, PID-I E1, PID-D E1, PID-C E1, PID Autotune E1
        	// PID-P E2, PID-I E2, PID-D E2, PID-C E2, PID Autotune E2
        	// PID-P E3, PID-I E3, PID-D E3, PID-C E3, PID Autotune E3
        	// PID-P E4, PID-I E4, PID-D E4, PID-C E4, PID Autotune E4
        	//
#ifdef PIDTEMP

#define _PID_BASE_MENU_ITEMS(ELABEL, eindex) \
		eselect = eindex == 99 ? HEAT_BED:eindex;\
		raw_Ki = config.Get_Ki(eselect); \
		raw_Kd = config.Get_Kd(eselect); \
		raw_Kp = config.Get_Kp(eselect); \
		MENU_ITEM_EDIT_CALLBACK_T(float, T(MSG_PID_P ## eindex), &raw_Kp, 1, 9990,100,copy_and_scalePID_p_E ## eindex); \
		MENU_ITEM_EDIT_CALLBACK_T(float, T(MSG_PID_I ## eindex), &raw_Ki, 0.01, 9990, 100,copy_and_scalePID_i_E ## eindex); \
		MENU_ITEM_EDIT_CALLBACK_T(float,  T(MSG_PID_D ## eindex), &raw_Kd, 1, 9990, 100,copy_and_scalePID_d_E ## eindex)
#if defined PID_ADD_EXTRUSION_RATE && (eindex != 99)
#define _PID_MENU_ITEMS(ELABEL, eindex) \
		_PID_BASE_MENU_ITEMS(ELABEL, eindex); \
		MENU_ITEM_EDIT_T(float, T(MSG_PID_C ## eindex), &config.Kc[eindex], 0, 9990,1);\
		MENU_ITEM_EDIT_T(float, T(MSG_PID_F ## eindex), &config.Kf[eindex], 0, 1,50);
#endif

#ifdef PID_AUTOTUNE_MENU
#define PID_MENU_ITEMS(ELABEL, eindex) \
		_PID_MENU_ITEMS(ELABEL, eindex); \
		MENU_MULTIPLIER_ITEM_EDIT_CALLBACK(int3, MSG_PID_AUTOTUNE ELABEL, &autotune_temp[eindex], 150, heater_maxtemp[eindex] - 15, lcd_autotune_callback_E ## eindex)
#else
#define PID_MENU_ITEMS(ELABEL, eindex) _PID_MENU_ITEMS(ELABEL, eindex)
#endif

#if defined  PID_PARAMS_PER_EXTRUDER && (EXTRUDERS > 1)
        	PID_MENU_ITEMS(MSG_E1, 0);
        	PID_MENU_ITEMS(MSG_E2, 1);
#endif
#if TEMP_SENSOR_BED != 0
        	PID_MENU_ITEMS(MSG_E1, 99);
#endif
#endif //PIDTEMP
        	MENU_ITEM(submenu, T(MSG_Preheat_PLA), lcd_control_temperature_preheat_pla_settings_menu);
        	MENU_ITEM(submenu, T(MSG_Preheat_ABS), lcd_control_temperature_preheat_abs_settings_menu);
        	END_MENU();
        }

static void Config_StoreSettings()
{
	config.Store_EEPROM();
}

static void lcd_control_temperature_preheat_pla_settings_menu()
{
    START_MENU();
    MENU_ITEM(back, T(MSG_Temperature), lcd_control_temperature_menu);
    MENU_ITEM_EDIT_T(int, T(MSG_FanSpeed), &config.plaPreheatFanSpeed, 0, 255 ,1);
    MENU_ITEM_EDIT_T(int, T(MSG_Nozzle), &config.plaPreheatHotendTemp, 0, HEATER_0_MAXTEMP - 15,1);
#if TEMP_SENSOR_BED != 0
    MENU_ITEM_EDIT_T(int, T(MSG_Bed), &config.plaPreheatHPBTemp, 0, BED_MAXTEMP - 15,1);
#endif
#ifdef EEPROM_SETTINGS
    MENU_ITEM(function, T(MSG_Store_Eprom), Config_StoreSettings);
#endif
    END_MENU();
}

static void lcd_control_temperature_preheat_abs_settings_menu()
{
    START_MENU();
    MENU_ITEM(back, T(MSG_Temperature), lcd_control_temperature_menu);
    MENU_ITEM_EDIT_T(int, T(MSG_FanSpeed), &config.absPreheatFanSpeed, 0, 255 ,1);
    MENU_ITEM_EDIT_T(int, T(MSG_Nozzle), &config.absPreheatHotendTemp, 0, HEATER_0_MAXTEMP - 15,1);
#if TEMP_SENSOR_BED != 0
    MENU_ITEM_EDIT_T(int, T(MSG_Bed), &config.absPreheatHPBTemp, 0, BED_MAXTEMP - 15,1);
#endif
#ifdef EEPROM_SETTINGS
    MENU_ITEM(function, T(MSG_Store_Eprom), Config_StoreSettings);
#endif
    END_MENU();
}

static void reset_acceleration_rates()
{
	config.Reset_AccelerationRates();
}

static void lcd_control_motion_menu()
{
    START_MENU();
    MENU_ITEM(back, T(MSG_Control), lcd_control_menu);
#ifdef ENABLE_AUTO_BED_LEVELING
    MENU_ITEM_EDIT(float32, MSG_ZPROBE_ZOFFSET, &zprobe_zoffset, 0.5, 50);
#endif
    MENU_ITEM_EDIT_T(float, T(MSG_Accel), &config.acceleration, 500, 99000,0.1);
    MENU_ITEM_EDIT_T(float, T(MSG_VXY_Jerk), &config.max_jerk[X_AXIS], 1, 990,1);
    MENU_ITEM_EDIT_T(float, T(MSG_VZ_Jerk), &config.max_jerk[Z_AXIS], 0.1, 990,10);
    MENU_ITEM_EDIT_T(float, T(MSG_VE_Jerk), &config.max_jerk[E_AXIS], 1, 990,1);
    MENU_ITEM_EDIT_T(float, T(MSG_Vmax_X), &config.max_feedrate_mm_s[X_AXIS], 1, 999,1);
    MENU_ITEM_EDIT_T(float, T(MSG_Vmax_Y), &config.max_feedrate_mm_s[Y_AXIS], 1, 999,1);
    MENU_ITEM_EDIT_T(float, T(MSG_Vmax_Z), &config.max_feedrate_mm_s[Z_AXIS], 1, 999,1);
    MENU_ITEM_EDIT_T(float, T(MSG_Vmax_E), &config.max_feedrate_mm_s[E_AXIS], 1, 999,1);
    MENU_ITEM_EDIT_T(float, T(MSG_Vmin), &config.min_feedrate_mm_s, 0, 999,1);
    MENU_ITEM_EDIT_T(float, T(MSG_Vtrav_min), &config.min_travel_feedrate_mm_s, 0, 999,1);
    MENU_ITEM_EDIT_CALLBACK_T(long, T(MSG_Amax_X), &config.max_acceleration_units_per_sq_second[X_AXIS], 100, 99000, 0.01, reset_acceleration_rates);
    MENU_ITEM_EDIT_CALLBACK_T(long, T(MSG_Amax_Y), &config.max_acceleration_units_per_sq_second[Y_AXIS], 100, 99000, .001, reset_acceleration_rates);
    MENU_ITEM_EDIT_CALLBACK_T(long, T(MSG_Amax_Z), &config.max_acceleration_units_per_sq_second[Z_AXIS], 100, 99000, 0.01, reset_acceleration_rates);
    MENU_ITEM_EDIT_CALLBACK_T(long, T(MSG_Amax_E), &config.max_acceleration_units_per_sq_second[E_AXIS], 100, 99000, 10, reset_acceleration_rates);
    MENU_ITEM_EDIT_T(float, T(MSG_A_Retract), &config.retract_acceleration, 100, 99000,0.01);
    MENU_ITEM_EDIT_T(float, T(MSG_Xsteps), &config.axis_steps_per_unit[X_AXIS], 50, 9999,1);
    MENU_ITEM_EDIT_T(float, T(MSG_Ysteps), &config.axis_steps_per_unit[Y_AXIS], 50, 9999,1);
    MENU_ITEM_EDIT_T(float, T(MSG_Zsteps), &config.axis_steps_per_unit[Z_AXIS], 50, 9999,1);
    MENU_ITEM_EDIT_T(float, T(MSG_Esteps), &config.axis_steps_per_unit[E_AXIS], 50, 9999,1);
#ifdef ABORT_ON_ENDSTOP_HIT_FEATURE_ENABLED
    MENU_ITEM_EDIT(bool, MSG_ENDSTOP_ABORT, &abort_on_endstop_hit);
#endif
    END_MENU();
}

static void lcd_timer_menu()
{
START_MENU();
MENU_ITEM(back, T(MSG_Advanced), lcd_control_advanced_menu);
MENU_ITEM_EDIT_T(int, T(MSG_Flow_time), &Extrude_time, 0, 90,1);
MENU_ITEM(function, T(MSG_Start_Timer), lcd_flow_extrude_start);
MENU_ITEM(function, T(MSG_Stop_Timer), lcd_flow_extrude_stop);
END_MENU();
}

void set_Extruder_T0()
{
	service.Print(Service_Set_T0);
	watch_temp_callback_E0();
}

void set_Extruder_T1()
{
#if EXTRUDERS > 1
	service.Print(Service_Set_T1);
	watch_temp_callback_E1();
#endif
}

static void lcd_control_advanced_menu()
{
    START_MENU();
    MENU_ITEM(back, T(MSG_Control), lcd_control_menu);
#ifdef ENABLE_AUTO_BED_LEVELING
    MENU_ITEM_EDIT(float32, MSG_ZPROBE_ZOFFSET, &zprobe_zoffset, 0.5, 50);
#endif

    MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(int, T(MSG_Nozzle1), &heater.target_temperature[0], 0, HEATER_0_MAXTEMP - 15, 1, set_Extruder_T0);

  #if EXTRUDERS > 1
    MENU_MULTIPLIER_ITEM_EDIT_CALLBACK_T(int, T(MSG_Nozzle2), &heater.target_temperature[1], 0, HEATER_1_MAXTEMP - 15, 1, set_Extruder_T1);

#endif // EXTRUDERS > 1

    MENU_ITEM(submenu, T(MSG_Stopwatch), lcd_timer_menu);
    MENU_ITEM_EDIT_T(float, T(MSG_Adv_Press1), &config.advance_factor[0], 0, 255, 1.0);
    MENU_ITEM_EDIT_T(int, T(MSG_Adv_delay1), &config.advance_delay[0], 0, 255, 1);
    MENU_ITEM_EDIT_T(float, T(MSG_Adv_Press2), &config.advance_factor[1], 0, 255, 1.0);
    MENU_ITEM_EDIT_T(int, T(MSG_Adv_delay2), &config.advance_delay[1], 0, 255, 1);

    END_MENU();
}

#ifdef DOGLCD
static void lcd_set_contrast()
{
    if (encoderPosition != 0)
    {
        lcd_contrast -= encoderPosition;
        if (lcd_contrast < 0) lcd_contrast = 0;
        else if (lcd_contrast > 63) lcd_contrast = 63;
        encoderPosition = 0;
        lcdDrawUpdate = 1;
        u8g.setContrast(lcd_contrast);
    }
    if (lcdDrawUpdate)
    {
        lcd_implementation_drawedit(PSTR(MSG_CONTRAST), itostr2(lcd_contrast));
    }
    if (LCD_CLICKED)
    {
        lcd_quick_feedback();
        currentMenu = lcd_control_menu;
        encoderPosition = 0;
    }
}
#endif

#ifdef FWRETRACT
static void lcd_control_retract_menu()
{
    START_MENU();
    MENU_ITEM(back, MSG_CONTROL, lcd_control_menu);
    MENU_ITEM_EDIT(bool, MSG_AUTORETRACT, &autoretract_enabled);
    MENU_ITEM_EDIT(float52, MSG_CONTROL_RETRACT, &retract_length, 0, 100);
    MENU_ITEM_EDIT(float3, MSG_CONTROL_RETRACTF, &retract_feedrate, 1, 999);
    MENU_ITEM_EDIT(float52, MSG_CONTROL_RETRACT_ZLIFT, &retract_zlift, 0, 999);
    MENU_ITEM_EDIT(float52, MSG_CONTROL_RETRACT_RECOVER, &retract_recover_length, 0, 100);
    MENU_ITEM_EDIT(float3, MSG_CONTROL_RETRACT_RECOVERF, &retract_recover_feedrate, 1, 999);
    END_MENU();
}
#endif

#if SDCARDDETECT == -1
static void lcd_sd_refresh()
{
    sdcard.initsd();
    currentMenuViewOffset = 0;
}
#endif
static void lcd_sd_updir()
{
    sdcard.updir();
    currentMenuViewOffset = 0;
}

void lcd_sdcard_menu()
{
    if (lcdDrawUpdate == 0 && LCD_CLICKED == 0)
        return;  // nothing to do (so don't thrash the SD card)
    uint16_t fileCnt = sdcard.getnrfilenames();
    START_MENU();
    MENU_ITEM(back,T(MSG_Main_Menu) , lcd_main_menu);
    sdcard.getWorkDirName();
    if(sdcard.getWorkDirName()[0]=='/')
    {
#if SDCARDDETECT == -1
        MENU_ITEM(function, T(MSG_Refresh), lcd_sd_refresh);

#endif
    }else{
        MENU_ITEM(function, T(MSG_Folder_Up), lcd_sd_updir);
    }

    for(uint16_t i=0;i<fileCnt;i++)
    {
        if (_menuItemNr == _lineNr)
        {
            #ifndef SDCARD_RATHERRECENTFIRST
              sdcard.getfilename(i);
            #else
              sdcard.getfilename(fileCnt-1-i);
            #endif
            if (sdcard.IsDir())
            {
                MENU_ITEM(sddirectory, T(MSG_Card_Menu), sdcard.FileName());
            }else{
                MENU_ITEM(sdfile, T(MSG_Card_Menu), sdcard.FileName());
            }
        }else{
            MENU_ITEM_DUMMY();
        }
    }
    END_MENU();
}

template <class T> void menu_edit_T ()
{
	T temp;
    if ((int32_t)encoderPosition < minEditValue)
        encoderPosition = minEditValue;
    if ((int32_t)encoderPosition > maxEditValue)
        encoderPosition = maxEditValue;
    if (lcdDrawUpdate)
    {
    	temp =encoderPosition / scale;
          lcd_implementation_drawedit_T(editLabel, &(temp));
    }
    if (LCD_CLICKED)
    {
    	temp =encoderPosition / scale;
        *((T*)editValue) = temp;
        lcd_quick_feedback();
        currentMenu = prevMenu;
        encoderPosition = prevEncoderPosition;
    }
}

template <class T> void menu_edit_callback ()
{
	T temp;

    if ((int32_t)encoderPosition < minEditValue)
        encoderPosition = minEditValue;
    if ((int32_t)encoderPosition > maxEditValue)
        encoderPosition = maxEditValue;
    if (lcdDrawUpdate)
    {
    	temp =encoderPosition / scale;
          lcd_implementation_drawedit_T(editLabel, &(temp));
    }
     if (LCD_CLICKED)
    {
        *((T*)editValue) = ((T)encoderPosition) / scale;
        lcd_quick_feedback();
        currentMenu = prevMenu;
        encoderPosition = prevEncoderPosition;
        (*callbackFunc)();
    }
}



template <class T> static void menu_action_setting_edit (const char* pstr, T* ptr, float minValue, float maxValue , float nscale)
{
    prevMenu = currentMenu;
    prevEncoderPosition = encoderPosition;
    scale = nscale;
    lcdDrawUpdate = 2;
    currentMenu = menu_edit_T<T> ;

    editLabel = pstr;
    editValue = ptr;
    minEditValue = minValue * scale;
    maxEditValue = maxValue * scale;
    encoderPosition = (*ptr) * scale;
}

template <class T> static void menu_action_setting_edit_callback(const char* pstr, T* ptr, float minValue, float maxValue,float nscale, menuFunc_t callback)
{
	    prevMenu = currentMenu;
	    prevEncoderPosition = encoderPosition;
	    scale = nscale;

	    lcdDrawUpdate = 2;
    currentMenu = menu_edit_callback<T> ;

	    editLabel = pstr;
	    editValue = ptr;
	    minEditValue = minValue * scale;
	    maxEditValue = maxValue * scale;
	    encoderPosition = (*ptr) * scale;
	    callbackFunc = callback;
}

#ifdef REPRAPWORLD_KEYPAD
  static void reprapworld_keypad_move_z_up() {
    encoderPosition = 1;
    move_menu_scale = REPRAPWORLD_KEYPAD_MOVE_STEP;
    lcd_move_z();
  }
  static void reprapworld_keypad_move_z_down() {
    encoderPosition = -1;
    move_menu_scale = REPRAPWORLD_KEYPAD_MOVE_STEP;
    lcd_move_z();
  }
  static void reprapworld_keypad_move_x_left() {
    encoderPosition = -1;
    move_menu_scale = REPRAPWORLD_KEYPAD_MOVE_STEP;
    lcd_move_x();
  }
  static void reprapworld_keypad_move_x_right() {
    encoderPosition = 1;
    move_menu_scale = REPRAPWORLD_KEYPAD_MOVE_STEP;
    lcd_move_x();
  }
  static void reprapworld_keypad_move_y_down() {
    encoderPosition = 1;
    move_menu_scale = REPRAPWORLD_KEYPAD_MOVE_STEP;
    lcd_move_y();
  }
  static void reprapworld_keypad_move_y_up() {
    encoderPosition = -1;
    move_menu_scale = REPRAPWORLD_KEYPAD_MOVE_STEP;
    lcd_move_y();
  }
  static void reprapworld_keypad_move_home() {
    enquecommand_P((PSTR("G28"))); // move all axis home
  }
#endif

/** End of menus **/

static void lcd_quick_feedback()
{
    lcdDrawUpdate = 2;
    blocking_enc = millis() + 500;
    lcd_implementation_quick_feedback();
}

/** Menu action functions **/
static void menu_action_back(menuFunc_t data)
{
    currentMenu = data;
    encoderPosition = 0;
    encoderPosition = 0;my_printer.Reset_Status_Timeout();
}
static void menu_action_submenu(menuFunc_t data)
{
    currentMenu = data;
    encoderPosition = 0;my_printer.Reset_Status_Timeout();
}

static void menu_action_function(menuFunc_t data)
{
    (*data)();
}
static void menu_action_service(enum Services servicenr)
{
    service.Print(servicenr);
}
static void menu_action_sdfile(char* filename)
{
    char cmd[30];
//    char* c;
//    sprintf_P(cmd, PSTR("M23 %s"), filename);
//    for(c = &cmd[4]; *c; c++)
//
//    enquecommand(cmd);
//    enquecommand_P(PSTR("M24"));

	sdcard.openFile(sf_message(PSTR("%s%s"),sdcard.getPath(),sdcard.FileName()),true,true);
	service.Print(Service_Start_SD);
    lcd_return_to_status();
}
static void menu_action_sddirectory(const char* filename)
{
    sdcard.chdir(filename);
    encoderPosition = 0;
}
static void menu_action_setting_edit_bool(const char* pstr, bool* ptr)
{
    *ptr = !(*ptr);
}
#endif//ULTIPANEL

String logosplash[] = {
  "\xff\x20\xff\x20\x20\xff\xff\xff\x20\x20\xff\xff\xff\x20\x20\xff\xff\xff\x20\x20\xff\xff\xff\x20\x20\xff\x20\xff\x20\x20",
  "\xff\x20\xff\x20\x20\xff\xff\x20\x20\x20\xff\xff\xff\x20\x20\x20\xff\x20\x20\x20\xff\xff\x20\x20\x20\x20\xff\x20\x20\x20",
  "\xff\x20\xff\x20\x20\xff\x20\x20\x20\x20\xff\xff\x20\x20\x20\x20\xff\x20\x20\x20\xff\x20\x20\x20\x20\x20\xff\x20\x20\x20",
  "\x20\xff\x20\x20\x20\xff\xff\xff\x20\x20\xff\x20\xff\x20\x20\x20\xff\x20\x20\x20\xff\xff\xff\x20\x20\xff\x20\xff\x20\x20"
};

void lcd_firmwarescreen()
{ 
  lcd.display();
  lcd.setCursor(0, 0);
  lcd.print((VERSION_SPLASH_NAME));
  lcd.setCursor(0, 1);
  lcd.print((VERSION_SPLASH_FIRMWARE));
  lcd.setCursor(0, 2);
  lcd.print((VERSION_SPLASH_WEBSITE1));
  lcd.setCursor(0, 3);
  lcd.print((VERSION_SPLASH_WEBSITE2));
}

void lcd_firmware_menu()
{
  START_MENU();
  MENU_ITEM(function, PSTR(VERSION_SPLASH_NAME_SMALL), lcd_return_to_status);
  MENU_ITEM(function, PSTR(VERSION_SPLASH_FIRMWARE_SMALL), lcd_return_to_status);
  MENU_ITEM(function, PSTR(VERSION_SPLASH_WEBSITE1_SMALL), lcd_return_to_status);
  MENU_ITEM(function, PSTR(VERSION_SPLASH_WEBSITE2_SMALL), lcd_return_to_status);
  END_MENU();
}

void lcd_splashscreen()
{ 
  lcd_implementation_init();   //
  lcd.clear();
  
  pinMode(BTN_ENC,INPUT);
  pinMode(BTN_EN1,INPUT);
  pinMode(BTN_EN2,INPUT);
  WRITE(BTN_EN1,HIGH);
  WRITE(BTN_EN2,HIGH);
  WRITE(BTN_ENC,HIGH);
  
  #ifdef LEDCONTROL
   pinMode(REDPIN, OUTPUT);
   pinMode(GREENPIN, OUTPUT);
   pinMode(BLUEPIN, OUTPUT);
   digitalWrite(REDPIN, REDSTARTVAL);
   digitalWrite(GREENPIN, GREENSTARTVAL);
   digitalWrite(BLUEPIN, BLUESTARTVAL);
  #endif
 
  int width = 20;
  int stp = 2;
  int lngth = 30;
  
    for (int m=width-1; m>=0; m-=stp) {
    for (int i=0; i<4; i++) {
      String line = logosplash[i];
      lcd.setCursor(m, i);
      lcd.print(line.substring(0, width - m));
    }
    lcd.display();
    for (int n=0; n<255; n++) {
      delay(1);
      if((READ(BTN_ENC)==0) || (READ(BTN_EN1)==0) || (READ(BTN_EN2)==0))
      {
        goto bailout;
      }
    }
    lcd.noDisplay();
   }
  
    for (int m=1; m<lngth; m+=stp) {
      for (int i=0; i<4; i++) {
        String line = logosplash[i];
        lcd.setCursor(0, i);
        lcd.print(line.substring(m, 20+m));
      }
      lcd.display();
      for (int b=0; b<255; b++) {
        delay(1);
        if((READ(BTN_ENC)==0) || (READ(BTN_EN1)==0) || (READ(BTN_EN2)==0))
        {
          goto bailout;
        }
      }
      lcd.noDisplay();
    } 

  lcd_firmwarescreen();
  
  for (int b=0; b<255; b++) {
        delay(11);
        if((READ(BTN_ENC)==0) || (READ(BTN_EN1)==0) || (READ(BTN_EN2)==0))
        {
          goto bailout;
        }
      }
      
  bailout:
  
  lcd.clear();
  _delay_ms(100);
  
}

/** LCD API **/
void lcd_init()
{
    lcd_implementation_init();

#ifdef NEWPANEL
    pinMode(BTN_EN1,INPUT);
    pinMode(BTN_EN2,INPUT);
    WRITE(BTN_EN1,HIGH);
    WRITE(BTN_EN2,HIGH);
  #if BTN_ENC > 0
    pinMode(BTN_ENC,INPUT);
    WRITE(BTN_ENC,HIGH);
  #endif
  #ifdef REPRAPWORLD_KEYPAD
    pinMode(SHIFT_CLK,OUTPUT);
    pinMode(SHIFT_LD,OUTPUT);
    pinMode(SHIFT_OUT,INPUT);
    WRITE(SHIFT_OUT,HIGH);
    WRITE(SHIFT_LD,HIGH);
  #endif
#else  // Not NEWPANEL
  #ifdef SR_LCD_2W_NL // Non latching 2 wire shift register
     pinMode (SR_DATA_PIN, OUTPUT);
     pinMode (SR_CLK_PIN, OUTPUT);
  #elif defined(SHIFT_CLK) 
     pinMode(SHIFT_CLK,OUTPUT);
     pinMode(SHIFT_LD,OUTPUT);
     pinMode(SHIFT_EN,OUTPUT);
     pinMode(SHIFT_OUT,INPUT);
     WRITE(SHIFT_OUT,HIGH);
     WRITE(SHIFT_LD,HIGH);
     WRITE(SHIFT_EN,LOW);
  #else
     #ifdef ULTIPANEL
     #error ULTIPANEL requires an encoder
     #endif
  #endif // SR_LCD_2W_NL
#endif//!NEWPANEL

#if defined (SDSUPPORT) && defined(SDCARDDETECT) && (SDCARDDETECT > 0)
    pinMode(SDCARDDETECT,INPUT);
    WRITE(SDCARDDETECT, HIGH);
    lcd_oldcardstatus = IS_SD_INSERTED;
#endif//(SDCARDDETECT > 0)
#ifdef LCD_HAS_SLOW_BUTTONS
    slow_buttons = 0;
#endif
    lcd_buttons_update();
#ifdef ULTIPANEL
    encoderDiff = 0;
#endif
    lcd_setstatuspgm (T(MSG_Welcome));
    my_printer.Reset_Status_Timeout();
}


void lcd_update()
{
//    static unsigned long timeoutToStatus = 0;

    #ifdef LCD_HAS_SLOW_BUTTONS
    slow_buttons = lcd_implementation_read_slow_buttons(); // buttons which take too long to read in interrupt context
    #endif

    lcd_buttons_update();

    #if (SDCARDDETECT > 0)
    if((IS_SD_INSERTED != lcd_oldcardstatus))
    {
        lcdDrawUpdate = 2;
        lcd_oldcardstatus = IS_SD_INSERTED;
        lcd_implementation_init(); // to maybe revive the LCD if static electricity killed it.

        if(lcd_oldcardstatus)
        {
            card.initsd();
            LCD_MESSAGEPGM(MSG_SD_INSERTED);
        }
        else
        {
            card.release();
            LCD_MESSAGEPGM(MSG_SD_REMOVED);
        }
    }
    #endif//CARDINSERTED

    if (my_printer.LCD_Timeout())
    {
#ifdef ULTIPANEL
    #ifdef REPRAPWORLD_KEYPAD
          if (REPRAPWORLD_KEYPAD_MOVE_Z_UP) {
            reprapworld_keypad_move_z_up();
          }
          if (REPRAPWORLD_KEYPAD_MOVE_Z_DOWN) {
            reprapworld_keypad_move_z_down();
          }
          if (REPRAPWORLD_KEYPAD_MOVE_X_LEFT) {
            reprapworld_keypad_move_x_left();
          }
          if (REPRAPWORLD_KEYPAD_MOVE_X_RIGHT) {
            reprapworld_keypad_move_x_right();
          }
          if (REPRAPWORLD_KEYPAD_MOVE_Y_DOWN) {
            reprapworld_keypad_move_y_down();
          }
          if (REPRAPWORLD_KEYPAD_MOVE_Y_UP) {
            reprapworld_keypad_move_y_up();
          }
          if (REPRAPWORLD_KEYPAD_MOVE_HOME) {
            reprapworld_keypad_move_home();
          }
    #endif
        if (abs(encoderDiff) >= ENCODER_PULSES_PER_STEP)
        {
            lcdDrawUpdate = 1;
            encoderPosition += encoderDiff / ENCODER_PULSES_PER_STEP;
            encoderDiff = 0;
            my_printer.Reset_Status_Timeout(5000);
        }
        if (LCD_CLICKED)
        {
        	if (currentMenu == lcd_timer_menu)
        		my_printer.Reset_Status_Timeout((4*(unsigned long)LCD_TIMEOUT_TO_STATUS));
        	else
        		my_printer.Reset_Status_Timeout(LCD_TIMEOUT_TO_STATUS);
        }
#endif//ULTIPANEL

#ifdef DOGLCD        // Changes due to different driver architecture of the DOGM display
        blink++;     // Variable for fan animation and alive dot
        u8g.firstPage();
        do
        {
            u8g.setFont(u8g_font_6x10_marlin);
            u8g.setPrintPos(125,0);
            if (blink % 2) u8g.setColorIndex(1); else u8g.setColorIndex(0); // Set color for the alive dot
            u8g.drawPixel(127,63); // draw alive dot
            u8g.setColorIndex(1); // black on white
            (*currentMenu)();
            if (!lcdDrawUpdate)  break; // Terminate display update, when nothing new to draw. This must be done before the last dogm.next()
        } while( u8g.nextPage() );
#else
        (*currentMenu)();
#endif

#ifdef LCD_HAS_STATUS_INDICATORS
        lcd_implementation_update_indicators();
#endif

#ifdef ULTIPANEL
        if (my_printer.Status_Timeout())
        {
        	lcd_reset_alert_level();

        	if(currentMenu != lcd_status_screen)
        	{
        		my_printer.Reset_Status_Timeout();
        		lcd_return_to_status();
        		lcdDrawUpdate = 2;
        	}
        }
#endif//ULTIPANEL
        if (lcdDrawUpdate == 2)
            lcd_implementation_clear();
        if (lcdDrawUpdate)
            lcdDrawUpdate--;

        blink = !(Position.axis_known_position[X_AXIS] && Position.axis_known_position[Y_AXIS] && Position.axis_known_position[Z_AXIS]);

        my_printer.Reset_LCD_Timeout();
    }
}

void set_utf_strlen(char* s, uint8_t n) {
  uint8_t i = 0, j = 0;
  while (s[i] && (j < n)) {
    if ((s[i] & 0xc0u) != 0x80u) j++;
    i++;
  }
  while (j++ < n) s[i++] = ' ';
  s[i] = 0;
}

void lcd_setstatus(char* message)
{
    if (lcd_status_message_level > 0)
        return;
    strncpy(lcd_status_message, message, 3*LCD_WIDTH);
    set_utf_strlen(lcd_status_message, LCD_WIDTH);
    lcdDrawUpdate = 2;
}
void lcd_setstatuspgm(const char* message)
{
    if (lcd_status_message_level > 0)
        return;
    strncpy_P(lcd_status_message, message, 3*LCD_WIDTH);
    set_utf_strlen(lcd_status_message, LCD_WIDTH);
    lcdDrawUpdate = 2;
}
void lcd_setalertstatuspgm(const char* message)
{
    lcd_setstatuspgm(message);
    if (lcd_status_message_level == 0)
        lcdDrawUpdate = 2;
    lcd_status_message_level = 1;
#ifdef ULTIPANEL
//    lcd_return_to_status();
#endif//ULTIPANEL
}
void lcd_reset_alert_level()
{
    lcd_status_message_level = 0;
}

#ifdef DOGLCD
void lcd_setcontrast(uint8_t value)
{
    lcd_contrast = value & 63;
    u8g.setContrast(lcd_contrast);
}
#endif

#ifdef ULTIPANEL
/* Warning: This function is called from interrupt context */
void lcd_buttons_update()
{
#ifdef NEWPANEL
    uint8_t newbutton=0;
    if(READ(BTN_EN1)==0)  newbutton|=EN_A;
    if(READ(BTN_EN2)==0)  newbutton|=EN_B;
  #if BTN_ENC > 0
    if((blocking_enc<millis()) && (READ(BTN_ENC)==0))
    {
        newbutton |= EN_C;
        my_printer.Reset_Status_Timeout();
    }
  #endif
    buttons = newbutton;
    #ifdef LCD_HAS_SLOW_BUTTONS
    buttons |= slow_buttons;
    #endif
    #ifdef REPRAPWORLD_KEYPAD
      // for the reprapworld_keypad
      uint8_t newbutton_reprapworld_keypad=0;
      WRITE(SHIFT_LD,LOW);
      WRITE(SHIFT_LD,HIGH);
      for(int8_t i=0;i<8;i++) {
          newbutton_reprapworld_keypad = newbutton_reprapworld_keypad>>1;
          if(READ(SHIFT_OUT))
              newbutton_reprapworld_keypad|=(1<<7);
          WRITE(SHIFT_CLK,HIGH);
          WRITE(SHIFT_CLK,LOW);
      }
      buttons_reprapworld_keypad=~newbutton_reprapworld_keypad; //invert it, because a pressed switch produces a logical 0
  #endif
#else   //read it from the shift register
    uint8_t newbutton=0;
    WRITE(SHIFT_LD,LOW);
    WRITE(SHIFT_LD,HIGH);
    unsigned char tmp_buttons=0;
    for(int8_t i=0;i<8;i++)
    {
        newbutton = newbutton>>1;
        if(READ(SHIFT_OUT))
            newbutton|=(1<<7);
        WRITE(SHIFT_CLK,HIGH);
        WRITE(SHIFT_CLK,LOW);
    }
    buttons=~newbutton; //invert it, because a pressed switch produces a logical 0
#endif//!NEWPANEL

    //manage encoder rotation
    uint8_t enc=0;
    if(buttons&EN_A)
        enc|=(1<<0);
    if(buttons&EN_B)
        enc|=(1<<1);
    if(enc != lastEncoderBits)
    {
        switch(enc)
        {
        case encrot0:
            if(lastEncoderBits==encrot3)
                encoderDiff++;
            else if(lastEncoderBits==encrot1)
                encoderDiff--;
            break;
        case encrot1:
            if(lastEncoderBits==encrot0)
                encoderDiff++;
            else if(lastEncoderBits==encrot2)
                encoderDiff--;
            break;
        case encrot2:
            if(lastEncoderBits==encrot1)
                encoderDiff++;
            else if(lastEncoderBits==encrot3)
                encoderDiff--;
            break;
        case encrot3:
            if(lastEncoderBits==encrot2)
                encoderDiff++;
            else if(lastEncoderBits==encrot0)
                encoderDiff--;
            break;
        }
    }
    lastEncoderBits = enc;
}

void lcd_buzz(long duration, uint16_t freq)
{
#ifdef LCD_USE_I2C_BUZZER
  lcd.buzz(duration,freq);
#endif
}

bool lcd_clicked()
{
  return LCD_CLICKED;
}
#endif//ULTIPANEL

/********************************/
/** Float conversion utilities **/
/********************************/
//  convert float to string with +123.4 format
char conv[8];
char *ftostr3(const float &x)
{
  return itostr3((int)x);
}

// Convert float to rj string with _123, -123, _-12, or __-1 format
char *ftostr4sign(const float& x) { return itostr4sign((int)x); }

char *itostr2(const uint8_t &x)
{
  //sprintf(conv,"%5.1f",x);
  int xx=x;
  conv[0]=(xx/10)%10+'0';
  conv[1]=(xx)%10+'0';
  conv[2]=0;
  return conv;
}

//  convert float to string with +123.4 format
char *ftostr31(const float &x)
{
  int xx=x*10;
  conv[0]=(xx>=0)?'+':'-';
  xx=abs(xx);
  conv[1]=(xx/1000)%10+'0';
  conv[2]=(xx/100)%10+'0';
  conv[3]=(xx/10)%10+'0';
  conv[4]='.';
  conv[5]=(xx)%10+'0';
  conv[6]=0;
  return conv;
}

//  convert float to string with 123.4 format
char *ftostr31ns(const float &x)
{
  int xx=x*10;
  //conv[0]=(xx>=0)?'+':'-';
  xx=abs(xx);
  conv[0]=(xx/1000)%10+'0';
  conv[1]=(xx/100)%10+'0';
  conv[2]=(xx/10)%10+'0';
  conv[3]='.';
  conv[4]=(xx)%10+'0';
  conv[5]=0;
  return conv;
}

char *ftostr32(const float &x)
{
  long xx=x*100;
  if (xx >= 0)
    conv[0]=(xx/10000)%10+'0';
  else
    conv[0]='-';
  xx=abs(xx);
  conv[1]=(xx/1000)%10+'0';
  conv[2]=(xx/100)%10+'0';
  conv[3]='.';
  conv[4]=(xx/10)%10+'0';
  conv[5]=(xx)%10+'0';
  conv[6]=0;
  return conv;
}

// Convert signed float to space-padded string with -_23.4_ format
char* ftostr32sp(const float& x) {
  long xx = x * 100;
  uint8_t dig;
  if (xx < 0) { // negative val = -_0
    xx = -xx;
    conv[0] = '-';
    dig = (xx / 1000) % 10;
    conv[1] = dig ? '0' + dig : ' ';
  }
  else { // positive val = __0
    dig = (xx / 10000) % 10;
    if (dig) {
      conv[0] = '0' + dig;
      conv[1] = '0' + (xx / 1000) % 10;
    }
    else {
      conv[0] = ' ';
      dig = (xx / 1000) % 10;
      conv[1] = dig ? '0' + dig : ' ';
    }
  }

  conv[2] = '0' + (xx / 100) % 10; // lsd always

  dig = xx % 10;
  if (dig) { // 2 decimal places
    conv[5] = '0' + dig;
    conv[4] = '0' + (xx / 10) % 10;
    conv[3] = '.';
  }
  else { // 1 or 0 decimal place
    dig = (xx / 10) % 10;
    if (dig) {
      conv[4] = '0' + dig;
      conv[3] = '.';
    }
    else {
      conv[3] = conv[4] = ' ';
    }
    conv[5] = ' ';
  }
  conv[6] = '\0';
  return conv;
}

char *itostr31(const int &xx)
{
  conv[0]=(xx>=0)?'+':'-';
  conv[1]=(xx/1000)%10+'0';
  conv[2]=(xx/100)%10+'0';
  conv[3]=(xx/10)%10+'0';
  conv[4]='.';
  conv[5]=(xx)%10+'0';
  conv[6]=0;
  return conv;
}

char *itostr3(const int &xx)
{
  if (xx >= 100)
    conv[0]=(xx/100)%10+'0';
  else
    conv[0]=' ';
  if (xx >= 10)
    conv[1]=(xx/10)%10+'0';
  else
    conv[1]=' ';
  conv[2]=(xx)%10+'0';
  conv[3]=0;
  return conv;
}

char *itostr3left(const int &xx)
{
  if (xx >= 100)
  {
    conv[0]=(xx/100)%10+'0';
    conv[1]=(xx/10)%10+'0';
    conv[2]=(xx)%10+'0';
    conv[3]=0;
  }
  else if (xx >= 10)
  {
    conv[0]=(xx/10)%10+'0';
    conv[1]=(xx)%10+'0';
    conv[2]=0;
  }
  else
  {
    conv[0]=(xx)%10+'0';
    conv[1]=0;
  }
  return conv;
}

char *itostr4(const int &xx)
{
  if (xx >= 1000)
    conv[0]=(xx/1000)%10+'0';
  else
    conv[0]=' ';
  if (xx >= 100)
    conv[1]=(xx/100)%10+'0';
  else
    conv[1]=' ';
  if (xx >= 10)
    conv[2]=(xx/10)%10+'0';
  else
    conv[2]=' ';
  conv[3]=(xx)%10+'0';
  conv[4]=0;
  return conv;
}
// Convert signed int to rj string with _123, -123, _-12, or __-1 format
char *itostr4sign(const int& x) {
  int xx = abs(x);
  int sign = 0;
  if (xx >= 100) {
    conv[1] = (xx / 100) % 10 + '0';
    conv[2] = (xx / 10) % 10 + '0';
  }
  else if (xx >= 10) {
    conv[0] = ' ';
    sign = 1;
    conv[2] = (xx / 10) % 10 + '0';
  }
  else {
    conv[0] = ' ';
    conv[1] = ' ';
    sign = 2;
  }
  conv[sign] = x < 0 ? '-' : ' ';
  conv[3] = xx % 10 + '0';
  conv[4] = 0;
  return conv;
}

//  convert float to string with 12345 format
char *ftostr5(const float &x)
{
  long xx=abs(x);
  if (xx >= 10000)
    conv[0]=(xx/10000)%10+'0';
  else
    conv[0]=' ';
  if (xx >= 1000)
    conv[1]=(xx/1000)%10+'0';
  else
    conv[1]=' ';
  if (xx >= 100)
    conv[2]=(xx/100)%10+'0';
  else
    conv[2]=' ';
  if (xx >= 10)
    conv[3]=(xx/10)%10+'0';
  else
    conv[3]=' ';
  conv[4]=(xx)%10+'0';
  conv[5]=0;
  return conv;
}

//  convert float to string with +1234.5 format
char *ftostr51(const float &x)
{
  long xx=x*10;
  conv[0]=(xx>=0)?'+':'-';
  xx=abs(xx);
  conv[1]=(xx/10000)%10+'0';
  conv[2]=(xx/1000)%10+'0';
  conv[3]=(xx/100)%10+'0';
  conv[4]=(xx/10)%10+'0';
  conv[5]='.';
  conv[6]=(xx)%10+'0';
  conv[7]=0;
  return conv;
}

//  convert float to string with +123.45 format
char *ftostr52(const float &x)
{
  long xx=x*100;
  conv[0]=(xx>=0)?'+':'-';
  xx=abs(xx);
  conv[1]=(xx/10000)%10+'0';
  conv[2]=(xx/1000)%10+'0';
  conv[3]=(xx/100)%10+'0';
  conv[4]='.';
  conv[5]=(xx/10)%10+'0';
  conv[6]=(xx)%10+'0';
  conv[7]=0;
  return conv;
}

char* sf_message(const char* formatstr, ...)
{
	static char buffer[40];
	static char format[10];
	char ch;
	int bufp = 0;
	int formp = 0;
	bool resolution = false;
	int length = 0;
	int digits = 2;
	va_list ap;
	va_start (ap, formatstr);
	ch=pgm_read_byte(formatstr);
	while(ch)
	{
		if (ch == '%')
		{
			// handle format string
			digits = 2;
			length = 0;
			resolution = false;
			formp=0;
			format[formp++] = ch;
			while (true)
			{
				ch=pgm_read_byte(++formatstr);
				if (ch == '%')
				{
					buffer[bufp++] = ch; // is only a '%'
					break;
				}
				if (isdigit(ch))
				{
					format[formp++] = ch;
					if (resolution)
					{
						digits = ch - '0';
					}
					else
					{
						length = ch -'0';
					}
				}
				if (ch == '.')
				{
					format[formp++] = ch;
					resolution = true;
				}
				if (ch == 'i' || ch == 'I')
				{
					format[formp++] = ch;
					format[formp++] = '\0';
					bufp += sprintf(&buffer[bufp], &format[0], va_arg(ap, int));
					//					    ch=pgm_read_byte(++formatstr);
					break;
				}
				if (ch == 'l' || ch == 'L')
				{
					format[formp++] = ch;
					format[formp++] = '\0';
					bufp += sprintf(&buffer[bufp], &format[0], va_arg(ap, long));
					//					    ch=pgm_read_byte(++formatstr);
					break;
				}
				if (ch == 'f' || ch == 'F')
				{
//					size_t n = 0;
					double number = va_arg(ap,double);

					//				    	  if (isnan(number)) return print("nan");
					//				    	  if (isinf(number)) return print("inf");
					//				    	  if (number > 4294967040.0) return print ("ovf");  // constant determined empirically
					//				    	  if (number <-4294967040.0) return print ("ovf");  // constant determined empirically

					// Handle negative numbers
					if (number < 0.0)
					{
						buffer[bufp++] = '-';
						number = -number;
					}

					// Round correctly so that print(1.999, 2) prints as "2.00"
					double rounding = 0.5;
					for (uint8_t i=0; i<digits; ++i)
						rounding /= 10.0;

					number += rounding;

					// Extract the integer part of the number and print it
					unsigned long int_part = (unsigned long)number;
					double remainder = number - (double)int_part;
					switch (length)
					{
					case 1:
						bufp += sprintf_P(&buffer[bufp], PSTR("%1i"), int_part);
						break;
					case 2:
						bufp += sprintf_P(&buffer[bufp], PSTR("%2i"), int_part);
						break;
					case 3:
						bufp += sprintf_P(&buffer[bufp], PSTR("%3i"), int_part);
						break;
					default:
						bufp += sprintf_P(&buffer[bufp], PSTR("%i"), int_part);

					}
					// Print the decimal point, but only if there are digits beyond
					if (digits > 0) {
						buffer[bufp++] = '.';
					}

					// Extract digits from the remainder one at a time
					while (digits-- > 0)
					{
						remainder *= 10.0;
						int toPrint = int(remainder);
						buffer[bufp++] = '0'+toPrint;
						remainder -= toPrint;
					}

					//						    ch=pgm_read_byte(++formatstr);
					break;
				}
				if (ch == 's' || ch == 'S')
				{
					char *str = va_arg(ap,char*);
					while (*str)
					{
						buffer[bufp++] = *str++;
					}
					break;
				}
			}
		}
		else
			buffer[bufp++] = ch;
		ch=pgm_read_byte(++formatstr);
	}
	va_end (ap);
	buffer[bufp++] = '\0';
return &buffer[0];
}



#endif //ULTRA_LCD
